#!/bin/sh

#   configure the requirements
AMV="automake (GNU automake) 1.15"
ACV="autoconf (GNU Autoconf) 2.69"
LTV="libtoolize (GNU libtool) 2.4.6"
GTT="gettextize (GNU gettext-tools) 0.19.8.1"
USAGE="
To build RPM from plain CVS sources the following
installed developer tools are mandatory:
    $AMV
    $ACV
    $LTV
    $GTT
"

#   wrapper for running GNU libtool's libtoolize(1)
libtoolize () {
    _libtoolize=`which glibtoolize 2>/dev/null`
    _libtoolize_args="$*"
    case "$_libtoolize" in
        /* ) ;;
        *  ) _libtoolize=`which libtoolize 2>/dev/null`
             case "$_libtoolize" in
                 /* ) ;;
                 *  ) _libtoolize="libtoolize" ;;
             esac
             ;;
    esac
    _libtoolize_version="`$_libtoolize --version | sed -e '1q' | sed -e 's;^[^0-9]*;;'`"
    case "$_libtoolize_version" in
        1.* ) _libtoolize_args=`echo "X$_libtoolize_args" | sed -e 's;^X;;' -e 's;--quiet;;' -e 's;--install;;'` ;;
    esac
    eval $_libtoolize $_libtoolize_args
}

#   requirements sanity check
[ "`automake   --version | head -1`" != "$AMV" ] && echo "$USAGE" # && exit 1
[ "`autoconf   --version | head -1`" != "$ACV" ] && echo "$USAGE" # && exit 1
[ "`libtoolize --version | head -1`" != "$LTV" ] && echo "$USAGE" # && exit 1
[ "`gettextize --version | head -1 | sed -e 's;^.*/\\(gettextize\\);\\1;'`" != "$GTT" ] && echo "$USAGE" # && exit 1

gnulib_lgpl3='
_Exit
accept4
acos
acosf
acosl
alphasort
amemxfrm
argp
asin
asinf
asinl
astrxfrm
atan
atan2
atan2f
atanf
atanl
atoll
c-strstr
canon-host
cbrt
cbrt-ieee
cbrtf
cbrtf-ieee
cbrtl
cbrtl-ieee
ceil
ceil-ieee
ceilf
ceilf-ieee
ceill
ceill-ieee
closedir
config-h
copysign
copysignf
copysignl
cos
cosf
cosh
coshf
cosl
crc
dprintf
dprintf-posix
dup3
duplocale
erf
erfc
error
exp
exp-ieee
exp2
exp2-ieee
exp2f
exp2f-ieee
exp2l
exp2l-ieee
expf
expf-ieee
expl
expl-ieee
expm1
expm1-ieee
expm1f
expm1f-ieee
expm1l
expm1l-ieee
fabs
fabs-ieee
fabsf
fabsf-ieee
fabsl
fabsl-ieee
fbufmode
fchdir
filevercmp
floor
floor-ieee
floorf
floorf-ieee
floorl
floorl-ieee
fma
fma-ieee
fmaf
fmaf-ieee
fmal
fmal-ieee
fmod
fmod-ieee
fmodf
fmodf-ieee
fmodl
fmodl-ieee
forkpty
fpieee
fprintf-posix
fpucw
freadable
freopen
frexp
frexp-ieee
frexp-nolibm
frexpf
frexpf-ieee
frexpl
frexpl-ieee
frexpl-nolibm
fseterr
fsusage
fwritable
fwriting
getpass-gnu
getrusage
getsubopt
hypot
hypot-ieee
hypotf
hypotf-ieee
hypotl
hypotl-ieee
i-ring
iconv_open-utf
ilogb
ilogbf
ilogbl
imaxabs
imaxdiv
isapipe
isfinite
isinf
isnan
isnand
isnand-nolibm
isnanf
isnanf-nolibm
isnanl
isnanl-nolibm
j0
j1
jn
ldexp-ieee
ldexpf
ldexpf-ieee
ldexpl
ldexpl-ieee
lgamma
libunistring
libunistring-optional
log
log-ieee
log10
log10-ieee
log10f
log10f-ieee
log10l
log10l-ieee
log1p
log1p-ieee
log1pf
log1pf-ieee
log1pl
log1pl-ieee
log2
log2-ieee
log2f
log2f-ieee
log2l
log2l-ieee
logb
logb-ieee
logbf
logbf-ieee
logbl
logbl-ieee
logf
logf-ieee
login_tty
logl
logl-ieee
mbchar
mbfile
mbiter
mbmemcasecmp
mbrlen
mbscasecmp
mbscasestr
mbschr
mbscspn
mbslen
mbsncasecmp
mbsnlen
mbsnrtowcs
mbspbrk
mbspcasecmp
mbsrchr
mbssep
mbsspn
mbsstr
mbstok_r
mbuiter
memcmp2
memrchr
mkfifo
mknod
modf
modf-ieee
modff
modff-ieee
modfl
modfl-ieee
nanosleep
nextafter
no-c++
obstack
opendir
openmp
pclose
popen
pow
powf
printf-frexp
printf-frexpl
printf-posix
printf-safe
putenv
random
readdir
remainder
remainder-ieee
remainderf
remainderf-ieee
remainderl
remainderl-ieee
remove
rewinddir
rint
rint-ieee
rintf
rintf-ieee
rintl
rintl-ieee
round
round-ieee
roundf
roundf-ieee
roundl
roundl-ieee
scandir
setlocale
signbit
sin
sinf
sinh
sinhf
sinl
snprintf-posix
sprintf-posix
sqrt
sqrt-ieee
sqrtf
sqrtf-ieee
sqrtl
sqrtl-ieee
strftime
strtod
strtoimax
strtoll
strtoull
strtoumax
symlink
sys_resource
system-posix
tan
tanf
tanh
tanhf
tanl
tcgetsid
time_rz
tmpdir
tmpfile
towctrans
trunc
trunc-ieee
truncf
truncf-ieee
truncl
truncl-ieee
tsearch
tzset
ucs4-utf16
ucs4-utf8
unicase/base
unicase/cased
unicase/empty-prefix-context
unicase/empty-suffix-context
unicase/ignorable
unicase/locale-language
unicase/special-casing
unicase/tocasefold
unicase/tolower
unicase/totitle
unicase/toupper
unicase/u16-casecmp
unicase/u16-casecoll
unicase/u16-casefold
unicase/u16-casemap
unicase/u16-casexfrm
unicase/u16-ct-casefold
unicase/u16-ct-tolower
unicase/u16-ct-totitle
unicase/u16-ct-toupper
unicase/u16-is-cased
unicase/u16-is-casefolded
unicase/u16-is-invariant
unicase/u16-is-lowercase
unicase/u16-is-titlecase
unicase/u16-is-uppercase
unicase/u16-prefix-context
unicase/u16-suffix-context
unicase/u16-tolower
unicase/u16-totitle
unicase/u16-toupper
unicase/u32-casecmp
unicase/u32-casecoll
unicase/u32-casefold
unicase/u32-casemap
unicase/u32-casexfrm
unicase/u32-ct-casefold
unicase/u32-ct-tolower
unicase/u32-ct-totitle
unicase/u32-ct-toupper
unicase/u32-is-cased
unicase/u32-is-casefolded
unicase/u32-is-invariant
unicase/u32-is-lowercase
unicase/u32-is-titlecase
unicase/u32-is-uppercase
unicase/u32-prefix-context
unicase/u32-suffix-context
unicase/u32-tolower
unicase/u32-totitle
unicase/u32-toupper
unicase/u8-casecmp
unicase/u8-casecoll
unicase/u8-casefold
unicase/u8-casemap
unicase/u8-casexfrm
unicase/u8-ct-casefold
unicase/u8-ct-tolower
unicase/u8-ct-totitle
unicase/u8-ct-toupper
unicase/u8-is-cased
unicase/u8-is-casefolded
unicase/u8-is-invariant
unicase/u8-is-lowercase
unicase/u8-is-titlecase
unicase/u8-is-uppercase
unicase/u8-prefix-context
unicase/u8-suffix-context
unicase/u8-tolower
unicase/u8-totitle
unicase/u8-toupper
unicase/ulc-casecmp
unicase/ulc-casecoll
unicase/ulc-casexfrm
uniconv/u16-conv-from-enc
uniconv/u16-conv-to-enc
uniconv/u16-strconv-from-enc
uniconv/u16-strconv-from-locale
uniconv/u16-strconv-to-enc
uniconv/u16-strconv-to-locale
uniconv/u32-conv-from-enc
uniconv/u32-conv-to-enc
uniconv/u32-strconv-from-enc
uniconv/u32-strconv-from-locale
uniconv/u32-strconv-to-enc
uniconv/u32-strconv-to-locale
uniconv/u8-conv-to-enc
uniconv/u8-strconv-to-enc
uniconv/u8-strconv-to-locale
unictype/bidiclass-all
unictype/bidiclass-byname
unictype/bidiclass-longname
unictype/bidiclass-name
unictype/bidiclass-test
unictype/block-all
unictype/block-list
unictype/block-of
unictype/block-test
unictype/category-C
unictype/category-Cc
unictype/category-Cf
unictype/category-Cn
unictype/category-Co
unictype/category-Cs
unictype/category-L
unictype/category-LC
unictype/category-Ll
unictype/category-Lm
unictype/category-Lo
unictype/category-Lt
unictype/category-Lu
unictype/category-Mc
unictype/category-Me
unictype/category-Mn
unictype/category-N
unictype/category-Nd
unictype/category-Nl
unictype/category-No
unictype/category-P
unictype/category-Pc
unictype/category-Pd
unictype/category-Pe
unictype/category-Pf
unictype/category-Pi
unictype/category-Po
unictype/category-Ps
unictype/category-S
unictype/category-Sc
unictype/category-Sk
unictype/category-Sm
unictype/category-So
unictype/category-Z
unictype/category-Zl
unictype/category-Zp
unictype/category-Zs
unictype/category-all
unictype/category-and
unictype/category-and-not
unictype/category-byname
unictype/category-longname
unictype/category-name
unictype/category-or
unictype/combining-class-all
unictype/combining-class-byname
unictype/combining-class-longname
unictype/combining-class-name
unictype/ctype-alnum
unictype/ctype-alpha
unictype/ctype-blank
unictype/ctype-cntrl
unictype/ctype-digit
unictype/ctype-graph
unictype/ctype-lower
unictype/ctype-print
unictype/ctype-punct
unictype/ctype-space
unictype/ctype-upper
unictype/ctype-xdigit
unictype/decimal-digit
unictype/digit
unictype/joininggroup-all
unictype/joininggroup-byname
unictype/joininggroup-name
unictype/joininggroup-of
unictype/joiningtype-all
unictype/joiningtype-byname
unictype/joiningtype-longname
unictype/joiningtype-name
unictype/mirror
unictype/numeric
unictype/property-all
unictype/property-alphabetic
unictype/property-ascii-hex-digit
unictype/property-bidi-arabic-digit
unictype/property-bidi-arabic-right-to-left
unictype/property-bidi-block-separator
unictype/property-bidi-boundary-neutral
unictype/property-bidi-common-separator
unictype/property-bidi-control
unictype/property-bidi-embedding-or-override
unictype/property-bidi-eur-num-separator
unictype/property-bidi-eur-num-terminator
unictype/property-bidi-european-digit
unictype/property-bidi-hebrew-right-to-left
unictype/property-bidi-left-to-right
unictype/property-bidi-non-spacing-mark
unictype/property-bidi-other-neutral
unictype/property-bidi-pdf
unictype/property-bidi-segment-separator
unictype/property-bidi-whitespace
unictype/property-byname
unictype/property-case-ignorable
unictype/property-cased
unictype/property-changes-when-casefolded
unictype/property-changes-when-casemapped
unictype/property-changes-when-lowercased
unictype/property-changes-when-titlecased
unictype/property-changes-when-uppercased
unictype/property-combining
unictype/property-composite
unictype/property-currency-symbol
unictype/property-dash
unictype/property-decimal-digit
unictype/property-default-ignorable-code-point
unictype/property-deprecated
unictype/property-diacritic
unictype/property-extender
unictype/property-format-control
unictype/property-grapheme-base
unictype/property-grapheme-extend
unictype/property-grapheme-link
unictype/property-hex-digit
unictype/property-hyphen
unictype/property-id-continue
unictype/property-id-start
unictype/property-ideographic
unictype/property-ids-binary-operator
unictype/property-ids-trinary-operator
unictype/property-ignorable-control
unictype/property-iso-control
unictype/property-join-control
unictype/property-left-of-pair
unictype/property-line-separator
unictype/property-logical-order-exception
unictype/property-lowercase
unictype/property-math
unictype/property-non-break
unictype/property-not-a-character
unictype/property-numeric
unictype/property-other-alphabetic
unictype/property-other-default-ignorable-code-point
unictype/property-other-grapheme-extend
unictype/property-other-id-continue
unictype/property-other-id-start
unictype/property-other-lowercase
unictype/property-other-math
unictype/property-other-uppercase
unictype/property-paired-punctuation
unictype/property-paragraph-separator
unictype/property-pattern-syntax
unictype/property-pattern-white-space
unictype/property-private-use
unictype/property-punctuation
unictype/property-quotation-mark
unictype/property-radical
unictype/property-sentence-terminal
unictype/property-soft-dotted
unictype/property-space
unictype/property-terminal-punctuation
unictype/property-test
unictype/property-titlecase
unictype/property-unassigned-code-value
unictype/property-unified-ideograph
unictype/property-uppercase
unictype/property-variation-selector
unictype/property-white-space
unictype/property-xid-continue
unictype/property-xid-start
unictype/property-zero-width
unictype/scripts-all
unictype/syntax-c-ident
unictype/syntax-c-whitespace
unictype/syntax-java-ident
unictype/syntax-java-whitespace
unigbrk/base
unigbrk/u16-grapheme-breaks
unigbrk/u16-grapheme-next
unigbrk/u16-grapheme-prev
unigbrk/u32-grapheme-breaks
unigbrk/u32-grapheme-next
unigbrk/u32-grapheme-prev
unigbrk/u8-grapheme-breaks
unigbrk/u8-grapheme-next
unigbrk/u8-grapheme-prev
unigbrk/uc-gbrk-prop
unigbrk/uc-is-grapheme-break
unigbrk/ulc-grapheme-breaks
unilbrk/base
unilbrk/tables
unilbrk/u16-possible-linebreaks
unilbrk/u16-width-linebreaks
unilbrk/u32-possible-linebreaks
unilbrk/u32-width-linebreaks
unilbrk/u8-possible-linebreaks
unilbrk/u8-width-linebreaks
unilbrk/ulc-common
unilbrk/ulc-possible-linebreaks
unilbrk/ulc-width-linebreaks
uniname/base
uniname/uniname
uninorm/compat-decomposition
uninorm/decomposing-form
uninorm/decomposition
uninorm/filter
uninorm/nfkc
uninorm/nfkd
uninorm/u16-normalize
uninorm/u16-normcmp
uninorm/u16-normcoll
uninorm/u16-normxfrm
uninorm/u32-normcmp
uninorm/u32-normcoll
uninorm/u32-normxfrm
uninorm/u8-normalize
uninorm/u8-normcmp
uninorm/u8-normcoll
uninorm/u8-normxfrm
unistdio/base
unistdio/u-printf-args
unistdio/u16-asnprintf
unistdio/u16-asprintf
unistdio/u16-printf-parse
unistdio/u16-snprintf
unistdio/u16-sprintf
unistdio/u16-u16-asnprintf
unistdio/u16-u16-asprintf
unistdio/u16-u16-snprintf
unistdio/u16-u16-sprintf
unistdio/u16-u16-vasnprintf
unistdio/u16-u16-vasprintf
unistdio/u16-u16-vsnprintf
unistdio/u16-u16-vsprintf
unistdio/u16-vasnprintf
unistdio/u16-vasprintf
unistdio/u16-vsnprintf
unistdio/u16-vsprintf
unistdio/u32-asnprintf
unistdio/u32-asprintf
unistdio/u32-printf-parse
unistdio/u32-snprintf
unistdio/u32-sprintf
unistdio/u32-u32-asnprintf
unistdio/u32-u32-asprintf
unistdio/u32-u32-snprintf
unistdio/u32-u32-sprintf
unistdio/u32-u32-vasnprintf
unistdio/u32-u32-vasprintf
unistdio/u32-u32-vsnprintf
unistdio/u32-u32-vsprintf
unistdio/u32-vasnprintf
unistdio/u32-vasprintf
unistdio/u32-vsnprintf
unistdio/u32-vsprintf
unistdio/u8-asnprintf
unistdio/u8-asprintf
unistdio/u8-printf-parse
unistdio/u8-snprintf
unistdio/u8-sprintf
unistdio/u8-u8-asnprintf
unistdio/u8-u8-asprintf
unistdio/u8-u8-snprintf
unistdio/u8-u8-sprintf
unistdio/u8-u8-vasnprintf
unistdio/u8-u8-vasprintf
unistdio/u8-u8-vsnprintf
unistdio/u8-u8-vsprintf
unistdio/u8-vasnprintf
unistdio/u8-vasprintf
unistdio/u8-vsnprintf
unistdio/u8-vsprintf
unistdio/ulc-asnprintf
unistdio/ulc-asprintf
unistdio/ulc-fprintf
unistdio/ulc-printf-parse
unistdio/ulc-snprintf
unistdio/ulc-sprintf
unistdio/ulc-vasnprintf
unistdio/ulc-vasprintf
unistdio/ulc-vfprintf
unistdio/ulc-vsnprintf
unistdio/ulc-vsprintf
unistr/u16-check
unistr/u16-chr
unistr/u16-cmp
unistr/u16-cmp2
unistr/u16-cpy
unistr/u16-cpy-alloc
unistr/u16-endswith
unistr/u16-mblen
unistr/u16-mbsnlen
unistr/u16-mbtouc
unistr/u16-mbtouc-unsafe
unistr/u16-mbtoucr
unistr/u16-move
unistr/u16-next
unistr/u16-prev
unistr/u16-set
unistr/u16-startswith
unistr/u16-stpcpy
unistr/u16-stpncpy
unistr/u16-strcat
unistr/u16-strchr
unistr/u16-strcmp
unistr/u16-strcoll
unistr/u16-strcpy
unistr/u16-strcspn
unistr/u16-strdup
unistr/u16-strlen
unistr/u16-strmblen
unistr/u16-strmbtouc
unistr/u16-strncat
unistr/u16-strncmp
unistr/u16-strncpy
unistr/u16-strnlen
unistr/u16-strpbrk
unistr/u16-strrchr
unistr/u16-strspn
unistr/u16-strstr
unistr/u16-strtok
unistr/u16-to-u32
unistr/u16-to-u8
unistr/u16-uctomb
unistr/u32-check
unistr/u32-chr
unistr/u32-cmp
unistr/u32-cmp2
unistr/u32-cpy-alloc
unistr/u32-endswith
unistr/u32-mblen
unistr/u32-mbsnlen
unistr/u32-mbtouc
unistr/u32-mbtoucr
unistr/u32-move
unistr/u32-next
unistr/u32-prev
unistr/u32-set
unistr/u32-startswith
unistr/u32-stpcpy
unistr/u32-stpncpy
unistr/u32-strcat
unistr/u32-strchr
unistr/u32-strcmp
unistr/u32-strcoll
unistr/u32-strcpy
unistr/u32-strcspn
unistr/u32-strdup
unistr/u32-strlen
unistr/u32-strmblen
unistr/u32-strmbtouc
unistr/u32-strncat
unistr/u32-strncmp
unistr/u32-strncpy
unistr/u32-strnlen
unistr/u32-strpbrk
unistr/u32-strrchr
unistr/u32-strspn
unistr/u32-strstr
unistr/u32-strtok
unistr/u32-to-u16
unistr/u8-chr
unistr/u8-cmp
unistr/u8-cmp2
unistr/u8-cpy
unistr/u8-cpy-alloc
unistr/u8-endswith
unistr/u8-mbsnlen
unistr/u8-move
unistr/u8-next
unistr/u8-set
unistr/u8-startswith
unistr/u8-stpcpy
unistr/u8-stpncpy
unistr/u8-strcat
unistr/u8-strchr
unistr/u8-strcmp
unistr/u8-strcoll
unistr/u8-strcpy
unistr/u8-strcspn
unistr/u8-strdup
unistr/u8-strmblen
unistr/u8-strmbtouc
unistr/u8-strncat
unistr/u8-strncmp
unistr/u8-strncpy
unistr/u8-strnlen
unistr/u8-strpbrk
unistr/u8-strrchr
unistr/u8-strspn
unistr/u8-strstr
unistr/u8-strtok
unistr/u8-to-u16
uniwbrk/base
uniwbrk/table
uniwbrk/u16-wordbreaks
uniwbrk/u32-wordbreaks
uniwbrk/u8-wordbreaks
uniwbrk/ulc-wordbreaks
uniwbrk/wordbreak-property
uniwidth/base
uniwidth/u16-strwidth
uniwidth/u16-width
uniwidth/u32-strwidth
uniwidth/u32-width
uniwidth/u8-strwidth
uniwidth/u8-width
uniwidth/width
utf16-ucs4
utf16-ucs4-unsafe
utf8-ucs4
utf8-ucs4-unsafe
va-args
vararrays
vasnprintf-posix
vasprintf-posix
vdprintf
vdprintf-posix
version-etc
version-etc-fsf
vfprintf-posix
vfscanf
vprintf-posix
vscanf
vsnprintf-posix
vsprintf-posix
wcpncpy
wcscasecmp
wcsdup
wcsncasecmp
wcsnlen
wcsnrtombs
wcsrtombs
wcsstr
wcstok
wcswidth
wcsxfrm
wctob
wctrans
wcwidth
wmemchr
wmemcmp
wmemcpy
wmemmove
wmemset
y0
y1
yn
'

gnulib_lgpl2_crypto='
crypto/arcfour
crypto/arctwo
crypto/des
crypto/gc
crypto/gc-arcfour
crypto/gc-arctwo
crypto/gc-camellia
crypto/gc-des
crypto/gc-hmac-md5
crypto/gc-hmac-sha1
crypto/gc-hmac-sha256
crypto/gc-hmac-sha512
crypto/gc-md2
crypto/gc-md4
crypto/gc-md5
crypto/gc-pbkdf2-sha1
crypto/gc-random
crypto/gc-rijndael
crypto/gc-sha1
crypto/hmac-md5
crypto/hmac-sha1
crypto/hmac-sha256
crypto/hmac-sha512
crypto/md2
crypto/md4
crypto/md5
crypto/rijndael
crypto/sha1
crypto/sha256
crypto/sha512
'

gnulib_lgpl2_files='
canonicalize-lgpl
filename
filenamecat-lgpl
fnmatch
fnmatch-gnu
glob
largefile
pathmax
read-file
tempname
'

gnulib_lgpl2_i18n='
gettext-h
hard-locale
iconv
iconv-h
iconv_open
langinfo
localcharset
localeconv
localename
mbrtowc
mbsinit
mbsrtowcs
mbtowc
nl_langinfo
striconv
striconveh
striconveha
'

gnulib_lgpl2_math='
bitrotate
count-leading-zeros
count-one-bits
count-trailing-zeros
ffs
ffsl
ffsll
float
integer_length
integer_length_l
integer_length_ll
intprops
inttostr
inttypes
inttypes-incomplete
ldexp
math
minmax
random_r
stdint
'

gnulib_lgpl2_memory='
alloca-opt
allocator
calloc-posix
malloca
malloc-gnu
malloc-posix
memchr
memchr2
memmem
memmem-simple
mempcpy
memxor
rawmemchr
realloc-posix
safe-alloc
'

gnulib_lgpl2_net='
byteswap
getaddrinfo
hostent
inet_ntop
inet_pton
net_if
servent
socketlib
sockets
'

gnulib_lgpl2_pty='
grantpt
openpty
posix_openpt
pt_chown
ptsname
ptsname_r
pty
unlockpt
'

gnulib_lgpl2_stdio='
fflush
fopen
fpurge
freadahead
freading
freadptr
freadseek
fseek
fseeko
ftell
ftello
getdelim
getline
getndelim2
snprintf
stdio
vasnprintf
vasprintf
vsnprintf
'

gnulib_lgpl2_signal='
exitfail
raise
sigaction
signal
sigpipe
strsignal
'

gnulib_lgpl2_spawn='
posix_spawn
posix_spawnattr_destroy
posix_spawnattr_getflags
posix_spawnattr_getpgroup
posix_spawnattr_getschedparam
posix_spawnattr_getschedpolicy
posix_spawnattr_getsigdefault
posix_spawnattr_getsigmask
posix_spawnattr_init
posix_spawnattr_setflags
posix_spawnattr_setpgroup
posix_spawnattr_setschedparam
posix_spawnattr_setschedpolicy
posix_spawnattr_setsigdefault
posix_spawnattr_setsigmask
posix_spawn_file_actions_addclose
posix_spawn_file_actions_adddup2
posix_spawn_file_actions_addopen
posix_spawn_file_actions_destroy
posix_spawn_file_actions_init
posix_spawn-internal
posix_spawnp
spawn
'

gnulib_lgpl2_string='
base32
base64
btowc
c-ctype
c-strcase
c-strcaseeq
c-strcasestr
isblank
iswblank
regex
strcase
strcasestr
strcasestr-simple
strchrnul
strdup-posix
streq
strerror
strerror-override
strerror_r-posix
string
strings
strncat
strndup
strnlen
strnlen1
strptime
strsep
strstr
strstr-simple
strtok_r
strverscmp
wchar
wcrtomb
wctomb
wctype-h
'

gnulib_lgpl2_sys='
areadlink
areadlink-with-size
careadlinkat
chdir
chown
cloexec
dirfd
dirname-lgpl
dosname
double-slash-root
d-type
execinfo
fdatasync
fd-hook
full-read
full-write
getcwd-lgpl
getgroups
getlogin
getlogin_r
getopt-posix
getpagesize
getpass
getugroups
lchown
link
lstat
mkdir
mkdtemp
mkostemp
mkostemps
mkstemp
mkstemps
mgetgroups
nonblocking
nproc
open
perror
physmem
pipe2
pipe-posix
poll-h
pread
pselect
pwrite
readlink
rename
rmdir
root-uid
safe-read
safe-write
same-inode
sched
sethostname
stat
stat-time
sysexits
sys_file
sys_ioctl
sys_select
sys_uio
termios
ttyname_r
uname
unlink
waitpid
xsize
'

gnulib_lgpl2_thread='
cond
lock
pthread
pthread_sigmask
thread
threadlib
tls
yield
'

gnulib_lgpl2_time='
clock-time
gettimeofday
mktime
mktime-internal
parse-duration
sleep
sys_times
time
timegm
time_r
timer-time
times
usleep
'

gnulib_lgpl2_unicode='
uniconv/base
uniconv/u8-conv-from-enc
uniconv/u8-strconv-from-enc
uniconv/u8-strconv-from-locale
unictype/base
unictype/bidiclass-of
unictype/category-M
unictype/category-none
unictype/category-of
unictype/category-test
unictype/category-test-withtable
unictype/combining-class
unictype/joiningtype-of
unictype/scripts
uninorm/base
uninorm/canonical-decomposition
uninorm/composition
uninorm/decompose-internal
uninorm/decomposition-table
uninorm/nfc
uninorm/nfd
uninorm/u32-normalize
unistr/base
unistr/u32-cpy
unistr/u32-mbtouc-unsafe
unistr/u32-to-u8
unistr/u32-uctomb
unistr/u8-check
unistr/u8-mblen
unistr/u8-mbtouc
unistr/u8-mbtoucr
unistr/u8-mbtouc-unsafe
unistr/u8-prev
unistr/u8-strlen
unistr/u8-to-u32
unistr/u8-uctomb
unitypes
'

gnulib_lgpl2_unknown='
alignof
ansi-c++-opt
argz
array-mergesort
assure
binary-io
check-version
concat-filename
extensions
findprog-lgpl
hash
hash-pjw
hash-pjw-bare
ignore-value
include_next
passfd
qsort_r
search
selinux-h
size_max
snippet/arg-nonnull
snippet/c++defs
snippet/link-warning
snippet/_Noreturn
snippet/unused-parameter
snippet/warn-on-use
stdalign
stdarg
stdbool
stddef
stdnoreturn
u64
verify
'

gnulib_lgpl2='
sys_types
stdlib
unistd
environ
sys_stat
sys_time
errno
secure_getenv
fcntl-h
signal-h
dirent
ctype
sys_utsname
sys_wait
getopt-gnu
stpcpy
stpncpy
realloc-gnu
setenv
unsetenv
netdb
netinet_in
arpa_inet
locale
sys_socket
assert-h

accept
bind
close
connect
dup2
fclose
fcntl
flock
fsync
gethostname
getpeername
getsockname
getsockopt
ioctl
listen
lseek
poll
recv
recvfrom
select
send
sendto
setsockopt
shutdown
socket
sigprocmask
write

getdtablesize
dup
fdopen
fstat
isatty
read

'

gnulib_modules="
absolute-header
configmake
lib-msvc-compat
lib-symbol-versions
msvc-inval
msvc-nothrow
relocatable-lib-lgpl
${gnulib_lgpl2}
"

gnulib_tool=.gnulib/gnulib-tool

for dir in .gnulib bash beecrypt brotli file flatcc gpsee libtpm libgit2 lz4 lzfse neon pcre popt rc snappy syck xar xz zstd; do

  [ -d $dir ] || continue
  echo "===> $dir"
  case $dir in
  .gnulib)
    $gnulib_tool                    \
      --lgpl=2                      \
      --m4-base=m4                  \
      --aux-dir=build-aux           \
      --source-base=gnulib          \
      --with-tests                  \
      --tests-base=gnulib/tests     \
      --local-dir=misc              \
      --libtool                     \
      --quiet                       \
      --avoid closedir              \
      --avoid opendir               \
      --avoid readdir               \
      --avoid fclose                \
      --avoid fflush                \
      --import $gnulib_modules
    ;;
  flatcc)
    ( cd $dir && cmake . && 
	echo "check: test" >> Makefile &&
	echo "distclean: clean" >> Makefile
	#cp *.la lib/
	#mkdir -p lib/.libs
	#ln -sf ../libflatccrt.a lib/.libs/libflatccrt.a
    )
    ;;
  zstd)
    ;;
  lz4)
    ;;
  snappy)
    ( cd $dir && sh ./autogen.sh )
    ;;
  brotli)
    ( cd $dir && ./configure-cmake --prefix=/usr --libdir=/usr/lib64 &&
	echo "check: test" >> ../Makefile &&
	echo "distclean: clean" >> ../Makefile
    )
    ;;
  lzfse)
    ( cd $dir
        cmake . \
            -DLZFSE_BUNDLE_MODE:BOOL=OFF \
            -DBUILD_SHARED_LIBS:BOOL=ON \
            -DCMAKE_INSTALL_BINDIR:PATH=/usr/bin \
            -DCMAKE_INSTALL_LIBDIR:PATH=/usr/lib64 \
            -DCMAKE_INSTALL_INCLUDEDIR:PATH=/usr/include
	echo "check: test" >> Makefile
	echo "distclean: clean"
    )
    ;;
  libgit2)
    ( cd $dir && cmake . && 
	echo "check: test" >> Makefile &&
	echo "distclean: clean" >> Makefile
    )
    ;;
  *)
    ( cd $dir && sh ./autogen.sh --noconfigure "$@" )
    ;;
  esac
  echo "<=== $dir"
done

echo "===> rpm"
rm -rf autom4te.cache || true
echo "---> generate files via GNU libtool (libtoolize)"
libtoolize --quiet --copy --force --install
echo "---> generate files via GNU gettext (autopoint)"
po_dir=./po
LANG=C
ls "$po_dir"/*.po 2>/dev/null |
	sed 's|.*/||; s|\.po$||' > "$po_dir/LINGUAS"
autopoint --force
echo "---> generate files via GNU autoconf (aclocal, autoheader)"
rm -f aclocal.m4
aclocal -I m4
autoheader -I m4
echo "---> generate files via GNU automake (automake)"
automake -Wall -Wno-override -a -c
echo "---> generate files via GNU autoconf (autoconf)"
autoconf -I m4
echo "<=== rpm"

