/** \ingroup js_c
 * \file js/rpmbf-js.c
 */

#include "system.h"

#include "rpmbf-js.h"
#include "rpmjs-debug.h"

#include <rpmmacro.h>
#include <rpmbf.h>

#include "debug.h"

/*@unchecked@*/
static int _debug = -1;
#if !defined(SPEW)
# define SPEW(_fmt, ...) \
    if (_debug || _rpmbf_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)
#endif
  
static bool
rpmbf_add(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);
    RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmbfClass, nullptr);
    rpmbf bf = (rpmbf) ptr;
    bool ok = true;

//_METHOD_DEBUG_ENTRY(_debug);

    RootedString str(cx, ToString(cx, args[0]));
    const char * _s = JS_EncodeStringToUTF8(cx, str);

    args.rval().setBoolean(rpmbfAdd(bf, _s, 0) == 0);
    return ok;
}

static bool
rpmbf_chk(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);
    RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmbfClass, nullptr);
    rpmbf bf = (rpmbf) ptr;
    bool ok = true;

//_METHOD_DEBUG_ENTRY(_debug);

    RootedString str(cx, ToString(cx, args[0]));
    const char * _s = JS_EncodeStringToUTF8(cx, str);

    args.rval().setBoolean(rpmbfChk(bf, _s, 0) > 0);
    return ok;
}

static bool
rpmbf_clr(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);
    RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmbfClass, nullptr);
    rpmbf bf = (rpmbf) ptr;
    bool ok = true;

//_METHOD_DEBUG_ENTRY(_debug);

    args.rval().setBoolean(rpmbfClr(bf) == 0);
    return ok;
}

static bool
rpmbf_del(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);
    RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmbfClass, nullptr);
    rpmbf bf = (rpmbf) ptr;
    bool ok = true;

//_METHOD_DEBUG_ENTRY(_debug);

    RootedString str(cx, ToString(cx, args[0]));
    const char * _s = JS_EncodeStringToUTF8(cx, str);

    args.rval().setBoolean(rpmbfDel(bf, _s, 0) == 0);
    return ok;
}

static bool
rpmbf_intersect(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);
    RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmbfClass, nullptr);
    rpmbf _a = (rpmbf) ptr;
    bool ok = true;

//_METHOD_DEBUG_ENTRY(_debug);

    RootedObject o(cx, ToObject(cx, args[0]));
    rpmbf _b = (rpmbf) JS_GetInstancePrivate(cx, o, &rpmbfClass, nullptr);

    args.rval().setBoolean(rpmbfIntersect(_a, _b) == 0);
    return ok;
}

static bool
rpmbf_union(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);
    RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmbfClass, nullptr);
    rpmbf _a = (rpmbf) ptr;
    bool ok = true;

//_METHOD_DEBUG_ENTRY(_debug);

    RootedObject o(cx, ToObject(cx, args[0]));
    rpmbf _b = (rpmbf) JS_GetInstancePrivate(cx, o, &rpmbfClass, nullptr);

    args.rval().setBoolean(rpmbfUnion(_a, _b) == 0);
    return ok;
}

static JSFunctionSpec rpmbf_funcs[] = {
    JS_FS("add",	rpmbf_add,		0,0),
    JS_FS("chk",	rpmbf_chk,		0,0),
    JS_FS("clr",	rpmbf_clr,		0,0),
    JS_FS("del",	rpmbf_del,		0,0),
    JS_FS("intersect",	rpmbf_intersect,	0,0),
    JS_FS("union",	rpmbf_union,		0,0),
    JS_FS_END
};

/* --- Object properties */
static bool
rpmbf_getprop(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);

    args.rval().setInt32(_debug);
    return true;
}

static bool
rpmbf_setprop(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);

    args.rval().setInt32(_debug);
    if (!ToInt32(cx, args[0], &_debug))
	return false;
    return true;
}

static JSPropertySpec rpmbf_props[] = {
    JS_PSGS("debug", rpmbf_getprop, rpmbf_setprop, JSPROP_ENUMERATE),
    JS_PS_END
};

#ifdef	NOTYET
static bool
rpmbf_addProperty(JSContext *cx,
		Handle<JSObject*> obj,
		Handle<jsid> id,
		Handle<Value> vp)
{
    return true;
}
#else
#define	rpmbf_addProperty	nullptr
#endif

#ifdef	NOTYET
static bool
rpmbf_delProperty(JSContext *cx,
		Handle<JSObject*> obj,
		Handle<jsid> id,
		ObjectOpResult& result)
{
    return true;
}
#else
#define	rpmbf_delProperty	nullptr
#endif

#ifdef	NOTYET
static bool
rpmbf_getProperty(JSContext *cx,
		Handle<JSObject*> obj,
		Handle<jsid> id,
		MutableHandle<Value> vp)
{
    return true;
}
#else
#define	rpmbf_getProperty	nullptr
#endif

#ifdef	NOTYET
static bool
rpmbf_setProperty(JSContext *cx,
		Handle<JSObject*> obj,
		Handle<jsid> id,
		MutableHandle<Value> vp,
		ObjectOpResult& result)
{
    return true;
}
#else
#define	rpmbf_setProperty	nullptr
#endif

#ifdef	NOTYET
static bool
rpmbf_resolve(JSContext *cx,
		Handle<JSObject*> obj,
		Handle<jsid> id,
		bool *result)
{
// _RESOLVE_DEBUG_ENTRY(_debug);

#ifdef	DYING
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmbfClass, nullptr);
    rpmbf bf = (rpmbf) ptr;

    if (ptr == NULL) {	/* don't resolve to parent prototypes objects. */
	*objp = NULL;
	goto exit;
    }

#ifdef	NOTYET
    /* Lazily resolve new strings, with duplication to Augeas defvar too. */
    if ((flags & JSRESOLVE_ASSIGNING) && JSVAL_IS_STRING(id)) {
        const char *name = JS_EncodeString(cx, JS_ValueToString(cx, id));
	const char * _path;
	const char * _value;
	int xx;
	JSFunctionSpec *fsp;

	/* XXX avoid "bf.print" method namess duped into defvar space? */
	for (fsp = rpmbf_funcs; fsp->name != NULL; fsp++) {
	    if (!strcmp(fsp->name, name)) {
		*objp = obj;	/* XXX always resolve in this object. */
		goto exit;
	    }
	}

	/* See if NAME exists in DEFVAR path. */
	_path = rpmGetPath(_defvar, "/", name, NULL);
	_value = NULL;

	switch (rpmbfGet(bf, _path, &_value)) {
	/* XXX For now, force all defvar's to be single valued strings. */
	case -1:/* multiply defined */
	case 0:	/* not found */
	    /* Create an empty Augeas defvar for NAME */
	    xx = rpmbfDefvar(bf, name, "");
	    /*@fallthrough@*/
	case 1:	/* found */
	    /* Reflect the Augeas defvar NAME into JS Aug properties. */
	    if (JS_DefineProperty(cx, obj, name,
			(_value != NULL
			    ? STRING_TO_JSVAL(JS_NewStringCopyZ(cx, _value))
			    : JSVAL_VOID),
                        NULL, NULL, JSPROP_ENUMERATE))
		break;
	    /*@fallthrough@*/
	default:
assert(0);
	    break;
	}

	_value = _free(_value);
	_path = _free(_path);
	name = _free(_name);
    }
#endif
    *objp = obj;	/* XXX always resolve in this object. */

exit:
#endif	/* DYING */

    return true;
}
#else
#define	rpmbf_resolve		nullptr
#endif

#ifdef	NOTYET
static bool
rpmbf_mayResolve(const JSAtomState& state, jsid id, JSObject *obj)
{
    return true;
}
#else
#define	rpmbf_mayResolve	nullptr
#endif

#ifdef	NOTYET
static bool
rpmbf_enumerate(JSContext *cx, JS::Handle<JSObject*> obj)
{
SPEW("==> %s(%p,%p)\n", __FUNCTION__, cx, obj);
// _ENUMERATE_DEBUG_ENTRY(_debug);

#ifdef	DYING
    switch (op) {
    case JSENUMERATE_INIT:
    case JSENUMERATE_INIT_ALL:
	*statep = JSVAL_VOID;
	if (idp)
	    *idp = JSVAL_ZERO;
        break;
    case JSENUMERATE_NEXT:
	*statep = JSVAL_VOID;
	if (*idp != (jsid)JSVAL_VOID)
	    break;
        /*@fallthrough@*/
    case JSENUMERATE_DESTROY:
	*statep = JSVAL_NULL;
        break;
    }
#endif	/* DYING */

    return true;
}
#else
#define	rpmbf_enumerate		nullptr
#endif

#ifdef	NOTYET
static bool
rpmbf_call(JSContext *cx, unsigned argc, Value* vp)
{
SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
    CallArgs args = CallArgsFromVp(argc, vp);
    RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
    void * ptr = JS_GetInstancePrivate(cx, obj, &rpmbfClass, nullptr);
    rpmbf _a = (rpmbf) ptr;

    return true;
}
#else
#define	rpmbf_call		nullptr
#endif

#ifdef	NOTYET
static bool
rpmbf_hasInstance(JSContext *cx,
		Handle<JSObject*> obj,
		MutableHandle<Value>,
		bool *result)
{
    return true;
}
#else
#define	rpmbf_hasInstance	nullptr
#endif

#ifdef	NOTYET
static void
rpmbf_trace(JSTracer *foo, JSObject *obj)
{
}
#else
#define	rpmbf_trace		nullptr
#endif

/* --- Object ctors/dtors */
static rpmbf
rpmbf_init(JSObject *obj, size_t _m, size_t _k, unsigned int _flags)
{
//SPEW("==> %s(%p, %zu, %zu, %u)\n", __FUNCTION__, obj, _m, _k, _flags);
_rpmbf_debug = _debug;
    rpmbf bf = rpmbfNew(_m, _k, _flags);
    if (bf)
	JS_SetPrivate(obj, bf);
    return bf;
}

#ifndef	NOTYET
static void
rpmbf_dtor(JSFreeOp* fop, JSObject *obj)
{
//SPEW("==> %s(%p, %p)\n", __FUNCTION__, fop, obj);
    void * ptr = JS_GetPrivate(obj);
    rpmbf bf = (rpmbf) ptr;
_DTOR_DEBUG_ENTRY(_debug);
    bf = rpmbfFree(bf);
}
#else
#define	rpmbf_dtor		nullptr
#endif

static bool
rpmbf_ctor(JSContext *cx, unsigned argc, Value* vp)
{
//SPEW("==> %s(%p,%u,%p)\n", __FUNCTION__, cx, argc, vp);
_CTOR_DEBUG_ENTRY(_debug);
    CallArgs args = CallArgsFromVp(argc, vp);

    RootedObject obj(cx, JS_NewObjectForConstructor(cx, &rpmbfClass, args));
    if (!obj
     || !JS_FreezeObject(cx, obj))
	return false;

    unsigned int _m = 0;
    unsigned int _k = 0;
    unsigned int _flags = 0;

    switch (argc) {
    default:
	return false;
    case 3:
	if (!ToUint32(cx, args[2], &_flags))
	    return false;
    case 2:
	if (!ToUint32(cx, args[1], &_k))
	    return false;
    case 1:
	if (!ToUint32(cx, args[0], &_m))
	    return false;
    case 0:
	break;
    }

    if (rpmbf_init(obj, _m, _k, _flags) == NULL)
	return false;

    args.rval().setObject(*obj);
    return true;
}

/* --- Class initialization */
JSClass rpmbfClass = {
    /* XXX class should be "Bloom" eventually, avoid name conflicts for now */
    "Bf",
    JSCLASS_HAS_PRIVATE,
    rpmbf_addProperty,	// addProperty
    rpmbf_delProperty,	// delProperty
    rpmbf_getProperty,	// getProperty
    rpmbf_setProperty,	// setProperty
    rpmbf_enumerate,	// enumerate
    rpmbf_resolve,	// resolve
    rpmbf_mayResolve,	// mayResolve,
    rpmbf_dtor,		// finalize
    rpmbf_call,		// call
    rpmbf_hasInstance,	// hasInstance
    rpmbf_ctor,		// construct
    rpmbf_trace,	// trace
};

JSObject *
rpmjs_InitBfClass(JSContext *cx, HandleObject obj)
{
SPEW("==> %s(%p,%p)\n", __FUNCTION__, cx, obj.get());
    RootedObject o(cx);
    o = JS_InitClass(cx, obj, NULL,
		&rpmbfClass, rpmbf_ctor, 0, rpmbf_props, rpmbf_funcs,
		NULL, NULL);
    return o;
}

JSObject *
rpmjs_NewBfObject(JSContext *cx, size_t _m, size_t _k, unsigned int _flags)
{
SPEW("==> %s(%p,%u,%u,%u)\n", __FUNCTION__, cx, _m, _k, _flags);
    JSObject *obj = JS_NewObject(cx, &rpmbfClass);
    if (!obj)
	return NULL;

    rpmbf bf = rpmbf_init(obj, _m, _k, _flags);
    if (!bf)
	return NULL;

    return obj;
}
