#ifndef H_FSM
#define H_FSM

/** \ingroup payload
 * \file lib/fsm.h
 * File state machine to handle a payload within an rpm package.
 */

#include <iosm.h>

/** \ingroup payload
 * File state machine data.
 */
typedef struct iosm_s * FSM_t;

extern int _fsm_debug;

/** \ingroup payload
 * Iterator across package file info, forward on install, backward on erase.
 */
typedef struct iosmIterator_s * FSMI_t;

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Create file state machine instance.
 * @return		file state machine
 */
IOSM_t newFSM(void);

/**
 * Destroy file state machine instance.
 * @param fsm		file state machine
 * @return		always NULL
 */
IOSM_t freeFSM(IOSM_t fsm);

/**
 * Load external data into file state machine.
 * @param _fsm		file state machine
 * @param goal		goal
 * @param afmt		archive format (NULL uses cpio)
 * @param _ts		transaction set
 * @param _fi		transaction element file info
 * @param cfd		payload descriptor
 * @retval archiveSize	pointer to archive size
 * @retval failedFile	pointer to first file name that failed.
 * @return		0 on success
 */
int fsmSetup(void * _fsm, iosmFileStage goal, const char * afmt,
		const void * _ts,
		const void * _fi,
		FD_t cfd,
		unsigned int * archiveSize,
		const char ** failedFile);

/**
 * Clean file state machine.
 * @param _fsm		file state machine
 * @return		0 on success
 */
int fsmTeardown(void * _fsm);

/**
 * Map next file path and action.
 * @param fsm		file state machine
 */
int fsmMapPath(IOSM_t fsm);

/**
 * Map file stat(2) info.
 * @param fsm		file state machine
 */
int fsmMapAttrs(IOSM_t fsm);

/**
 * File state machine driver.
 * @param fsm		file state machine
 * @param nstage	next stage
 * @return		0 on success
 */
int fsmNext(IOSM_t fsm, iosmFileStage nstage);

/**
 * File state machine driver.
 * @param fsm		file state machine
 * @param stage		next stage
 * @return		0 on success
 */
int fsmStage(IOSM_t fsm, iosmFileStage stage);

#ifdef __cplusplus
}
#endif

#endif	/* H_FSM */
