#ifndef H_RPMAL
#define H_RPMAL

/** \ingroup rpmdep rpmtrans
 * \file lib/rpmal.h
 * Structures used for managing added/available package lists.
 */

#include <rpmds.h>
#include <rpmfi.h>

extern int _rpmal_debug;

#define	RPMAL_NOMATCH	((alKey)-1L)

/** \ingroup rpmts
 * An added/available package retrieval index.
 */
typedef int alNum;

/**
 */
typedef struct rpmal_s *		rpmal;

#ifdef __cplusplus
extern "C" {
#endif

/** \ingroup rpmal
 * Unreference available list.
 * @param al		available list
 * @param msg
 * @return		NULL on last dereference
 */
rpmal rpmalUnlink (rpmal al, const char * msg);
#define	rpmalUnlink(_al, _msg)	\
    ((rpmal)rpmioUnlinkPoolItem((rpmioItem)(_al), _msg, __FILE__, __LINE__))

/** \ingroup rpmal
 * Reference available list.
 * @param al		available list
 * @param msg
 * @return		new available list reference
 */
rpmal rpmalLink (rpmal al, const char * msg);
#define	rpmalLink(_al, _msg)	\
    ((rpmal)rpmioLinkPoolItem((rpmioItem)(_al), _msg, __FILE__, __LINE__))

/** \ingroup rpmal
 * Destroy available list.
 * @param al		available list
 * @return		NULL on last dereference
 */
rpmal rpmalFree(rpmal al);
#define	rpmalFree(_al)	\
    ((rpmal)rpmioFreePoolItem((rpmioItem)(_al), __FUNCTION__, __FILE__, __LINE__))

/**
 * Initialize available list.
 * @param delta		no. of entries to add on each realloc
 * @return al		new available list
 */
rpmal rpmalNew(int delta);

/**
 * Delete package from available list.
 * @param al		available list
 * @param pkgKey	package key
 */
void rpmalDel(rpmal al, alKey pkgKey);

/**
 * Add package to available list.
 * @param alistp	address of available list
 * @param pkgKey	package key, RPMAL_NOMATCH to force an append
 * @param key		associated file name/python object
 * @param provides	provides dependency set
 * @param fi		file info set
 * @param tscolor	transaction color bits
 * @return		available package index
 */
alKey rpmalAdd(rpmal * alistp, alKey pkgKey, fnpyKey key,
		rpmds provides, rpmfi fi, rpmuint32_t tscolor);

/**
 * Add package provides to available list index.
 * @param al		available list
 * @param pkgKey	package key
 * @param provides	added package provides
 * @param tscolor	transaction color bits
 */
void rpmalAddProvides(rpmal al, alKey pkgKey,
		rpmds provides, rpmuint32_t tscolor);

/**
 * Generate index for available list.
 * @param al		available list
 */
void rpmalMakeIndex(rpmal al);

/**
 * Check added package file lists for package(s) that provide a file.
 * @param al		available list
 * @param ds		dependency set
 * @retval keyp		added package key pointer (or NULL)
 * @return		associated package key(s), NULL if none
 */
fnpyKey * rpmalAllFileSatisfiesDepend(const rpmal al,
		const rpmds ds, alKey * keyp);

/**
 * Check added package file lists for package(s) that have a provide.
 * @param al		available list
 * @param ds		dependency set
 * @retval keyp		added package key pointer (or NULL)
 * @return		associated package key(s), NULL if none
 */
fnpyKey * rpmalAllSatisfiesDepend(const rpmal al, const rpmds ds, alKey * keyp);

/**
 * Check added package file lists for first package that has a provide.
 * @todo Eliminate.
 * @param al		available list
 * @param ds		dependency set
 * @retval keyp		added package key pointer (or NULL)
 * @return		associated package key, NULL if none
 */
fnpyKey rpmalSatisfiesDepend(const rpmal al, const rpmds ds, alKey * keyp);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMAL */
