#ifndef H_RPMROLLBACK
#define H_RPMROLLBACK

/** \ingroup rpmts
 * \file lib/rpmrollback.h
 *
 */

#include <rpmcli.h>	/* XXX QVA_t typedef */

/**
 */
typedef struct IDT_s * IDT;

/**
 */
typedef struct IDTindex_s * IDTX;

#if defined(_RPMROLLBACK_INTERNAL)
/**
 * A rollback transaction id element.
 */
#if !defined(SWIG)
struct IDT_s {
    int done;			/*!< package processed? */
    unsigned int instance;	/*!< installed package transaction id. */
    const char * key;		/*! removed package file name. */
    Header h;			/*!< removed package header. */
    union {
	rpmuint32_t u32;	/*!< install/remove transaction id */
    } val;
};
#endif

/**
 * A rollback transaction id index.
 */
#if !defined(SWIG)
struct IDTindex_s {
    int delta;			/*!< no. elements to realloc as a chunk. */
    int size;			/*!< size of id index element. */
    int alloced;		/*!< current number of elements allocated. */
    int nidt;			/*!< current number of elements initialized. */
    IDT idt;			/*!< id index elements. */
};
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Destroy id index.
 * @param idtx		id index
 * @return		NULL always
 */
IDTX IDTXfree(IDTX idtx);

/**
 * Create id index.
 * @return		new id index
 */
IDTX IDTXnew(void);

/**
 * Insure that index has room for "need" elements.
 * @param idtx		id index
 * @param need		additional no. of elements needed
 * @return 		id index (with room for "need" elements)
 */
IDTX IDTXgrow(IDTX idtx, int need);

/**
 * Sort tag (instance,value) pairs.
 * @param idtx		id index
 * @return 		id index
 */
IDTX IDTXsort(IDTX idtx);

/**
 * Load tag (instance,value) pairs from rpm databse, and return sorted id index.
 * @param ts		transaction set
 * @param tag		rpm tag
 * @param rbtid		rollback goal
 * @return 		id index
 */
IDTX IDTXload(rpmts ts, rpmTag tag, rpmuint32_t rbtid);

/**
 * Load tag (instance,value) pairs from packages, and return sorted id index.
 * @param ts		transaction set
 * @param globstr	glob expression
 * @param tag		rpm tag
 * @param rbtid		rollback goal
 * @return 		id index
 */
IDTX IDTXglob(rpmts ts, const char * globstr, rpmTag tag, rpmuint32_t rbtid);

/** \ingroup rpmcli
 * Rollback transactions, erasing new, reinstalling old, package(s).
 * @param ts		transaction set
 * @param ia		mode flags and parameters
 * @param argv		array of arguments (NULL terminated)
 * @return		0 on success
 */
int rpmRollback(rpmts ts, QVA_t ia, const char ** argv);

#ifdef __cplusplus
}
#endif

#endif  /* H_RPMROLLBACK */
