#ifndef H_RPMWF
#define H_RPMWF

/**
 * \file rpmdb/rpmwf.h
 * Structure(s)and methods for a archive wrapper format (e.g. XAR).
 */

#include <rpmxar.h>

extern int _rpmwf_debug;

/**
 */
typedef struct rpmwf_s * rpmwf;

#ifdef	_RPMWF_INTERNAL
struct rpmwf_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * fn;
    FD_t fd;
    void * b;
    size_t nb;
    char * l;
    size_t nl;
    char * s;
    size_t ns;
    char * h;
    size_t nh;
    char * p;
    size_t np;
    rpmxar xar;
};
#endif


#ifdef __cplusplus
extern "C" {
#endif

rpmRC rpmwfPushXAR(rpmwf wf, const char * fn);

rpmRC rpmwfPullXAR(rpmwf wf, const char * fn);

rpmRC rpmwfFini(rpmwf wf);

rpmRC rpmwfInit(rpmwf wf, const char * fn, const char * fmode);

rpmRC rpmwfPushRPM(rpmwf wf, const char * fn);

/**
 * Unreference a wrapper format instance.
 * @param wf		wrapper format
 * @param msg
 * @return		NULL on last dereference
 */
rpmwf rpmwfUnlink (rpmwf wf, const char * msg);
#define	rpmwfUnlink(_wf, _msg)	\
    ((rpmwf)rpmioUnlinkPoolItem((rpmioItem)(_wf), _msg, __FILE__, __LINE__))

/**
 * Reference a wrapper format instance.
 * @param wf		wrapper format
 * @param msg
 * @return		new wrapper format reference
 */
rpmwf rpmwfLink (rpmwf wf, const char * msg);
#define	rpmwfLink(_wf, _msg)	\
    ((rpmwf)rpmioLinkPoolItem((rpmioItem)(_wf), _msg, __FILE__, __LINE__))

rpmwf rpmwfFree(rpmwf wf);
#define	rpmwfFree(_wf)	\
    ((rpmwf)rpmioFreePoolItem((rpmioItem)(_wf), __FUNCTION__, __FILE__, __LINE__))

rpmwf rpmwfNew(const char * fn);

rpmwf rdRPM(const char * rpmfn);

rpmwf rdXAR(const char * xarfn);

rpmRC wrXAR(const char * xarfn, rpmwf wf);

rpmRC wrRPM(const char * rpmfn, rpmwf wf);

#ifdef __cplusplus
}
#endif

#endif  /* H_RPMWF */
