/** \ingroup signature
 * \file rpmio/digest.c
 */

#include "system.h"

#include "rpmio_internal.h"

#include <rpmbc.h>

#include "crc.h"

#include "arirang.h"

#include "blake.h"
#include "blake2-rpm.h"

#include "bmw.h"

#include "chi.h"

#include "cubehash.h"

#include "echo.h"
#undef	BitSequence
#undef	DataLength
#undef	HashReturn
#undef	hashState
#undef	Init
#undef	Update
#undef	Final
#undef	Hash

#include "edon-r.h"

#include "fugue.h"

#include "groestl.h"
#undef	BitSequence
#undef	DataLength
#undef	HashReturn
#undef	hashState
#undef	Init
#undef	Update
#undef	Final
#undef	Hash

#include "hamsi.h"

#include "jh.h"

#include "keccak.h"
#undef	BitSequence
#undef	DataLength
#undef	HashReturn
#undef	hashState
#undef	Init
#undef	Update
#undef	Final
#undef	Hash

#include "lane.h"

#include "luffa.h"

#include "md2.h"
#include "md6.h"

#include "radiogatun.h"

#include "shabal.h"

#include "shavite3.h"
#undef	BitSequence
#undef	DataLength
#undef	HashReturn
#undef	hashState
#undef	Init
#undef	Update
#undef	Final
#undef	Hash

#include "simd.h"
#undef	BitSequence
#undef	DataLength
#undef	HashReturn
#undef	hashState
#undef	Init
#undef	Update
#undef	Final
#undef	Hash

#include "salsa10.h"
#include "salsa20.h"

#include "skein.h"

#include "tib3.h"
#undef	BitSequence
#undef	DataLength
#undef	HashReturn
#undef	hashState
#undef	Init
#undef	Update
#undef	Final
#undef	Hash

#include "tiger.h"

#include "debug.h"

int _rpmctx_debug = 0;

#ifdef	_DIGEST_DEBUG
#define	DPRINTF(_a)	if (_rpmctx_debug < 0) fprintf _a
#else
#define	DPRINTF(_a)
#endif

/* Include Bob Jenkins lookup3 hash */
#define	_JLU3_jlu32l
#include "lookup3.c"

/* Include Yann Collet XXH hash */
#define XXH_NAMESPACE   RPM_
#if 1
#define XXH_STATIC_LINKING_ONLY
#define XXH_PRIVATE_API
#include "xxhash.h"
#else
#include "xxhash.c"
#endif

typedef	XXH32_state_t xxh32Param;
static int xxh32Reset(void * _param)
{
    unsigned int seed = 0;
    return XXH32_reset(_param, seed);
}

static int xxh32Update(void * _param, const byte * data, size_t size)
{
    return XXH32_update(_param, (const void *) data, size);
}

static int xxh32Digest(void * _param, byte * digest)
{
    XXH32_canonicalFromHash((XXH32_canonical_t *) digest, XXH32_digest(_param));
    return 0;
}

typedef	XXH64_state_t xxh64Param;
static int xxh64Reset(void * _param)
{
    unsigned int seed = 0;
    return XXH64_reset(_param, seed);
}

static int xxh64Update(void * _param, const byte * data, size_t size)
{
    return XXH64_update(_param, (const void *) data, size);
}

static int xxh64Digest(void * _param, byte * digest)
{
    XXH64_canonicalFromHash((XXH64_canonical_t *) digest, XXH64_digest(_param));
    return 0;
}

/**
 * Digest private data.
 */
struct DIGEST_CTX_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * name;		/*!< Digest name. */
    size_t paramsize;		/*!< No. bytes of digest parameters. */
    size_t blocksize;		/*!< No. bytes in block of plaintext data. */
    size_t digestsize;		/*!< No. bytes of digest. */
    int (*Reset) (void * param);
				/*!< Digest initialize. */
    int (*Update) (void * param, const byte * data, size_t size);
				/*!< Digest update. */
    int (*Digest) (void * param, byte * digest);
				/*!< Digest finish. */
    pgpHashAlgo hashalgo;	/*!< RFC 2440/4880 hash algorithm id. */
    rpmDigestFlags flags;	/*!< Bit(s) to control digest operation. */
    const char * asn1;		/*!< RFC 3447 ASN1 oid string (in hex). */
    void * param;		/*!< Digest parameters. */
    void * salt;		/*!< Key salt[2*blocksize] */
};

static void rpmctxFini(void * _ctx)
{
    DIGEST_CTX ctx = (DIGEST_CTX) _ctx;
    if (ctx->param != NULL && ctx->paramsize > 0)
	memset(ctx->param, 0, ctx->paramsize);	/* In case it's sensitive */
    ctx->param = _free(ctx->param);
    if (ctx->salt != NULL && ctx->blocksize > 0)
	memset(ctx->salt, 0, 2*ctx->paramsize);	/* In case it's sensitive */
    ctx->salt = _free(ctx->salt);
    ctx->name = NULL;
    ctx->paramsize = 0;
    ctx->blocksize = 0;
    ctx->digestsize = 0;
    ctx->Reset = NULL;
    ctx->Update = NULL;
    ctx->Digest = NULL;
    ctx->hashalgo = (pgpHashAlgo)0;
    ctx->flags = (rpmDigestFlags)0;
    ctx->asn1 = NULL;
}

#define	rpmctxDbug	NULL
#define	rpmctxInit	NULL
RPMIOPOOL_MODULE(ctx)

pgpHashAlgo rpmDigestAlgo(DIGEST_CTX ctx)
{
    return (ctx != NULL ? ctx->hashalgo : PGPHASHALGO_NONE);
}

rpmDigestFlags rpmDigestF(DIGEST_CTX ctx)
{
    return (ctx != NULL ? ctx->flags : RPMDIGEST_NONE);
}

const char * rpmDigestName(DIGEST_CTX ctx)
{
    return (ctx != NULL ? ctx->name : "UNKNOWN");
}

size_t rpmDigestSize(DIGEST_CTX ctx)
{
    return (ctx != NULL ? ctx->digestsize : 0);
}

const char * rpmDigestASN1(DIGEST_CTX ctx)
{
    return (ctx != NULL ? ctx->asn1 : NULL);
}

DIGEST_CTX
rpmDigestDup(DIGEST_CTX octx)
{
    DIGEST_CTX nctx = rpmctxGetPool(_rpmctxPool);

    nctx->name = octx->name;
    nctx->digestsize = octx->digestsize;
    nctx->blocksize = octx->blocksize;
    nctx->paramsize = octx->paramsize;
    nctx->Reset = octx->Reset;
    nctx->Update = octx->Update;
    nctx->Digest = octx->Digest;
    nctx->hashalgo = octx->hashalgo;
    nctx->flags = octx->flags;
    nctx->asn1 = octx->asn1;
    nctx->param = (octx->param != NULL && octx->paramsize > 0)
	    ? memcpy(DRD_xmalloc(nctx->paramsize), octx->param, nctx->paramsize)
	    : NULL;
    nctx->salt = (octx->salt != NULL && octx->blocksize > 0)
	    ? memcpy(DRD_xmalloc(nctx->blocksize), octx->salt, nctx->blocksize)
	    : NULL;
    return (DIGEST_CTX)rpmioLinkPoolItem((rpmioItem)nctx, __FUNCTION__, __FILE__, __LINE__);
}

static int noopReset(void * param)
{
    return 0;
}

/* XXX impedance match bytes -> bits length. */
static int md6_Update(void * param, const byte * _data, size_t _len)
{
    return md6_update((md6_state *)param, (unsigned char *) _data, (rpmuint64_t)(8 * _len));
}

DIGEST_CTX
rpmDigestInit(pgpHashAlgo hashalgo, rpmDigestFlags flags)
{
    DIGEST_CTX ctx = rpmctxGetPool(_rpmctxPool);
    int xx;

    ctx->name = "";
    ctx->paramsize = 0;
    ctx->blocksize = 64;
    ctx->digestsize = 0;
    ctx->Reset = NULL;
    ctx->Update = NULL;
    ctx->Digest = NULL;
    ctx->hashalgo = hashalgo;
    ctx->flags = flags;
    ctx->asn1 = NULL;
    ctx->param = NULL;
    ctx->salt = NULL;

#define	RPMDIGEST_INIT(XY, NAME, DIGESTBITS, BLOCKBITS)	\
{   ctx->name = #NAME;					\
    ctx->digestsize = DIGESTBITS/8;			\
    ctx->blocksize = BLOCKBITS;				\
    ctx->paramsize = sizeof(XY##Param);			\
    ctx->param = DRD_xcalloc(1, ctx->paramsize);	\
    ctx->Reset = (int (*)(void *)) XY##Reset;		\
    ctx->Update = (int (*)(void *, const byte *, size_t)) XY##Update; \
    ctx->Digest = (int (*)(void *, byte *)) XY##Digest;	\
}

#define	RPMDIGEST_INIT_FAMILY(XY, NAME)			\
    case PGPHASHALGO_##NAME##_224:			\
	RPMDIGEST_INIT(XY, #NAME-224, 224, 64)		\
	goto XY;					\
    case PGPHASHALGO_##NAME##_256:			\
	RPMDIGEST_INIT(XY, #NAME-256, 256, 64)		\
	goto XY;					\
    case PGPHASHALGO_##NAME##_384:			\
	RPMDIGEST_INIT(XY, #NAME-384, 384, 64)		\
	goto XY;					\
    case PGPHASHALGO_##NAME##_512:			\
	RPMDIGEST_INIT(XY, #NAME-512, 512, 64)		\
	goto XY;					\
XY:

    /* iso.identified-organization.dod.internet.private.enterprise.36106 */
    /* 1.3.6.1.4.1.36106 */
    /* enterprise(1).wraptastic(36106).hash(2).XX */
    /* 06 09 2B 06 01 04 01 82 9A 0A 02 XX */

    switch (hashalgo) {
    case PGPHASHALGO_MD5:
	RPMDIGEST_INIT(md5, MD5, 128, 64)
	/* Object ID is 1.2.840.113549.2.5 */
	ctx->asn1 = "3020300c" "06082a864886f70d0205" "05000410";
	/* iso.member-body.us.rsadsi.pkcs.pkcs-1.4 (md5WithRSAEncryption) */
	/* 1.2.840.113549.1.1.4 */
	/* RSADSI digestAlgorithm MD5 */
	/* 1.2.840.113549.2.5 */
	break;
    case PGPHASHALGO_SHA1:
	RPMDIGEST_INIT(sha1, SHA1, 160, 64)
	/* Object ID is 1.3.14.3.2.26 */
	ctx->asn1 = "30213009" "06052b0e03021a" "05000414";
	/* iso.member-body.us.rsadsi.pkcs.pkcs-1.5 (sha1WithRSAEncryption) */
	/* 1.2.840.113549.1.1.5 */
	/* iso.member-body.us.x9-57.x9cm.3 (dsaWithSha1)*/
	/* 1.2.840.10040.4.3 */
	/* from NIST's OIW  (sha1) */
	/* 1.3.14.3.2.26 */
	/* from NIST OIW (sha-1WithRSAEncryption) */
	/* 1.3.14.3.2.29 */
	/* iso.member-body.us.ansi-x9-62.signatures.ecdsa-with-sha1 */
	/* 1.2.840.10045.4.1 */
	break;
    case PGPHASHALGO_RIPEMD128:
	RPMDIGEST_INIT(ripemd128, RIPEMD128, 128, 64)
	/* Object ID is 1.3.6.1.4.1.36106.2.105 */
	ctx->asn1 = "3022300c" "060A2B06010401829A0A0269" "05000410";
	break;
    case PGPHASHALGO_RIPEMD160:
	RPMDIGEST_INIT(ripemd160, RIPEMD160, 160, 64)
	/* Object ID is 1.3.36.3.2.1 */
	ctx->asn1 = "30213009" "06052b24030201" "05000414";
	/* rsaSignatureWithripemd160 */
	/* 1.3.36.3.3.1.2 */
	/* TeleTrust hash algorithm. */
	/* 1.3.36.3.2.1 */
	break;
    case PGPHASHALGO_RIPEMD256:
	RPMDIGEST_INIT(ripemd256, RIPEMD256, 256, 64)
	/* Object ID is 1.3.6.1.4.1.36106.2.111 */
	ctx->asn1 = "3032300c" "060A2B06010401829A0A026F" "05000420";
	break;
    case PGPHASHALGO_RIPEMD320:
	RPMDIGEST_INIT(ripemd320, RIPEMD320, 320, 64)
	/* Object ID is 1.3.6.1.4.1.36106.2.112 */
	ctx->asn1 = "303A300c" "060A2B06010401829A0A0270" "05000428";
	break;
    case PGPHASHALGO_SALSA10:
	RPMDIGEST_INIT(salsa10, SALSA10, 512, 64)
	/* Object ID is 1.3.6.1.4.1.36106.2.113 */
	ctx->asn1 = "3052300C" "060A2B06010401829A0A0271" "05000440";
	break;
    case PGPHASHALGO_SALSA20:
	RPMDIGEST_INIT(salsa20, SALSA20, 512, 64)
	/* Object ID is 1.3.6.1.4.1.36106.2.114 */
	ctx->asn1 = "3052300C" "060A2B06010401829A0A0272" "05000440";
	break;
    case PGPHASHALGO_TIGER192:
	/* XXX TIGER2 has different padding */
	RPMDIGEST_INIT(tiger, TIGER192, 192, 64)
	/* Object ID is 1.3.6.1.4.1.11591.12.2 */
	ctx->asn1 = "3029300d" "06092b06010401da470c02" "05000418";
	/* GNU.digestAlgorithm TIGER */
	/* 1.3.6.1.4.1.11591.12.2 */
	break;
    case PGPHASHALGO_MD2:
	RPMDIGEST_INIT(md2, MD2, 128, 16)
	/* Object ID is 1.2.840.113549.2.2 */
	ctx->asn1 = "3020300c" "06082a864886f70d0202" "05000410";
	/* iso.member-body.us.rsadsi.digestAlgorithm.md2 */
	/* 1.2.840.113549.2.2 */
	break;
    case PGPHASHALGO_MD4:
	RPMDIGEST_INIT(md4, MD4, 128, 64)
	/* Object ID is 1.2.840.113549.2.4 */
	ctx->asn1 = "3020300c" "06082a864886f70d0204" "05000410";
	/* iso.member-body.us.rsadsi.digestAlgorithm.md4 */
	/* 1.2.840.113549.2.4 */
	break;
    case PGPHASHALGO_CRC32:
	RPMDIGEST_INIT(sum32, CRC32, 32, 8)
	{   sum32Param * mp = (sum32Param *) ctx->param;
	    mp->update = (rpmuint32_t (*)(rpmuint32_t, const byte *, size_t)) __crc32;
	    mp->combine = (rpmuint32_t (*)(rpmuint32_t, rpmuint32_t, size_t)) __crc32_combine;
	}
	/* Object ID is 1.3.6.1.4.1.36106.2.?? */
	/* ctx->asn1 = "???" "060A2B06010401829A0A02??" "05000404"; */
	break;
    case PGPHASHALGO_ADLER32:
	RPMDIGEST_INIT(sum32, ADLER32, 32, 8)
	{   sum32Param * mp = (sum32Param *) ctx->param;
	    mp->update = (rpmuint32_t (*)(rpmuint32_t, const byte *, size_t)) __adler32;
	    mp->combine = (rpmuint32_t (*)(rpmuint32_t, rpmuint32_t, size_t)) __adler32_combine;
	}
	/* Object ID is 1.3.6.1.4.1.36106.2.?? */
	/* ctx->asn1 = "???" "060A2B06010401829A0A02??" "05000404"; */
	break;
    case PGPHASHALGO_JLU32:
	RPMDIGEST_INIT(sum32, JLU32, 32, 8)
	{   sum32Param * mp = (sum32Param *) ctx->param;
	    mp->update = (rpmuint32_t (*)(rpmuint32_t, const byte *, size_t)) jlu32l;
	}
	/* Object ID is 1.3.6.1.4.1.36106.2.?? */
	/* ctx->asn1 = "???" "060A2B06010401829A0A02??" "05000404"; */
	break;
    case PGPHASHALGO_XXH32:
	RPMDIGEST_INIT(xxh32, XXH32, 32, 8)
	/* Object ID is 1.3.6.1.4.1.36106.2.?? */
	/* ctx->asn1 = "???" "060A2B06010401829A0A02??" "05000404"; */
	break;
    case PGPHASHALGO_XXH64:
	RPMDIGEST_INIT(xxh64, XXH64, 64, 8)
	/* Object ID is 1.3.6.1.4.1.36106.2.?? */
	/* ctx->asn1 = "???" "060A2B06010401829A0A02??" "05000404"; */
	break;
    case PGPHASHALGO_CRC64:
	RPMDIGEST_INIT(sum64, CRC64, 64, 8)
	{   sum64Param * mp = (sum64Param *) ctx->param;
	    mp->update = (rpmuint64_t (*)(rpmuint64_t, const byte *, size_t)) __crc64;
	    mp->combine = (rpmuint64_t (*)(rpmuint64_t, rpmuint64_t, size_t)) __crc64_combine;
	}
	/* Object ID is 1.3.6.1.4.1.36106.2.?? */
	/* ctx->asn1 = "???" "060A2B06010401829A0A02??" "05000408"; */
	break;
    case PGPHASHALGO_SHA224:
	RPMDIGEST_INIT(sha224, SHA224, 224, 64)
	/* Object ID is 2.16.840.1.101.3.4.2.4 */
	ctx->asn1 = "302d300d" "0609608648016503040204" "0500041C";
	/* From RFC3874, Section 4 */
	/* 2.16.840.1.101.3.4.2.4 */
	break;
    case PGPHASHALGO_SHA256:
	RPMDIGEST_INIT(sha256, SHA256, 256, 64)
	/* Object ID is  2.16.840.1.101.3.4.2.1 */
	ctx->asn1 = "3031300d" "0609608648016503040201" "05000420";
	/* According to the OpenPGP draft rfc2440-bis06 */
	/* 2.16.840.1.101.3.4.2.1 */
	/* PKCS#1 sha256WithRSAEncryption */
	/* 1.2.840.113549.1.1.11 */
	break;
    case PGPHASHALGO_SHA384:
	RPMDIGEST_INIT(sha384, SHA384, 384, 128)
	/* Object ID is 2.16.840.1.101.3.4.2.2 */
	ctx->asn1 = "3041300d" "0609608648016503040202" "05000430";
	/* ??? */
	/* 2.16.840.1.101.3.4.2.2 */
	/* PKCS#1 sha384WithRSAEncryption */
	/* 1.2.840.113549.1.1.12 */
	break;
    case PGPHASHALGO_SHA512:
	RPMDIGEST_INIT(sha512, SHA512, 512, 128)
	/* Object ID is 2.16.840.1.101.3.4.2.3 */
	ctx->asn1 = "3051300d" "0609608648016503040203" "05000440";
	/* ??? */
	/* 2.16.840.1.101.3.4.2.3 */
	/* PKCS#1 sha512WithRSAEncryption */
	/* 1.2.840.113549.1.1.13 */
	break;
    case PGPHASHALGO_SKEIN_224: ctx->digestsize = 224/8; goto skein256;
    case PGPHASHALGO_SKEIN_256: ctx->digestsize = 256/8; goto skein256;
skein256:
	ctx->name = "SKEIN256";
	ctx->paramsize = sizeof(Skein_256_Ctxt_t);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	(void) Skein_256_Init((Skein_256_Ctxt_t *)ctx->param,
				(int)(8 * ctx->digestsize));
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) Skein_256_Update;
	ctx->Digest = (int (*)(void *, byte *)) Skein_256_Final;
	break;
    case PGPHASHALGO_SKEIN_384: ctx->digestsize = 384/8; goto skein512;
    case PGPHASHALGO_SKEIN_512: ctx->digestsize = 512/8; goto skein512;
skein512:
	ctx->name = "SKEIN512";
	ctx->paramsize = sizeof(Skein_512_Ctxt_t);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	(void) Skein_512_Init((Skein_512_Ctxt_t *)ctx->param,
				(int)(8 * ctx->digestsize));
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) Skein_512_Update;
	ctx->Digest = (int (*)(void *, byte *)) Skein_512_Final;
	break;
    case PGPHASHALGO_SKEIN_1024:
	ctx->name = "SKEIN1024";
	ctx->digestsize = 1024/8;
	ctx->paramsize = sizeof(Skein1024_Ctxt_t);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	(void) Skein1024_Init((Skein1024_Ctxt_t *)ctx->param,
				(int)(8 * ctx->digestsize));
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) Skein1024_Update;
	ctx->Digest = (int (*)(void *, byte *)) Skein1024_Final;
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.208+{0,1,2,3} */
	RPMDIGEST_INIT_FAMILY(arirang, ARIRANG)
	(void) arirangInit((arirangParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.184+{0,1,2,3} */
	/* ctx->asn1 = "???" "060B2B06010401829A0A028138" "0500041c"; */
	RPMDIGEST_INIT_FAMILY(blake, BLAKE)
	(void) blakeInit((blakeParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

    case PGPHASHALGO_BLAKE2B:
	RPMDIGEST_INIT(blake2b, BLAKE2B, 8*BLAKE2B_OUTBYTES, 8*BLAKE2B_BLOCKBYTES)
	(void) blake2bInit((blake2bParam *)ctx->param, (int)(8 * ctx->digestsize));
	/* Object ID is 1.3.6.1.4.1.36106.2.188 */
	ctx->asn1 = "3033300D" "060B2B06010401829A0A02813C" "05000420";
	break;
    case PGPHASHALGO_BLAKE2BP:
	RPMDIGEST_INIT(blake2bp, BLAKE2BP, 8*BLAKE2B_OUTBYTES, 8*BLAKE2B_BLOCKBYTES)
	(void) blake2bpInit((blake2bpParam *)ctx->param, (int)(8 * ctx->digestsize));
	/* Object ID is 1.3.6.1.4.1.36106.2.189 */
	ctx->asn1 = "3033300D" "060B2B06010401829A0A02813D" "05000420";
	break;
    case PGPHASHALGO_BLAKE2S:
	RPMDIGEST_INIT(blake2s, BLAKE2S, 8*BLAKE2S_OUTBYTES, 8*BLAKE2S_BLOCKBYTES)
	(void) blake2sInit((blake2sParam *)ctx->param, (int)(8 * ctx->digestsize));
	/* Object ID is 1.3.6.1.4.1.36106.2.190 */
	ctx->asn1 = "3033300D" "060B2B06010401829A0A02813E" "05000420";
	break;
    case PGPHASHALGO_BLAKE2SP:
	RPMDIGEST_INIT(blake2sp, BLAKE2SP, 8*BLAKE2S_OUTBYTES, 8*BLAKE2S_BLOCKBYTES)
	(void) blake2spInit((blake2spParam *)ctx->param, (int)(8 * ctx->digestsize));
	/* Object ID is 1.3.6.1.4.1.36106.2.191 */
	ctx->asn1 = "3033300D" "060B2B06010401829A0A02813F" "05000420";
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.168+{0,1,2,3} */
	RPMDIGEST_INIT_FAMILY(bmw, BMW)
	(void) bmwInit((bmwParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.224+{0,1,2,3} */
	RPMDIGEST_INIT_FAMILY(chi, CHI)
	(void) chiInit((chiParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.136+{0,1,2,3} */
	RPMDIGEST_INIT_FAMILY(cubehash, CUBEHASH)
	(void) cubehashInit((cubehashParam *)ctx->param, (int)(8 * ctx->digestsize),
				(int)((ctx->flags >> 8) & 0xff),
				(int)((ctx->flags     ) & 0xff));
	break;

    case PGPHASHALGO_ECHO_224:
	ctx->digestsize = 224/8;
	goto echo;
    case PGPHASHALGO_ECHO_256:
	ctx->digestsize = 256/8;
	goto echo;
    case PGPHASHALGO_ECHO_384:
	ctx->digestsize = 384/8;
	goto echo;
    case PGPHASHALGO_ECHO_512:
	ctx->digestsize = 512/8;
	goto echo;
echo:
	ctx->name = "ECHO";
	ctx->paramsize = sizeof(echo_hashState);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) _echo_Update;
	ctx->Digest = (int (*)(void *, byte *)) echo_Final;
	(void) echo_Init((echo_hashState *)ctx->param,
				(int)(8 * ctx->digestsize));
	break;

    case PGPHASHALGO_EDONR_224:
	ctx->digestsize = 224/8;
	goto edonr;
    case PGPHASHALGO_EDONR_256:
	ctx->digestsize = 256/8;
	goto edonr;
    case PGPHASHALGO_EDONR_384:
	ctx->digestsize = 384/8;
	goto edonr;
    case PGPHASHALGO_EDONR_512:
	ctx->digestsize = 512/8;
	goto edonr;
edonr:
	ctx->name = "EDON-R";
	ctx->paramsize = sizeof(edonr_hashState);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) edonr_Update;
	ctx->Digest = (int (*)(void *, byte *)) edonr_Final;
	(void) edonr_Init((edonr_hashState *)ctx->param,
				(int)(8 * ctx->digestsize));
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.156+{0,1,2,3| */
	RPMDIGEST_INIT_FAMILY(fugue, FUGUE)
	(void) fugueInit((fugueParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

    case PGPHASHALGO_GROESTL_224:
	ctx->digestsize = 224/8;
	goto groestl;
    case PGPHASHALGO_GROESTL_256:
	ctx->digestsize = 256/8;
	goto groestl;
    case PGPHASHALGO_GROESTL_384:
	ctx->digestsize = 384/8;
	goto groestl;
    case PGPHASHALGO_GROESTL_512:
	ctx->digestsize = 512/8;
	goto groestl;
groestl:
	ctx->name = "GROESTL";
	ctx->paramsize = sizeof(groestl_hashState);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) _groestl_Update;
	ctx->Digest = (int (*)(void *, byte *)) groestl_Final;
	(void) groestl_Init((groestl_hashState *)ctx->param,
				(int)(8 * ctx->digestsize));
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.248+{0,1,2,3| */
	RPMDIGEST_INIT_FAMILY(hamsi, HAMSI)
	(void) hamsiInit((hamsiParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.232+{0,1,2,3| */
	RPMDIGEST_INIT_FAMILY(jh, JH)
	(void) jhInit((jhParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

#ifdef	REF
    case GOSTR3411_CP:
	/* Object ID is 1.2.643.2.2.3 */
	/* ctx->asn1 = "???" "06062A8503020203" "???"; */
	ctx->asn1 = "2a8503020203";	/* XXX bogus in libgcrypt */
	/* iso.member-body.ru.rans.cryptopro.3 (gostR3411-94-with-gostR3410-2001)   */
	/* 1.2.643.2.2.3 */
	/* iso.member-body.ru.rans.cryptopro.9 (gostR3411-94) */
	/* 1.2.643.2.2.9 */
	break;
#endif
    case PGPHASHALGO_KECCAK_224:
	ctx->digestsize = 224/8;
	/* ctx->asn1 = "???" "0609608648016503040207" "0500041C"; */
	/* joint-iso-itu-t(2) country(16) us(840) organization(1) gov(101) csor(3) nistAlgorithm(4) hashAlgs(2) 7 */
	/* 2.16.840.1.101.3.4.2.7 */
	/* PKCS#1 sha3_224WithRSAEncryption */
	/* ? */
	goto keccak;
    case PGPHASHALGO_KECCAK_256:
	ctx->digestsize = 256/8;
	/* ctx->asn1 = "???" "0609608648016503040208" "05000420"; */
	/* joint-iso-itu-t(2) country(16) us(840) organization(1) gov(101) csor(3) nistAlgorithm(4) hashAlgs(2) 8 */
	/* 2.16.840.1.101.3.4.2.8 */
	/* PKCS#1 sha3_256WithRSAEncryption */
	/* ? */
	goto keccak;
    case PGPHASHALGO_KECCAK_384:
	ctx->digestsize = 384/8;
	/* ctx->asn1 = "???" "0609608648016503040209" "05000430"; */
	/* joint-iso-itu-t(2) country(16) us(840) organization(1) gov(101) csor(3) nistAlgorithm(4) hashAlgs(2) 9 */
	/* 2.16.840.1.101.3.4.2.9 */
	/* PKCS#1 sha3_384WithRSAEncryption */
	/* ? */
	goto keccak;
    case PGPHASHALGO_KECCAK_512:
	ctx->digestsize = 512/8;
	/* ctx->asn1 = "???" "060960864801650304020a" "05000440"; */
	/* joint-iso-itu-t(2) country(16) us(840) organization(1) gov(101) csor(3) nistAlgorithm(4) hashAlgs(2) 10 */
	/* 2.16.840.1.101.3.4.2.10 */
	/* PKCS#1 sha3_512WithRSAEncryption */
	/* ? */
	goto keccak;
#ifdef	REF
    case SHAKE128:
	/* ctx->asn1 = "???" "060960864801650304020b" "05000410"; */
	/* joint-iso-itu-t(2) country(16) us(840) organization(1) gov(101) csor(3) nistAlgorithm(4) hashAlgs(2) 11 */
	/* 2.16.840.1.101.3.4.2.11 */
	/* PKCS#1 shake128WithRSAEncryption */
	/* ? */
	break;
    case SHAKE256:
	/* ctx->asn1 = "???" "060960864801650304020c" "05000420"; */
	/* joint-iso-itu-t(2) country(16) us(840) organization(1) gov(101) csor(3) nistAlgorithm(4) hashAlgs(2) 12 */
	/* 2.16.840.1.101.3.4.2.12 */
	/* PKCS#1 shake256WithRSAEncryption */
	/* ? */
	break;
#endif
keccak:
	ctx->name = "KECCAK";
	ctx->paramsize = sizeof(keccak_hashState);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) _keccak_Update;
	ctx->Digest = (int (*)(void *, byte *)) keccak_Final;
	(void) keccak_Init((keccak_hashState *)ctx->param,
				(int)(8 * ctx->digestsize));
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.212+{0,1,2,3| */
	RPMDIGEST_INIT_FAMILY(lane, LANE)
	(void) laneInit((laneParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

	/* Object ID is 1.3.6.1.4.1.36106.2.216+{0,1,2,3| */
	RPMDIGEST_INIT_FAMILY(luffa, LUFFA)
	(void) luffaInit((luffaParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

    case PGPHASHALGO_MD6_224:
	ctx->digestsize = 224/8;
	goto md6;
    case PGPHASHALGO_MD6_256:
	ctx->digestsize = 256/8;
	goto md6;
    case PGPHASHALGO_MD6_384:
	ctx->digestsize = 384/8;
	goto md6;
    case PGPHASHALGO_MD6_512:
	ctx->digestsize = 512/8;
	goto md6;
md6:
	ctx->name = "MD6";
	ctx->paramsize = sizeof(md6_state);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) md6_Update;
	ctx->Digest = (int (*)(void *, byte *)) md6_final;
	{   int d = (8 * ctx->digestsize);	/* no. of bits in digest */
	    int L = md6_default_L;		/* no. of parallel passes */
	    unsigned char *K = NULL;		/* key */
	    int keylen = 0;			/* key length (bytes) */
	    int r = md6_default_r(d, keylen);	/* no. of rounds */

	    if (ctx->flags != 0) {
		r = ((ctx->flags >> 8) & 0xffff);
		L = ((ctx->flags     ) & 0xff);
		if (r <= 0 || r > 255) r = md6_default_r(d, keylen);
	    }
	    (void) md6_full_init((md6_state *)ctx->param,
			d, K, keylen, L, r);
	}
	break;

    case PGPHASHALGO_RG32_256:
	RPMDIGEST_INIT(rg32, RG32, 256, 64)
	(void) rg32Init((rg32Param *)ctx->param, (int)(8 * ctx->digestsize));
	/* Object ID is 1.3.6.1.4.1.36106.2.192 */
	ctx->asn1 = "3033300D" "060B2B06010401829A0A028140" "05000420";
	break;
    case PGPHASHALGO_RG64_256:
	RPMDIGEST_INIT(rg64, RG64, 256, 64)
	(void) rg64Init((rg64Param *)ctx->param, (int)(8 * ctx->digestsize));
	/* Object ID is 1.3.6.1.4.1.36106.2.204 */
	ctx->asn1 = "3033300D" "060B2B06010401829A0A02814C" "05000420";
	break;

#ifdef	NOTYET
    case PGPHASHALGO_SHABAL_192: ctx->digestsize = 192/8; goto shabal;
#endif
	/* Object ID is 1.3.6.1.4.1.36106.2.176+{0,1,2,3| */
	RPMDIGEST_INIT_FAMILY(shabal, SHABAL)
	(void) shabalInit((shabalParam *)ctx->param, (int)(8 * ctx->digestsize));
	break;

    case PGPHASHALGO_SHAVITE3_224: ctx->digestsize = 224/8; goto shavite3;
    case PGPHASHALGO_SHAVITE3_256: ctx->digestsize = 256/8; goto shavite3;
    case PGPHASHALGO_SHAVITE3_384: ctx->digestsize = 384/8; goto shavite3;
    case PGPHASHALGO_SHAVITE3_512: ctx->digestsize = 512/8; goto shavite3;
shavite3:
	ctx->name = "SHAVITE3";
	ctx->paramsize = sizeof(shavite3_hashState);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) _shavite3_Update;
	ctx->Digest = (int (*)(void *, byte *)) shavite3_Final;
	(void) shavite3_Init((shavite3_hashState *)ctx->param,
				(int)(8 * ctx->digestsize));
	break;

    case PGPHASHALGO_SIMD_224: ctx->digestsize = 224/8; goto simd;
    case PGPHASHALGO_SIMD_256: ctx->digestsize = 256/8; goto simd;
    case PGPHASHALGO_SIMD_384: ctx->digestsize = 384/8; goto simd;
    case PGPHASHALGO_SIMD_512: ctx->digestsize = 512/8; goto simd;
simd:
	ctx->name = "SIMD";
	ctx->paramsize = sizeof(simd_hashState);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) _simd_Update;
	ctx->Digest = (int (*)(void *, byte *)) simd_Final;
	(void) simd_Init((simd_hashState *)ctx->param,
				(int)(8 * ctx->digestsize));
	break;

    case PGPHASHALGO_TIB3_224: ctx->digestsize = 224/8; goto tib3;
    case PGPHASHALGO_TIB3_256: ctx->digestsize = 256/8; goto tib3;
    case PGPHASHALGO_TIB3_384: ctx->digestsize = 384/8; goto tib3;
    case PGPHASHALGO_TIB3_512: ctx->digestsize = 512/8; goto tib3;
tib3:
	ctx->name = "TIB3";
	ctx->paramsize = sizeof(tib3_hashState);
	ctx->param = DRD_xcalloc(1, ctx->paramsize);
	ctx->Reset = (int (*)(void *)) noopReset;
	ctx->Update = (int (*)(void *, const byte *, size_t)) _tib3_Update;
	ctx->Digest = (int (*)(void *, byte *)) tib3_Final;
	(void) tib3_Init((tib3_hashState *)ctx->param,
				(int)(8 * ctx->digestsize));
	break;

    case PGPHASHALGO_HAVAL_5_160:
    default:
	(void)rpmioFreePoolItem((rpmioItem)ctx, __FUNCTION__, __FILE__, __LINE__);
	return NULL;
	break;
    }

    xx = (*ctx->Reset) (ctx->param);

DPRINTF((stderr, "==> ctx %p ==== Init(%s, %x) param %p\n", ctx, ctx->name, flags, ctx->param));
    return (DIGEST_CTX)rpmioLinkPoolItem((rpmioItem)ctx, __FUNCTION__, __FILE__, __LINE__);
}

int
rpmDigestUpdate(DIGEST_CTX ctx, const void * data, size_t len)
{
    if (ctx == NULL)
	return -1;

DPRINTF((stderr, "==> ctx %p ==== Update(%s,%p[%u]) param %p\n", ctx, ctx->name, data, (unsigned)len, ctx->param));
    return (*ctx->Update) (ctx->param, (byte *)data, len);
}

#define	HMAC_IPAD	0x36
#define	HMAC_OPAD	0x5c

int
rpmDigestFinal(DIGEST_CTX ctx, void * datap, size_t *lenp, int asAscii)
{
    byte * digest;
    char * t;

    if (ctx == NULL)
	return -1;
    digest = (byte *) DRD_xmalloc(ctx->digestsize);

DPRINTF((stderr, "==> ctx %p ==== Final(%s,%p,%p,%d) param %p digest %p[%u]\n", ctx, ctx->name, datap, lenp, asAscii, ctx->param, digest, (unsigned)ctx->digestsize));
    (void) (*ctx->Digest) (ctx->param, digest);

    /* If keyed HMAC, re-hash with key material. */
    if (ctx->salt != NULL) {
	DIGEST_CTX kctx = rpmDigestInit(ctx->hashalgo, RPMDIGEST_NONE);
	byte * salt = (byte *) ctx->salt;
	byte * kdigest = NULL;
	size_t kdigestlen = 0;
	unsigned i;
	for (i = 0; i < ctx->blocksize; i++)
	    salt[i] ^= HMAC_OPAD;
	rpmDigestUpdate(kctx, ctx->salt, ctx->blocksize);
	ctx->salt = _free(ctx->salt);
	rpmDigestUpdate(kctx, digest, ctx->digestsize);
	(void) rpmDigestFinal(kctx, &kdigest, &kdigestlen, 0);
	memcpy(digest, kdigest, kdigestlen);
	kdigest = _free(kdigest);
    }

    /* Return final digest. */
    if (!asAscii) {
	if (lenp) *lenp = ctx->digestsize;
	if (datap) {
	    *(byte **)datap = digest;
	    digest = NULL;
	}
    } else {
	if (lenp) *lenp = (2*ctx->digestsize);
	if (datap) {
	    const byte * s = (const byte *) digest;
	    static const char hex[] = "0123456789abcdef";
	    size_t i;

	    *(char **)datap = t = (char *) DRD_xmalloc((2*ctx->digestsize) + 1);
	    for (i = 0 ; i < ctx->digestsize; i++) {
		*t++ = hex[ (unsigned)((*s >> 4) & 0x0f) ];
		*t++ = hex[ (unsigned)((*s++   ) & 0x0f) ];
	    }
	    *t = '\0';
	}
    }
    if (digest) {
	memset(digest, 0, ctx->digestsize);	/* In case it's sensitive */
	free(digest);
    }
    (void)rpmioFreePoolItem((rpmioItem)ctx, __FUNCTION__, __FILE__, __LINE__);
    return 0;
}

int
rpmHmacInit(DIGEST_CTX ctx, const void * key, size_t keylen)
{
    int rc = 0;

    if (ctx == NULL)
	return -1;
    if (key != NULL) {
	byte * salt = (byte *) DRD_xcalloc(1, ctx->blocksize);
	unsigned i;
	if (keylen == 0) keylen = strlen((char *)key);
	ctx->salt = salt;
DPRINTF((stderr, "==> ctx %p ==== HMAC(%s,%p[%u])\n", ctx, ctx->name, key, (unsigned)keylen));
	if (keylen > ctx->blocksize) {
	    /* If key is larger than digestlen, then hash the material. */
	    DIGEST_CTX kctx = rpmDigestInit(ctx->hashalgo, RPMDIGEST_NONE);
	    byte * kdigest = NULL;
	    size_t kdigestlen = 0;
	    rpmDigestUpdate(kctx, key, keylen);
	    (void) rpmDigestFinal(kctx, &kdigest, &kdigestlen, 0);
	    memcpy(ctx->salt, kdigest, kdigestlen);
	    kdigest = _free(kdigest);
	} else
	    memcpy(ctx->salt, key, keylen);

	salt = (byte *)ctx->salt;
	for (i = 0; i < ctx->blocksize; i++)
	    salt[i] ^= HMAC_IPAD;
	rpmDigestUpdate(ctx, ctx->salt, ctx->blocksize);
	for (i = 0; i < ctx->blocksize; i++)
	    salt[i] ^= HMAC_IPAD;
    }
    return rc;
}
