
/** \ingroup rpmio
 * \file rpmio/mount.c
 */

#if defined(__linux__)

#include "system.h"

#include <sys/mount.h>

#include "rpmio.h"
#include "poptIO.h"
#include "debug.h"

unsigned rpmioMountFlags = 0;
struct poptOption rpmioMountFlagsTable[] = {

#if defined(MS_RDONLY)
 { "read-only", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_RDONLY,
	N_("read-only"), N_("MS_RDONLY") },
#endif
#if defined(MS_RDONLY)
 { "ro", '\0', POPT_BIT_SET|POPT_ARGFLAG_DOC_HIDDEN, &rpmioMountFlags,	MS_RDONLY,
	N_("read-only"), N_("MS_RDONLY") },
#endif
#if defined(MS_RDONLY)
 { "read-write", '\0', POPT_BIT_CLR,	&rpmioMountFlags,	MS_RDONLY,
	N_("read-write"), N_("~MS_RDONLY") },
#endif
#if defined(MS_RDONLY)
 { "rw", '\0', POPT_BIT_CLR|POPT_ARGFLAG_DOC_HIDDEN, &rpmioMountFlags,	MS_RDONLY,
	N_("read-write"), N_("~MS_RDONLY") },
#endif

#if defined(MNT_RDONLY)
 { "read-only", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_RDONLY,
	N_("read-only"), N_("MNT_RDONLY") },
#endif
#if defined(MNT_RDONLY)
 { "ro", '\0', POPT_BIT_SET|POPT_ARGFLAG_DOC_HIDDEN, &rpmioMountFlags,	MNT_RDONLY,
	N_("read-only"), N_("MNT_RDONLY") },
#endif
#if defined(MNT_RDONLY)
 { "read-write", '\0', POPT_BIT_CLR,	&rpmioMountFlags,	MNT_RDONLY,
	N_("read-write"), N_("~MNT_RDONLY") },
#endif
#if defined(MNT_RDONLY)
 { "rw", '\0', POPT_BIT_CLR|POPT_ARGFLAG_DOC_HIDDEN, &rpmioMountFlags,	MNT_RDONLY,
	N_("read-write"), N_("~MS_RDONLY") },
#endif

#if defined(MS_NOSUID)
 { "suid", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_NOSUID,
	N_("suid"), N_("MS_NOSUID") },
#endif
#if defined(MNT_NOSUID)
 { "suid", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NOSUID,
	N_("nosuid"), N_("MNT_NOSUID") },
#endif

#if defined(MS_NOEXEC)
 { "exec", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_NOEXEC,
	N_("exec"), N_("MS_NOEXEC") },
#endif
#if defined(MNT_NOEXEC)
 { "exec", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NOEXEC,
	N_("exec"), N_("MNT_NOEXEC") },
#endif

#if defined(MS_SYNCHRONOUS)
 { "sync", '\0', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_SYNC,
	N_("sync"), N_("MS_SYNCHRONOUS") },
#endif
#if defined(MNT_SYNCHRONOUS)
 { "sync", '\0', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_SYNC,
	N_("sync"), N_("MNT_SYNCHRONOUS") },
#endif

#if defined(MS_REMOUNT)
 { "remount", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_REMOUNT,
	N_("remount"), N_("MS_REMOUNT") },
#endif
#if defined(MS_MANDLOCK)
 { "mand", '\0', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_MANDLOCK,
	N_("mand"), N_("MS_MANDLOCK") },
#endif
#if defined(MS_DIRSYNC)
 { "dirsync", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_DIRSYNC,
	N_("dirsync"), N_("MS_DIRSYNC") },
#endif

#if defined(MS_NOATIME)
 { "atime", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_NOATIME,
	N_("atime"), N_("MS_NOATIME") },
#endif
#if defined(MNT_NOATIME)
 { "atime", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NOATIME,
	N_("atime"), N_("MNT_NOATIME") },
#endif

#if defined(MS_NODIRATIME)
 { "diratime", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_NODIRATIME,
	N_("diratime"), N_("MS_NODIRATIME") },
#endif
#if defined(MS_BIND)
 { "bind", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MS_BIND,
	N_("bind"), N_("MS_BIND") },
#endif
#if defined(MS_MOVE)
 { "move", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MS_MOVE,
	N_("move"), N_("MS_MOVE") },
#endif
#if defined(MS_REC)
 { "rec", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MS_REC,
	N_("rec"), N_("MS_REC") },
#endif
#if defined(MS_SILENT)
 { "silent", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_SILENT,
	N_("silent"), N_("MS_SILENT") },
#endif

#if defined(MS_POSIXACL)
 { "posixacl", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_POSIXACL,
	N_("posixacl"), N_("MS_POSIXACL") },
#endif
#if defined(MNT_ACLS)
 { "acls", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MNT_ACLS,
	N_("acls"), N_("MNT_ACLS") },
#endif
#if defined(MNT_NFS4ACLS)
 { "nfsv4acls", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MNT_NFS4ACLS,
	N_("nfsv4acls"), N_("MNT_NFS4ACLS") },
#endif

#if defined(MS_UNBINDABLE)
 { "unbindable", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_UNBINDABLE,
	N_("bindable"), N_("MS_UNBINDABLE") },
#endif
#if defined(MS_PRIVATE)
 { "private", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_PRIVATE,
	N_("private"), N_("MS_PRIVATE") },
#endif
#if defined(MS_SLAVE)
 { "slave", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MS_SLAVE,
	N_("slave"), N_("MS_SLAVE") },
#endif
#if defined(MS_SHARED)
 { "shared", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_SHARED,
	N_("shared"), N_("MS_SHARED") },
#endif
#if defined(MS_RELATIME)
 { "relatime", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_RELATIME,
	N_("relatime"), N_("MS_RELATIME") },
#endif
#if defined(MS_KERNMOUNT)
 { "kernmount", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_KERNMOUNT,
	N_("kernmount"), N_("MS_KERNMOUNT") },
#endif
#if defined(MS_I_VERSION)
 { "iversion", '\0', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_I_VERSION,
	N_("iversion"), N_("MS_I_VERSION") },
#endif
#if defined(MS_STRICTATIME)
 { "strictatime", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_STRICTATIME,
	N_("strictatime"), N_("MS_STRICTATIME") },
#endif
#if defined(MS_LAZYTIME)
 { "lazytime", '\0', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_LAZYTIME,
	N_("lazytime"), N_("MS_LAZYTIME") },
#endif
#if defined(MS_ACTIVE)
 { "active", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MS_ACTIVE,
	N_("active"), N_("MS_ACTIVE") },
#endif
#if defined(MS_NOUSER)
 { "user", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MS_NOUSER,
	N_("user"), N_("MS_NOUSER") },
#endif

#if defined(MNT_NODEV)
 { "dev", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NODEV,
	N_("dev"), N_("MNT_NODEV") },
#endif
#if defined(MNT_UNION)
 { "union", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MNT_UNION,
	N_("union"), N_("MNT_UNION") },
#endif
#if defined(MNT_ASYNC)
 { "async", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MNT_ASYNC,
	N_("async"), N_("MNT_ASYNC") },
#endif
#if defined(MNT_CPROTECT)
 { "cprotect", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_CPROTECT,
	N_("cprotect"), N_("MNT_CPROTECT") },
#endif
#if defined(MNT_EXPORTED)
 { "exported", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_EXPORTED,
	N_("exported"), N_("MNT_EXPORTED") },
#endif
#if defined(MNT_QUARANTINE)
 { "quarantine", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_QUARANTINE,
	N_("quarantine"), N_("MNT_QUARANTINE") },
#endif
#if defined(MNT_LOCAL)
 { "local", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MNT_LOCAL,
	N_("local"), N_("MNT_LOCAL") },
#endif
#if defined(MNT_QUOTA)
 { "quota", '\0', POPT_BIT_SET,		&rpmioMountFlags,	MNT_QUOTA,
	N_("quota"), N_("MNT_QUOTA") },
#endif
#if defined(MNT_ROOTFS)
 { "rootfs", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_ROOTFS,
	N_("rootfs"), N_("MNT_ROOTFS") },
#endif
#if defined(MNT_DOVOLFS)
 { "dovolfs", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_DOVOLFS,
	N_("dovolfs"), N_("MNT_DOVOLFS") },
#endif
#if defined(MNT_DONTBROWSE)
 { "dontbrowse", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_DONTBROWSE,
	N_("dontbrowse"), N_("MNT_DONTBROWSE") },
#endif

#if defined(MNT_IGNORE_OWNERSHIP)
 { "ignore-ownership", '\0', POPT_BIT_SET, &rpmioMountFlags,	MNT_IGNORE_OWNERSHIP,
	N_("ignore-ownership"), N_("MNT_IGNORE_OWNERSHIP") },
#endif
#if defined(MNT_UNKNOWNPERMOSSIONS)
 { "unknownpermissions", '\0', POPT_BIT_SET|POPT_ARGFLAG_DOC_HIDDEN, &rpmioMountFlags,	MNT_UNKNOWNPERMISSIONS,
	N_("unknownpermissions"), N_("MNT_UNKNOWNPERMISSIONS") },
#endif

#if defined(MNT_AUTOMOUNTED)
 { "automounted", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_AUTOMOUNTED,
	N_("automounted"), N_("MNT_AUTOMOUNTED") },
#endif
#if defined(MNT_JOURNALED)
 { "journaled", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_JOURNALED,
	N_("journaled"), N_("MNT_JOURNALED") },
#endif
#if defined(MNT_NOUSERXATTR)
 { "userxattr", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NOUSERXATTR,
	N_("userxattr"), N_("MNT_NOUSERXATTR") },
#endif
#if defined(MNT_DEFWRITE)
 { "defwrite", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_DEFWRITE,
	N_("defwrite"), N_("MNT_DEFWRITE") },
#endif
#if defined(MNT_MULTILABEL)
 { "multilabel", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_MULTILABEL,
	N_("multilabel"), N_("MNT_MULTILABEL") },
#endif

#if defined(MNT_UPDATE)
 { "update", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_UPDATE,
	N_("update"), N_("MNT_UPDATE") },
#endif
#if defined(MNT_NOBLOCK)
 { "block", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NOBLOCK,
	N_("block"), N_("MNT_NOBLOCK") },
#endif
#if defined(MNT_RELOAD)
 { "reload", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_RELOAD,
	N_("reload"), N_("MNT_RELOAD") },
#endif
#if defined(MNT_CMDFLAGS)
 { "cmdflags", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_CMDFLAGS,
	N_("cmdflags"), N_("MNT_CMDFLAGS") },
#endif

#if defined(MNT_NOSYMFOLLOW)
 { "symfollow", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NOSYMFOLLOW,
	N_("symfollow"), N_("MNT_NOSYMFOLLOW") },
#endif
#if defined(MNT_NOCLUSTERR)
 { "clusterr", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NOCLUSTERR,
	N_("clusterr"), N_("MNT_NOCLUSTERR") },
#endif
#if defined(MNT_NOCLUSTERW)
 { "clusterw", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioMountFlags,	MNT_NOCLUSTERW,
	N_("clusterw"), N_("MNT_NOCLUSTERW") },
#endif
#if defined(MNT_SUIDDIR)
 { "suiddir", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_SUIDDIR,
	N_("suiddir"), N_("MNT_SUIDDIR") },
#endif
#if defined(MNT_SOFTDEP)
 { "soft-updates", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_SOFTDEP,
	N_("soft-updates"), N_("MNT_SOFTDEP") },
#endif
#if defined(MNT_SUJ)
 { "journaled-soft-updates", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_SUJ,
	N_("journaled-soft-updates"), N_("MNT_SUJ") },
#endif
#if defined(MNT_MULTILABEL)
 { "multlabel", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_MULTILABEL,
	N_("multlabel"), N_("MNT_MULTILABEL") },
#endif
#if defined(MNT_GJOURNAL)
 { "gjournal", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_GJOURNAL,
	N_("gjournal"), N_("MNT_GJOURNAL") },
#endif
#if defined(MNT_GJOURNAL)
 { "gjournal", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_GJOURNAL,
	N_("gjournal"), N_("MNT_GJOURNAL") },
#endif
#if defined(MNT_VERIFIED)
 { "verified", '\0', POPT_BIT_SET,	&rpmioMountFlags,	MNT_VERIFIED,
	N_("verified"), N_("MNT_VERIFIED") },
#endif

  POPT_TABLEEND
};

unsigned rpmioUmountFlags = 0;
struct poptOption rpmioUmountFlagsTable[] = {
#if defined(MNT_FORCE)
 { "force", '\0', POPT_BIT_SET,		&rpmioUmountFlags,	MNT_FORCE,
	N_("force"), N_("MNT_FORCE") },
#endif
#if defined(MNT_DETACH)
 { "detach", '\0', POPT_BIT_SET,	&rpmioUmountFlags,	MNT_DETACH,
	N_("detach"), N_("MNT_DETACH") },
#endif
#if defined(MNT_EXPIRE)
 { "expire", '\0', POPT_BIT_SET,	&rpmioUmountFlags,	MNT_EXPIRE,
	N_("expire"), N_("MNT_EXPIRE") },
#endif
#if defined(UMOUNT_NOFOLLOW)
 { "follow", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioUmountFlags,	UMOUNT_NOFOLLOW,
	N_("follow"), N_("UMOUNT_NOFOLLOW") },
#endif
  POPT_TABLEEND
};

int Mount(const char *source, const char *target,
		const char *filesystemtype, unsigned long mountflags,
		const void *data)
{
    int rc = mount(source, target, filesystemtype, mountflags, data);
    char * mflags = rpmioB2N(rpmioMountFlagsTable, mountflags);
fprintf(stderr, "<-- %s(%s, %s, %s, %s, %p) rc %d\n", __FUNCTION__, source, target, filesystemtype, mflags, data, rc);
    mflags = _free(mflags);
    return rc;
}

int Umount(const char *target)
{
    int rc = umount(target);
fprintf(stderr, "<-- %s(%s) rc %d\n", __FUNCTION__, target, rc);
    return rc;
}

int Umount2(const char *target, int flags)
{
    int rc = umount2(target, flags);
    char * uflags = rpmioB2N(rpmioUmountFlagsTable, flags);
fprintf(stderr, "<-- %s(%s, %s) rc %d\n", __FUNCTION__, target, uflags, rc);
    uflags = _free(uflags);
    return rc;
}

#endif	/* __linux__ */
