/** \ingroup rpmio
 * \file rpmio/poptIO.c
 *  Popt tables for all rpmio-only executables.
 */

#include "system.h"

#if defined(HAVE_SYS_MOUNT_H)
# include <sys/mount.h>
#endif
#if defined(HAVE_SYS_IPC_H)
# include <sys/ipc.h>
#endif
#if defined(HAVE_SYS_RESOURCE_H)
# include <sys/resource.h>
#endif
#if defined(HAVE_MALLOC_H)
# include <malloc.h>
#endif

#include <poptIO.h>

#if defined(WITH_BEECRYPT)
#include <rpmbc.h>
#endif
#if defined(WITH_CDSA)
#include <rpmcdsa.h>
#endif
#if defined(WITH_GCRYPT)
#include <rpmgc.h>
#endif
#if defined(WITH_NSS)
#include <rpmnss.h>
#endif
#if defined(WITH_SSL)
#include <rpmssl.h>
#endif
#if defined(WITH_TOMCRYPT)
#include <rpmltc.h>
#endif

#include <rpmasn.h>
#include <rpmtpm.h>
#include <rpmgfs.h>

#include <rpmaio.h>
#include <rpmaug.h>
#include <rpmbag.h>
#include <rpmbf.h>
#include <rpmcap.h>
#include <rpmct.h>
#include <rpmcudf.h>
#include <rpmcvs.h>
#include <rpmeio.h>
#include <rpmev.h>
#include <rpmficl.h>
#include <rpmgit.h>
#include <rpmhkp.h>
#include <rpmjss.h>
#include <rpmmrb.h>
#include <rpmmsq.h>
#include <rpmnix.h>
#include <rpmodbc.h>
#include <rpmperl.h>
#include <rpmpython.h>
#include <rpmruby.h>
#include <rpmsed.h>
#include <rpmsql.h>
#include <rpmsquirrel.h>
#include <rpmsvn.h>
extern int _rpmsvn_debug;
#include <rpmtcl.h>
#include <rpmw.h>

#include <rpmsm.h>
#include <rpmsp.h>
#include <rpmsx.h>

#include <rpmzstd.h>

#include <set.h>

#include "debug.h"

#ifdef __cplusplus
GENfree(rpmioP)
#endif	/* __cplusplus */

const char *__progname;
int __progname_key;
int __progname_projid;

#if !defined(POPT_ARGFLAG_TOGGLE)	/* XXX compat with popt < 1.15 */
#define	POPT_ARGFLAG_TOGGLE	0
#endif
#define POPT_SHOWVERSION	-999
#define POPT_UNDEFINE		-994
#define	POPT_CRYPTO		-993

int __debug = 0;

extern int _ar_debug;
extern int _av_debug;
extern int _cpio_debug;
extern int _rpmctx_debug;
extern int _dav_cb_debug;
extern int _dav_debug;
extern int _dav_hook_debug;
extern int _ftp_debug;
extern int _fts_debug;
extern int _rpmht_debug;
extern int _html_debug;
extern int _iosm_debug;
extern int noLibio;
extern int _pgp_debug;
extern int _pgp_print;
extern int _rpmio_debug;
extern int _rpmiob_debug;
extern int _rpmlua_debug;
extern int _rpmmqtt_debug;
extern int _rpmsq_debug;
extern int _rpmzq_debug;
extern int _tar_debug;
extern int _xar_debug;

const char * rpmioPipeOutput = NULL;

const char * rpmioRootDir = "/";

const char *rpmioEVR = VERSION;

static int rpmioInitialized = -1;

#ifdef	NOTYET
#ifdef WITH_LUA
extern const char *rpmluaFiles;
#endif

static char *rpmpoptfiles = RPMPOPTFILES;
#endif

int _rpmio_popt_context_flags = 0;

pgpHashAlgo rpmioDigestHashAlgo = (pgpHashAlgo) -1;

/**
 * Digest options using popt.
 */
struct poptOption rpmioDigestPoptTable[] = {
 { "md2", '\0', POPT_ARG_VAL, 	&rpmioDigestHashAlgo, PGPHASHALGO_MD2,
	N_("MD2 digest (RFC-1319)"), NULL },
 { "md4", '\0', POPT_ARG_VAL, 	&rpmioDigestHashAlgo, PGPHASHALGO_MD4,
	N_("MD4 digest"), NULL },
 { "md5", '\0', POPT_ARG_VAL, 	&rpmioDigestHashAlgo, PGPHASHALGO_MD5,
	N_("MD5 digest (RFC-1321)"), NULL },
 { "sha1",'\0', POPT_ARG_VAL, 	&rpmioDigestHashAlgo, PGPHASHALGO_SHA1,
	N_("SHA-1 digest (FIPS-180-1)"), NULL },
 { "sha224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHA224,
	N_("SHA-224 digest (FIPS-180-2)"), NULL },
 { "sha256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHA256,
	N_("SHA-256 digest (FIPS-180-2)"), NULL },
 { "sha384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHA384,
	N_("SHA-384 digest (FIPS-180-2)"), NULL },
 { "sha512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHA512,
	N_("SHA-512 digest (FIPS-180-2)"), NULL },
 { "skein-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SKEIN_224,
	N_("SKEIN-224 digest"), NULL },
 { "skein-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SKEIN_256,
	N_("SKEIN-256 digest"), NULL },
 { "skein-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SKEIN_384,
	N_("SKEIN-384 digest"), NULL },
 { "skein-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SKEIN_512,
	N_("SKEIN-512 digest"), NULL },
 { "skein-1024",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SKEIN_1024,
	N_("SKEIN-1024 digest"), NULL },
 { "arirang-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_ARIRANG_224,
	N_("Arirang-224 digest"), NULL },
 { "arirang-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_ARIRANG_256,
	N_("Arirang-256 digest"), NULL },
 { "arirang-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_ARIRANG_384,
	N_("Arirang-384 digest"), NULL },
 { "arirang-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_ARIRANG_512,
	N_("Arirang-512 digest"), NULL },
 { "blake-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BLAKE_224,
	N_("Blake-224 digest"), NULL },
 { "blake-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BLAKE_256,
	N_("Blake-256 digest"), NULL },
 { "blake-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BLAKE_384,
	N_("Blake-384 digest"), NULL },
 { "blake-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BLAKE_512,
	N_("Blake-512 digest"), NULL },
 { "blake2b",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BLAKE2B,
	N_("Blake2b digest"), NULL },
 { "blake2bp",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BLAKE2BP,
	N_("Blake2bp digest"), NULL },
 { "blake2s",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BLAKE2S,
	N_("Blake2s digest"), NULL },
 { "blake2sp",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BLAKE2SP,
	N_("Blake2sp digest"), NULL },
 { "bmw-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BMW_224,
	N_("BlueMidnightWish-224 digest"), NULL },
 { "bmw-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BMW_256,
	N_("BlueMidnightWish-256 digest"), NULL },
 { "bmw-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BMW_384,
	N_("BlueMidnightWish-384 digest"), NULL },
 { "bmw-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_BMW_512,
	N_("BlueMidnightWish-512 digest"), NULL },
 { "chi-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_CHI_224,
	N_("Chi-224 digest"), NULL },
 { "chi-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_CHI_256,
	N_("Chi-256 digest"), NULL },
 { "chi-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_CHI_384,
	N_("Chi-384 digest"), NULL },
 { "chi-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_CHI_512,
	N_("Chi-512 digest"), NULL },
 { "cubehash-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_CUBEHASH_224,
	N_("CubeHash-224 digest"), NULL },
 { "cubehash-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_CUBEHASH_256,
	N_("CubeHash-256 digest"), NULL },
 { "cubehash-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_CUBEHASH_384,
	N_("CubeHash-384 digest"), NULL },
 { "cubehash-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_CUBEHASH_512,
	N_("CubeHash-512 digest"), NULL },
 { "echo-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_ECHO_224,
	N_("Echo-224 digest"), NULL },
 { "echo-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_ECHO_256,
	N_("Echo-256 digest"), NULL },
 { "echo-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_ECHO_384,
	N_("Echo-384 digest"), NULL },
 { "echo-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_ECHO_512,
	N_("Echo-512 digest"), NULL },
 { "edonr-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_EDONR_224,
	N_("EdonR-224 digest"), NULL },
 { "edonr-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_EDONR_256,
	N_("EdonR-256 digest"), NULL },
 { "edonr-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_EDONR_384,
	N_("EdonR-384 digest"), NULL },
 { "edonr-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_EDONR_512,
	N_("EdonR-512 digest"), NULL },
 { "fugue-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_FUGUE_224,
	N_("Fugue-224 digest"), NULL },
 { "fugue-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_FUGUE_256,
	N_("Fugue-256 digest"), NULL },
 { "fugue-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_FUGUE_384,
	N_("Fugue-384 digest"), NULL },
 { "fugue-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_FUGUE_512,
	N_("Fugue-512 digest"), NULL },
 { "groestl-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_GROESTL_224,
	N_("Groestl-224 digest"), NULL },
 { "groestl-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_GROESTL_256,
	N_("Groestl-256 digest"), NULL },
 { "groestl-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_GROESTL_384,
	N_("Groestl-384 digest"), NULL },
 { "groestl-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_GROESTL_512,
	N_("Groestl-512 digest"), NULL },
 { "hamsi-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_HAMSI_224,
	N_("Hamsi-224 digest"), NULL },
 { "hamsi-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_HAMSI_256,
	N_("Hamsi-256 digest"), NULL },
 { "hamsi-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_HAMSI_384,
	N_("Hamsi-384 digest"), NULL },
 { "hamsi-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_HAMSI_512,
	N_("Hamsi-512 digest"), NULL },
 { "jh-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_JH_224,
	N_("JH-224 digest"), NULL },
 { "jh-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_JH_256,
	N_("JH-256 digest"), NULL },
 { "jh-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_JH_384,
	N_("JH-384 digest"), NULL },
 { "jh-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_JH_512,
	N_("JH-512 digest"), NULL },
 { "keccak-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_KECCAK_224,
	N_("Keccak-224 digest"), NULL },
 { "keccak-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_KECCAK_256,
	N_("Keccak-256 digest"), NULL },
 { "keccak-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_KECCAK_384,
	N_("Keccak-384 digest"), NULL },
 { "keccak-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_KECCAK_512,
	N_("Keccak-512 digest"), NULL },
 { "lane-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_LANE_224,
	N_("Lane-224 digest"), NULL },
 { "lane-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_LANE_256,
	N_("Lane-256 digest"), NULL },
 { "lane-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_LANE_384,
	N_("Lane-384 digest"), NULL },
 { "lane-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_LANE_512,
	N_("Lane-512 digest"), NULL },
 { "luffa-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_LUFFA_224,
	N_("Luffa-224 digest"), NULL },
 { "luffa-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_LUFFA_256,
	N_("Luffa-256 digest"), NULL },
 { "luffa-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_LUFFA_384,
	N_("Luffa-384 digest"), NULL },
 { "luffa-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_LUFFA_512,
	N_("Luffa-512 digest"), NULL },
 { "md6-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_MD6_224,
	N_("MD6-224 digest"), NULL },
 { "md6-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_MD6_256,
	N_("MD6-256 digest"), NULL },
 { "md6-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_MD6_384,
	N_("MD6-384 digest"), NULL },
 { "md6-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_MD6_512,
	N_("MD6-512digest"), NULL },
 { "rg32-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_RG32_256,
	N_("RadioGatun32-256 digest"), NULL },
 { "rg64-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_RG64_256,
	N_("RadioGatun64-256 digest"), NULL },
 { "shabal-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHABAL_224,
	N_("Shabal-224 digest"), NULL },
 { "shabal-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHABAL_256,
	N_("Shabal-256 digest"), NULL },
 { "shabal-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHABAL_384,
	N_("Shabal-384 digest"), NULL },
 { "shabal-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHABAL_512,
	N_("Shabal-512 digest"), NULL },
 { "shavite3-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHAVITE3_224,
	N_("SHAvite3-224 digest"), NULL },
 { "shavite3-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHAVITE3_256,
	N_("SHAvite3-256 digest"), NULL },
 { "shavite3-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHAVITE3_384,
	N_("SHAvite3-384 digest"), NULL },
 { "shavite3-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SHAVITE3_512,
	N_("SHAvite3-512 digest"), NULL },
 { "simd-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SIMD_224,
	N_("SIMD-224 digest"), NULL },
 { "simd-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SIMD_256,
	N_("SIMD-256 digest"), NULL },
 { "simd-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SIMD_384,
	N_("SIMD-384 digest"), NULL },
 { "simd-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_SIMD_512,
	N_("SIMD-512 digest"), NULL },
 { "tib3-224",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_TIB3_224,
	N_("TIB3-224 digest"), NULL },
 { "tib3-256",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_TIB3_256,
	N_("TIB3-256 digest"), NULL },
 { "tib3-384",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_TIB3_384,
	N_("TIB3-384 digest"), NULL },
 { "tib3-512",'\0', POPT_ARG_VAL, &rpmioDigestHashAlgo, PGPHASHALGO_TIB3_512,
	N_("TIB3-512 digest"), NULL },
 { "salsa10",'\0', POPT_ARG_VAL,&rpmioDigestHashAlgo, PGPHASHALGO_SALSA10,
	N_("SALSA-10 hash"), NULL },
 { "salsa20",'\0', POPT_ARG_VAL,&rpmioDigestHashAlgo, PGPHASHALGO_SALSA20,
	N_("SALSA-20 hash"), NULL },
 { "rmd128",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_RIPEMD128,
	N_("RIPEMD-128 digest"), NULL },
 { "rmd160",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_RIPEMD160,
	N_("RIPEMD-160 digest"), NULL },
 { "rmd256",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_RIPEMD256,
	N_("RIPEMD-256 digest"), NULL },
 { "rmd320",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_RIPEMD320,
	N_("RIPEMD-320 digest"), NULL },
 { "tiger",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_TIGER192,
	N_("TIGER digest"), NULL },
 { "crc32",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_CRC32,
	N_("CRC-32 checksum"), NULL },
 { "crc64",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_CRC64,
	N_("CRC-64 checksum"), NULL },
 { "adler32",'\0', POPT_ARG_VAL,&rpmioDigestHashAlgo, PGPHASHALGO_ADLER32,
	N_("ADLER-32 checksum"), NULL },
 { "jlu32",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_JLU32,
	N_("Lookup3 hash"), NULL },
 { "xxh32",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_XXH32,
	N_("XXH-32 hash"), NULL },
 { "xxh64",'\0', POPT_ARG_VAL,	&rpmioDigestHashAlgo, PGPHASHALGO_XXH64,
	N_("XXH-64 hash"), NULL },
 { "nodigest",'\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &rpmioDigestHashAlgo, PGPHASHALGO_NONE,
	N_("No hash algorithm"), NULL },
 { "alldigests",'\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &rpmioDigestHashAlgo, 256,
	N_("All hash algorithm(s)"), NULL },
    POPT_TABLEEND
};

/**
 * Display rpm version.
 */
static void printVersion(FILE * fp)
{
    fprintf(fp, _("%s (" RPM_NAME ") %s\n"), __progname, rpmioEVR);
}

void rpmioConfigured(void)
{

    if (rpmioInitialized < 0) {
	/* XXX TODO: add initialization side-effects. */
	rpmioInitialized = 0;
    }
    if (rpmioInitialized)
	exit(EXIT_FAILURE);
}

void rpmioInitInfo(int lvl, const char * msg)
{
    char * t;

    t = rpmExpand("%now ", __progname, " (" RPM_NAME "-", rpmioEVR, ")", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("progname: %progname\t pid: %pid\t tid: %tid", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("\t\t\tpgrp: %pgrp\tpgid: %pgid", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("\t\t\tppid: %ppid\t sid: %sid", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("   umask:  %umask", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("    time:  %{now:%%T}", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("    date:  %{now:%%F}", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("    host: %host", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("  ipaddr: %ipaddr", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("    user: %user\tuid:%uid euid:%euid suid:%suid fsuid:%fsuid", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("   group: %group\tgid:%gid egid:%egid sgid:%sgid fsgid:%fsgid", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);
    t = rpmExpand("priority: %priority\t\tpersona:%personality", NULL);
	rpmlog(lvl, "%s\n", t);	t = _free(t);

    rpmcapDump(lvl, msg);
}

static struct rpmop_s rpmio_totalops;

int rpmioInitUsage(int lvl, const char *msg)
{
    rpmop op = &rpmio_totalops;
    int rc = 0;
    rpmswEnter(op, -1);
    return rc;
}

#if defined(HAVE_GETRUSAGE)
static void rpmioPrintUsage(int lvl, const char * msg, struct rusage *ru)
{
    struct timeval *tv;

    if (msg)
	rpmlog(lvl, "---------- %s\n", msg);

#define	TV_MSEC	tv_usec/1000
    tv = &ru->ru_utime;
    rpmlog(lvl, "%9s:%6ld.%02ld\t\t%s CPU time used\n",
		"utime",
		tv->tv_sec, tv->TV_MSEC/10,
		"user");
    tv = &ru->ru_stime;
    rpmlog(lvl, "%9s:%6ld.%02ld\t\t%s CPU time used\n",
		"stime",
		tv->tv_sec, tv->TV_MSEC/10,
		"system");
#undef	TV_MSEC

#define	RULOG(_v, _desc) \
    if (ru->ru_##_v) rpmlog(lvl, "%9s:%9lu\t\t%s\n", #_v, ru->ru_##_v, _desc)
    RULOG(maxrss,   "maximum resident set size");
    RULOG(ixrss,    "intergral shared memory size");
    RULOG(idrss,    "intergral unshared data size");
    RULOG(isrss,    "integral unshared stack size");
    RULOG(minflt,   "page reclaims (soft page faults)");
    RULOG(majflt,   "page faults (hard page faults)");
    RULOG(nswap,    "swaps");
    RULOG(inblock,  "block input operations");
    RULOG(oublock,  "block output operations");
    RULOG(msgsnd,   "IPC messages sent");
    RULOG(msgrcv,   "IPC messages received");
    RULOG(nsignals, "signals received");
    RULOG(nvcsw,    "voluntary context switches");
    RULOG(nivcsw,   "involuntary context switches");
#undef	RULOG

}
#endif

int rpmioFiniUsage(int lvl, const char * msg)
{
    rpmop op = &rpmio_totalops;
    rpmtime_t usecs = rpmswExit(op, -2);
    struct timeval _tv = {
	usecs / (1000*1000),
	usecs % (1000*1000)
    }, *tv = &_tv;
    int rc = 0;

    /* Display on stdout with --stats. */
    if (_rpmsw_stats && lvl > RPMLOG_NOTICE)
	lvl = RPMLOG_NOTICE;

    /* Display elapsed time. */
    if (tv->tv_sec >= 3600)
	_rpmlog(lvl, "%9s:%3ld:%02ld:%02ld\t\t%s time\n",
		"elapsed",
		(tv->tv_sec / 3600),
		(tv->tv_sec % 3600) / 60,
		(tv->tv_sec %   60),
		"elapsed");
    else
	_rpmlog(lvl, "%9s:%3ld:%02ld.%02ld\t\t%s time\n",
		"elapsed",
		(tv->tv_sec / 60),
		(tv->tv_sec % 60),
		(tv->tv_usec/1000)/10,
		"elapsed");

    /* Display resource usage. */
#if defined(HAVE_GETRUSAGE)
    struct rusage _ru, *ru = &_ru;
    rc = getrusage(RUSAGE_SELF, ru);
    if (!rc)
	rpmioPrintUsage(lvl, "getrusage(self)", ru);
    rc = getrusage(RUSAGE_CHILDREN, ru);
    if (!rc)
	rpmioPrintUsage(lvl, "getrusage(children)", ru);
    rc = getrusage(RUSAGE_THREAD, ru);
    if (!rc)
	rpmioPrintUsage(lvl, "getrusage(thread)", ru);
#endif

    return rc;
}

void rpmioFiniInfo(int lvl, const char * msg)
{
    (void) rpmioFiniUsage(lvl, NULL);
    rpmioMallocInfo(NULL);

    if (msg)
	rpmlog(lvl, "%s exit.\n", msg);
}

#ifdef	DYING
__attribute__((constructor(101)))
static void rpmioCtor(void)
{
    fprintf(stderr, "--> %s at %s:%u\n", __FUNCTION__, __FILE__, __LINE__);
    const char * cmdline = rpmGetPath("/proc/%pid/cmdline", NULL);
    fprintf(stderr, "\tcmdline: %s\n", cmdline);
    cmdline = _free(cmdline);
}

__attribute__((destructor(101)))
static void rpmioDtor(void)
{
    fprintf(stderr, "<-- %s at %s:%u\n", __FUNCTION__, __FILE__, __LINE__);
}

void rpmioMain(int argc, char **argv, char **env);
__attribute__((section(".init_array")))
static void *rpmioPreInit = &rpmioMain;

void rpmioMain(int argc, char **argv, char **env)
{
    (void)rpmioPreInit;
#ifdef	NOTYET
    fprintf(stderr, "--> %s(%d,%p,%p) %s\n", __func__, argc, argv, env, *argv);
#else
    fprintf(stderr, "--> %s at %s:%u\n", __FUNCTION__, __FILE__, __LINE__);
#endif	/* NOTYET */
}
#endif	/* DYING */

/**
 */
static void rpmioAllArgCallback(poptContext con,
                enum poptCallbackReason reason,
                const struct poptOption * opt, const char * arg,
                const void * data)
{

    /* XXX avoid accidental collisions with POPT_BIT_SET for flags */
    if (opt->arg == NULL)
    switch (opt->val) {
    case 'q':
	rpmSetVerbosity(RPMLOG_WARNING);
	break;
    case 'v':
	rpmIncreaseVerbosity();
	break;
#ifdef	NOTYET
    case 'D':
    {	char *s, *t;
	/* XXX Convert '-' in macro name to underscore, skip leading %. */
	s = t = xstrdup(arg);
	while (*t && !xisspace((int)*t)) {
	    if (*t == '-') *t = '_';
	    t++;
	}
	t = s;
	if (*t == '%') t++;
	rpmioConfigured();
	/* XXX adding macro to global context isn't Right Thing Todo. */
	(void) rpmDefineMacro(NULL, t, RMIL_CMDLINE);
	(void) rpmDefineMacro(rpmCLIMacroContext, t, RMIL_CMDLINE);
	s = _free(s);
    }	break;
    case POPT_UNDEFINE:
    {	char *s, *t;
	/* XXX Convert '-' in macro name to underscore, skip leading %. */
	s = t = xstrdup(arg);
	while (*t && !xisspace((int)*t)) {
	    if (*t == '-') *t = '_';
	    t++;
	}
	t = s;
	if (*t == '%') t++;
	rpmioConfigured();
	(void) rpmUndefineMacro(NULL, t);
	(void) rpmUndefineMacro(rpmCLIMacroContext, t);
	s = _free(s);
    }	break;
    case 'E':
	rpmioConfigured();
	{   const char *val = rpmExpand(arg, NULL);
            size_t val_len;
            val_len = strlen(val);
            if (val[val_len - 1] == '\n')
                fwrite(val, val_len, 1, stdout);
            else
		fprintf(stdout, "%s\n", val);
	    val = _free(val);
	}
	break;
#endif	/* NOTYET */
    case POPT_CRYPTO:
	{   const char *val;
#ifdef	NOTYET
	    rpmioConfigured();
	    val = rpmExpand(arg, NULL);
#else
	    val = xstrdup(arg);
#endif	/* NOTYET */
	    if (!xstrcasecmp(val, "beecrypt") || !xstrcasecmp(val, "bc")) {
#if defined(WITH_BEECRYPT)
		pgpImplVecs = &rpmbcImplVecs;
#else
                rpmlog(RPMLOG_ERR, "BeeCrypt (\"beecrypt\") based cryptography implementation not available\n");
                exit(EXIT_FAILURE);
#endif
            }
	    else if (!xstrcasecmp(val, "gcrypt") || !xstrcasecmp(val, "gc")) {
#if defined(WITH_GCRYPT)
		pgpImplVecs = &rpmgcImplVecs;
#else
                rpmlog(RPMLOG_ERR, "GCrypt (\"gcrypt\") based cryptography implementation not available\n");
                exit(EXIT_FAILURE);
#endif
            }
	    else if (!xstrcasecmp(val, "tomcrypt") || !xstrcasecmp(val, "ltc")) {
#if defined(WITH_TOMCRYPT)
		pgpImplVecs = &rpmltcImplVecs;
#else
                rpmlog(RPMLOG_ERR, "TomCrypt (\"tomcrypt\") based cryptography implementation not available\n");
                exit(EXIT_FAILURE);
#endif
            }
	    else if (!xstrcasecmp(val, "NSS")) {
#if defined(WITH_NSS)
		pgpImplVecs = &rpmnssImplVecs;
#else
                rpmlog(RPMLOG_ERR, "Mozilla NSS (\"nss\") based cryptography implementation not available\n");
                exit(EXIT_FAILURE);
#endif
            }
	    else if (!xstrcasecmp(val, "OpenSSL") || !xstrcasecmp(val, "ssl")) {
#if defined(WITH_SSL)
		pgpImplVecs = &rpmsslImplVecs;
#else
                rpmlog(RPMLOG_ERR, "OpenSSL (\"openssl\") based cryptography implementation not available\n");
                exit(EXIT_FAILURE);
#endif
            }
	    else if (!xstrcasecmp(val, "CDSA")) {
#if defined(WITH_CDSA)
		pgpImplVecs = &rpmcdsaImplVecs;
#else
                rpmlog(RPMLOG_ERR, "CDSA (\"cdsa\") based cryptography implementation not available\n");
                exit(EXIT_FAILURE);
#endif
            }
            else {
                rpmlog(RPMLOG_ERR, "cryptography implementation \"%s\" not known\n", val);
                exit(EXIT_FAILURE);
            }
	    val = _free(val);
	}
	break;
    case POPT_SHOWVERSION:
	printVersion(stdout);
	con = rpmioFini(con);
	exit(EXIT_SUCCESS);
	break;
    }
}

unsigned rpmioOpenFlags = 0;
struct poptOption rpmioOpenFlagsTable[] = {
#ifdef	O_APPEND
 { "append", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_APPEND,
	N_("append"), N_("O_APPEND") },
#endif
#ifdef	O_ASYNC
 { "async", '\0', POPT_BIT_SET,		&rpmioOpenFlags, O_ASYNC,
	N_("async"), N_("O_ASYNC") },
#endif
#ifdef	O_CLOEXEC
 { "cloexec", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_CLOEXEC,
	N_("cloexec"), N_("O_CLOEXEC") },
#endif
#ifdef	O_CREAT
 { "creat", '\0', POPT_BIT_SET,		&rpmioOpenFlags, O_CREAT,
	N_("creat"), N_("O_CREAT") },
#endif
#ifdef	O_DIRECT
 { "direct", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_DIRECT,
	N_("direct"), N_("O_DIRECT") },
#endif

#ifdef	O_DIRECTORY
 { "directory", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_DIRECTORY,
	N_("directory"), N_("O_DIRECTORY") },
#endif

#ifdef	O_DSYNC
 { "dsync", '\0', POPT_BIT_SET,		&rpmioOpenFlags, O_DSYNC,
	N_("dsync"), N_("O_DSYNC") },
#endif
#ifdef	O_EXCL
 { "excl", '\0', POPT_BIT_SET,		&rpmioOpenFlags, O_EXCL,
	N_("excl"), N_("O_EXCL") },
#endif
#ifdef	O_LARGEFILE
 { "largefile", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_LARGEFILE,
	N_("largefile"), N_("O_LARGEFILE") },
#endif

#ifdef	O_NOATIME
 { "noatime", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_NOATIME,
	N_("noatime"), N_("O_NOATIME") },
#endif
#ifdef	O_NOCTTY
 { "noctty", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_NOCTTY,
	N_("noctty"), N_("O_NOCTTY") },
#endif
#ifdef	O_NOFOLLOW
 { "nofollow", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_NOFOLLOW,
	N_("nofollow"), N_("O_NOFOLLOW") },
#endif
#ifdef	O_NONBLOCK
 { "nonblock", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_NONBLOCK,
	N_("nonblock"), N_("O_NONBLOCK") },
#endif
#ifdef	O_NDELAY
 { "ndelay", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_NDELAY,
	N_("ndelay"), N_("O_NDELAY") },
#endif

#ifdef	O_PATH
 { "path", '\0', POPT_BIT_SET,		&rpmioOpenFlags, O_PATH,
	N_("path"), N_("O_PATH") },
#endif

#ifdef	NOTYET		/* XXX O_ACCMODE needed */
#ifdef	O_RDONLY
 { "rdonly", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_RDONLY,
	N_("rdonly"), N_("O_RDONLY") },
#endif
#ifdef	O_RDWR
 { "rdwr", '\0', POPT_BIT_SET,		&rpmioOpenFlags, O_RDWR,
	N_("rdwr"), N_("O_RDWR") },
#endif
#endif			/* XXX O_ACCMODE needed */

#ifdef	O_SYNC
 { "sync", '\0', POPT_BIT_SET,		&rpmioOpenFlags, O_SYNC,
	N_("sync"), N_("O_SYNC") },
#endif

#ifdef	O_TMPFILE
 { "tmpfile", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_TMPFILE,
	N_("tmpfile"), N_("O_TMPFILE") },
#endif

#ifdef	O_TRUNC
 { "trunc", '\0', POPT_BIT_SET,		&rpmioOpenFlags, O_TRUNC,
	N_("trunc"), N_("O_TRUNC") },
#endif

#ifdef	NOTYET		/* XXX O_ACCMODE needed */
#ifdef	O_WRONLY
 { "wronly", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_WRONLY,
	N_("wronly"), N_("O_WRONLY") },
#endif
#endif			/* XXX O_ACCMODE needed */

#ifdef	O_SHLOCK
 { "shlock", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_SHLOCK,
	N_("shlock"), N_("O_SHLOCK") },
#endif
#ifdef	O_EXLOCK
 { "exlock", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_EXLOCK,
	N_("exlock"), N_("O_EXLOCK") },
#endif
#ifdef	O_SYMLINK
 { "symlink", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_SYMLINK,
	N_("symlink"), N_("O_SYMLINK") },
#endif
#ifdef	O_EVTONLY
 { "evtonly", '\0', POPT_BIT_SET,	&rpmioOpenFlags, O_EVTONLY,
	N_("evtonly"), N_("O_EVTONLY") },
#endif
   POPT_TABLEEND
};

char * rpmioB2N(struct poptOption * optTable, unsigned long optFlags)
{
    char t[BUFSIZ];
    char * te = t;
    struct poptOption * opt;

#define	POPT_MASK   (POPT_ARGFLAG_LOGICALOPS|POPT_ARGFLAG_NOT|POPT_ARG_MASK)
    for (opt=optTable; opt->longName || opt->shortName || opt->arg; opt++) {
	int argInfo = opt->argInfo;

	switch (argInfo & POPT_MASK) {
	case POPT_BIT_SET:
	case POPT_BIT_CLR:
	    if (!(optFlags & opt->val))	/* XXX limited to 32 bits */
		continue;
	    break;
	default:
	    continue;
	    break;
	}
	if (te > t)
	    *te++ = '|';
	if (opt->argDescrip)
	    te = stpcpy(te, opt->argDescrip);
	else
	    te += sprintf(te, "0x%x", opt->val);
	optFlags &= ~opt->val;		/* XXX first found only. */
    }
#undef	POPT_MASK
    if (te == t)
	te = stpcpy(te, "0");
    *te = '\0';
    return xstrdup(t);
}

unsigned long rpmioN2B(struct poptOption * optTable, const char * options)
{
    ARGV_t av = NULL;
    int xx = argvSplit(&av, options, ",");
    int ac = argvCount(av);
    struct poptOption * opt;
    unsigned long optFlags = 0;
    (void)xx;

    for (int i = 0; i < ac; i++) {
	const char * o = av[i];
	int negate = 0;

	/* Handle !foo and nofoo negated options. */
	if (*o == '!') {
	    negate ^= 1;
	    o++;
	}
	if (strchr("Nn", o[0]) && strchr("Oo", o[1])) {
	    negate ^= 1;
	    o += 2;
	}
	if (!(o && *o))
	    continue;

	for (opt=optTable; opt->longName || opt->shortName || opt->arg; opt++) {

	    if (opt->longName == NULL)
		continue;

	    int argInfo = opt->argInfo;
	    if (argInfo & POPT_ARGFLAG_TOGGLE) {
		/* Is this the right bit (stripping optional prefix)? */
		if (strcasecmp(o, opt->longName))
		    continue;

		/* Toggle the value (if requested). */
		if (negate) {
		    argInfo &= ~POPT_ARGFLAG_TOGGLE;
		    if (argInfo == POPT_BIT_SET)
			argInfo = POPT_BIT_CLR;
		    else if (argInfo == POPT_BIT_CLR)
			argInfo = POPT_BIT_SET;
		    argInfo |= POPT_ARGFLAG_TOGGLE;
		}
	    } else {
		/* Is this the right bit (using full option name)? */
		if (strcasecmp(av[i], opt->longName))
		    continue;
	    }

	    switch (argInfo & POPT_ARG_MASK) {
	    case POPT_ARG_VAL:
		(void) poptSaveInt((int *)&optFlags, argInfo, (long)opt->val);
		break;
	    default:
		break;
	    }
	    o = NULL;		/* XXX option processed. */
	    break;
	}
	if (o)
	    fprintf(stderr, "*** unknown option \"%s\" was ignored\n", o);
    }
    av = argvFree(av);
    return optFlags;
}

unsigned rpmioFtsFlags = 0;
struct poptOption rpmioFtsFlagsTable[] = {
 { "comfollow", '\0', POPT_BIT_SET,	&rpmioFtsFlags, FTS_COMFOLLOW,
	N_("Follow command line symlinks"), N_("FTS_COMFOLLOW") },
 { "logical", '\0', POPT_BIT_SET,	&rpmioFtsFlags, FTS_LOGICAL,
	N_("Logical walk"), N_("FTS_LOGICAL") },
 { "nochdir", '\0', POPT_BIT_SET,	&rpmioFtsFlags, FTS_NOCHDIR,
	N_("Don't change directories"), N_("FTS_NOCHDIR") },
 { "nostat", '\0', POPT_BIT_SET,	&rpmioFtsFlags, FTS_NOSTAT,
	N_("Don't get stat info"), N_("FTS_NOSTAT") },
 { "physical", '\0', POPT_BIT_SET,	&rpmioFtsFlags, FTS_PHYSICAL,
	N_("Physical walk"), N_("FTS_PHYSICAL") },
 { "seedot", '\0', POPT_BIT_SET,	&rpmioFtsFlags, FTS_SEEDOT,
	N_("Return dot and dot-dot"), N_("FTS_SEEDOT") },
 { "xdev", '\0', POPT_BIT_SET,		&rpmioFtsFlags, FTS_XDEV,
	N_("Don't cross devices"), N_("FTS_XDEV") },
 { "whiteout", '\0', POPT_BIT_SET,	&rpmioFtsFlags, FTS_WHITEOUT,
	N_("Return whiteout information"), N_("FTS_WHITEOUT") },
   POPT_TABLEEND
};

struct poptOption rpmioAllPoptTable[] = {
 { NULL, '\0', POPT_ARG_CALLBACK | POPT_CBFLAG_INC_DATA | POPT_CBFLAG_CONTINUE,
        (void *)rpmioAllArgCallback, 0, NULL, NULL },

 { "debug", 'd', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &__debug, -1,
	N_("Debug generic operations"), NULL },

#ifdef	NOTYET
 { "define", 'D', POPT_ARG_STRING, NULL, (int)'D',
	N_("Define MACRO with value EXPR"),
	N_("'MACRO EXPR'") },
 { "undefine", '\0', POPT_ARG_STRING, NULL, POPT_UNDEFINE,
	N_("Undefine MACRO"),
	N_("'MACRO'") },
 { "eval", 'E', POPT_ARG_STRING, NULL, (int)'E',
	N_("Print macro expansion of EXPR"),
	N_("'EXPR'") },

 { "macros", '\0', POPT_ARG_STRING, &rpmMacrofiles, 0,
	N_("Read <FILE:...> instead of default file(s)"),
	N_("<FILE:...>") },
#ifdef WITH_LUA
 { "rpmlua", '\0', POPT_ARG_STRING, &rpmluaFiles, 0,
	N_("Read <FILE:...> instead of default RPM Lua file(s)"),
	N_("<FILE:...>") },
#endif
 { "rpmpopt", '\0', POPT_ARG_STRING, NULL, 0,
	N_("Read <FILE:...> instead of default POPT file(s)"),
	N_("<FILE:...>") },
#endif	/* NOTYET */

 { "pipe", '\0', POPT_ARG_STRING|POPT_ARGFLAG_DOC_HIDDEN, &rpmioPipeOutput, 0,
	N_("Send stdout to CMD"),
	N_("CMD") },
 { "root", 'r', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &rpmioRootDir, 0,
	N_("Use ROOT as top level directory"),
	N_("ROOT") },

 { "quiet", '\0', 0, NULL, (int)'q',
	N_("Provide less detailed output"), NULL},
 { "verbose", 'v', 0, NULL, (int)'v',
	N_("Provide more detailed output"), NULL},
 { "version", '\0', 0, NULL, POPT_SHOWVERSION,
	N_("Print the version"), NULL },

#if defined(HAVE_LIBIO_H) && defined(_G_IO_IO_FILE_VERSION)
 { "nolibio", '\0', POPT_ARG_VAL|POPT_ARGFLAG_TOGGLE|POPT_ARGFLAG_DOC_HIDDEN, &noLibio, -1,
	N_("Disable use of libio(3) API"), NULL},
#endif

 { "usecrypto",'\0', POPT_ARG_STRING|POPT_ARGFLAG_DOC_HIDDEN, NULL, POPT_CRYPTO,
        N_("Select cryptography implementation"),
	N_("CRYPTO") },

 { "ardebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_ar_debug, -1,
	N_("Debug ar archives"), NULL},
 { "avdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_av_debug, -1,
	N_("Debug argv collections"), NULL},
 { "cpiodebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_cpio_debug, -1,
	N_("Debug cpio archives"), NULL},
 { "ctxdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmctx_debug, -1,
	N_("Debug digest contexts"), NULL},
 { "davdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_dav_debug, -1,
	N_("Debug WebDAV data stream"), NULL},
 { "davcbdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_dav_cb_debug, -1,
	N_("Debug NEON callbacks"), NULL},
 { "davhookdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_dav_hook_debug, -1,
	N_("Debug NEON hooks"), NULL},
 { "ftpdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_ftp_debug, -1,
	N_("Debug FTP/HTTP data stream"), NULL},
 { "ftsdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_fts_debug, -1,
	N_("Debug Fts(3) traverse"), NULL},
 { "htdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmht_debug, -1,
	N_("Debug hash tables"), NULL},
 { "htmldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_html_debug, -1,
	N_("Debug HTML parsing"), NULL},
 { "iosmdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_iosm_debug, -1,
	N_("Debug I/O state machine"), NULL},
 { "miredebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmmire_debug, -1,
	N_("Debug miRE patterns"), NULL},
 { "pgpdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_pgp_debug, -1,
	N_("Debug PGP usage"), NULL},
 { "pgpprint", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_pgp_print, -1,
	N_("Print PGP keys"), NULL},
 { "rpmiodebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmio_debug, -1,
	N_("Debug rpmio I/O"), NULL},
 { "rpmiobdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmiob_debug, -1,
	N_("Debug rpmio I/O buffers"), NULL},

 { "rpmaiodebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmaio_debug, -1,
	N_("Debug AIO wrapper"), NULL},
 { "rpmasndebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmasn_debug, -1,
	N_("Debug embedded ASN.1 interpreter"), NULL},
 { "rpmbagdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmbag_debug, -1,
	N_("Debug depsolver wrappers "), NULL},
 { "rpmcvsdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmcvs_debug, -1,
	N_("Debug CVS wrappers"), NULL},
 { "rpmgitdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmgit_debug, -1,
	N_("Debug GIT wrappers"), NULL},
 { "rpmsetdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmset_debug, -1,
	N_("Debug SET-VERSION wrappers"), NULL},
 { "rpmsvndebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmsvn_debug, -1,
	N_("Debug Subversion wrappers"), NULL},
 { "rpmtpmdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmtpm_debug, -1,
	N_("Debug TPM emulator"), NULL},

 { "rpmaugdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmaug_debug, -1,
	N_("Debug embedded Augeas interpreter"), NULL},
 { "rpmbfdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmbf_debug, -1,
	N_("Debug Bloom filters"), NULL},
 { "rpmcapdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmcap_debug, -1,
	N_("Debug capabilities"), NULL},
 { "rpmctdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmct_debug, -1,
	N_("Debug cp(1)"), NULL},
 { "rpmseddebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmsed_debug, -1,
	N_("Debug sed(1)"), NULL},
 { "rpmcudfdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmcudf_debug, -1,
	N_("Debug embedded CUDF parser"), NULL},
 { "rpmeiodebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmeio_debug, -1,
	N_("Debug libeio wrappers"), NULL},
 { "rpmevdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmev_debug, -1,
	N_("Debug libev wrappers"), NULL},
 { "rpmficldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmficl_debug, -1,
	N_("Debug embedded FICL interpreter"), NULL},
 { "rpmgfsdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmgfs_debug, -1,
	N_("Debug Mongo DB gridfs connection"), NULL},
 { "rpmhkpdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmhkp_debug, -1,
	N_("Debug hkp:// keyring"), NULL},
 { "rpmjssdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmjss_debug, -1,
	N_("Debug embedded JavaScript interpreter"), NULL},
#ifdef WITH_LUA
 { "rpmluadebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmlua_debug, -1,
	N_("Debug embedded Lua interpreter"), NULL},
#endif
 { "rpmmgdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmmg_debug, -1,
	N_("Debug rpmmg magic"), NULL},
 { "rpmmqttdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmmqtt_debug, -1,
	N_("Debug rpmmqtt magic"), NULL},
 { "rpmmrbdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmmrb_debug, -1,
	N_("Debug embedded MRuby interpreter"), NULL},
 { "rpmmsqdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmmsq_debug, -1,
	N_("Debug message queues"), NULL},
 { "nixdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmnix_debug, -1,
	N_("Debug embedded Nix interpreter"), NULL},
 { "odbcdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_odbc_debug, -1,
	N_("Debug embedded ODBC interface"), NULL},
 { "rpmperldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmperl_debug, -1,
	N_("Debug embedded Perl interpreter"), NULL},
 { "rpmpythondebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmpython_debug, -1,
	N_("Debug embedded Python interpreter"), NULL},
 { "rpmrubydebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmruby_debug, -1,
	N_("Debug embedded Ruby interpreter"), NULL},
 { "rpmsqdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmsq_debug, -1,
	N_("Debug rpmsq Signal Queue"), NULL},
 { "rpmsqldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmsql_debug, -1,
	N_("Debug embedded SQL interpreter"), NULL},
 { "rpmsquirreldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmsquirrel_debug, -1,
	N_("Debug embedded SQUIRREL interpreter"), NULL},
 { "rpmsmdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmsm_debug, 1,
	N_("Debug semanage"), NULL},
 { "rpmspdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmsp_debug, 1,
	N_("Debug sepol"), NULL},
 { "rpmsxdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmsx_debug, 1,
	N_("Debug selinux"), NULL},
 { "rpmtcldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmtcl_debug, -1,
	N_("Debug embedded TCL interpreter"), NULL},
 { "rpmwdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmw_debug, -1,
	N_("Debug wrapper generator"), NULL},
 { "rpmvcdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmvc_debug, -1,
	N_("Debug embedded SQL virtual cursor"), NULL},
 { "rpmvtdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmvt_debug, -1,
	N_("Debug embedded SQL virtual table"), NULL},
#ifdef WITH_BZIP2
 { "rpmzqdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmzq_debug, -1,
	N_("Debug rpmzq Job Queuing"), NULL},
#endif
 { "rpmzstddebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmzstd_debug, -1,
	N_("Debug ZSTD wrapper"), NULL},
 { "xardebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_xar_debug, -1,
	N_("Debug xar archives"), NULL},
 { "tardebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_tar_debug, -1,
	N_("Debug tar archives"), NULL},
 { "stats", '\0', POPT_ARG_VAL,				&_rpmsw_stats, -1,
	N_("Display operation statistics"), NULL},
 { "urldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmurl_debug, -1,
	N_("Debug URL cache handling"), NULL},

   POPT_TABLEEND
};

poptContext
rpmioFini(poptContext optCon)
{
    /* XXX this should be done in the rpmioClean() wrapper. */
    /* keeps memory leak checkers quiet */
    rpmFreeMacros(NULL);
    rpmFreeMacros(rpmCLIMacroContext);

    rpmioClean();

    optCon = poptFreeContext(optCon);

    rpmioMuntrace();

    /* Spew time and usage information. */
    rpmioFiniInfo(RPMLOG_DEBUG, __FUNCTION__);

    return NULL;
}

static inline int checkfd(const char * devnull, int fdno, int flags)
{
    struct stat sb;
    int ret = 0;

    if (fstat(fdno, &sb) == -1 && errno == EBADF)
	ret = (open(devnull, flags) == fdno) ? 1 : 2;
    return ret;
}

poptContext
rpmioInit(int argc, char *const argv[], struct poptOption * optionsTable)
{
    poptContext optCon;
    char * arg0 = argv[0];
    int rc;

    rpmioMtrace();
#if defined(HAVE_FTOK)
    __progname_key = ftok(arg0, __progname_projid);
#endif
    rpmioInitUsage(RPMLOG_DEBUG, __FUNCTION__);

    setprogname(arg0);       /* Retrofit glibc __progname */

    /* XXX glibc churn sanity */
    if (__progname == NULL) {
	if ((__progname = strrchr(arg0, '/')) != NULL) __progname++;
	else __progname = arg0;
    }

    /* Insure that stdin/stdout/stderr are open, lest stderr end up in rpmdb. */
   {	static const char _devnull[] = "/dev/null";
#if defined(STDIN_FILENO)
	(void) checkfd(_devnull, STDIN_FILENO, O_RDONLY);
#endif
#if defined(STDOUT_FILENO)
	(void) checkfd(_devnull, STDOUT_FILENO, O_WRONLY);
#endif
#if defined(STDERR_FILENO)
	(void) checkfd(_devnull, STDERR_FILENO, O_WRONLY);
#endif
   }

#if defined(ENABLE_NLS)
    (void) setlocale(LC_ALL, "" );
    (void) bindtextdomain(PACKAGE, LOCALEDIR);
    (void) textdomain(PACKAGE);
#endif

    rpmSetVerbosity(RPMLOG_NOTICE);

    if (optionsTable == NULL) {
	/* Read rpm configuration (if not already read). */
	rpmioConfigured();
	return NULL;
    }

#ifdef	NOTYET
    /* read all RPM POPT configuration files */
    for (int i = 1; i < argc; i++) {
        if (strcmp(argv[i], "--rpmpopt") == 0 && i+1 < argc) {
            rpmpoptfiles = argv[i+1];
            break;
        }
        else if (strncmp(argv[i], "--rpmpopt=", 10) == 0) {
            rpmpoptfiles = argv[i]+10;
            break;
        }
    }
#endif	/* NOTYET */

    /* XXX strip off the "lt-" prefix so that rpmpopt aliases "work". */
  { static const char lt_[] = "lt-";
    const char * s = __progname;
    char * t;
    if (!strncmp(s, lt_, sizeof(lt_)-1))
        s += sizeof(lt_)-1;
    /* XXX strip off trailing -$(VERSION) suffix */
    if ((t = strrchr(s, '-')) != NULL && !strcmp(t+1, VERSION))
	*t = '\0';
    /* XXX undo wdj* -> rpm* renaming. */
    if (!strcmp(s, "wdj"))
	s = "rpm";
    if (!strcmp(s, "wdjbuild"))
	s = "rpmbuild";
    optCon = poptGetContext(s, argc, (const char **)argv, optionsTable, _rpmio_popt_context_flags);
  }

#ifdef	NOTYET
#if defined(RPM_VENDOR_OPENPKG) /* stick-with-rpm-file-sanity-checking */ || \
    !defined(POPT_ERROR_BADCONFIG)	/* XXX POPT 1.15 retrofit */
  { char * path_buf = xstrdup(rpmpoptfiles);
    char *path;
    char *path_next;

    for (path = path_buf; path != NULL && *path != '\0'; path = path_next) {
        const char **av;
        int ac, i;

        /* locate start of next path element */
        path_next = strchr(path, ':');
        if (path_next != NULL && *path_next == ':')
            *path_next++ = '\0';
        else
            path_next = path + strlen(path);

        /* glob-expand the path element */
        ac = 0;
        av = NULL;
        if ((i = rpmGlob(path, &ac, &av)) != 0)
            continue;

        /* work-off each resulting file from the path element */
        for (i = 0; i < ac; i++) {
            const char *fn = av[i];
            if (fn[0] == '@' /* attention */) {
                fn++;
                if (!rpmSecuritySaneFile(fn)) {
                    rpmlog(RPMLOG_WARNING, "existing POPT configuration file \"%s\" considered INSECURE -- not loaded\n", fn);
                    continue;
                }
            }
            (void)poptReadConfigFile(optCon, fn);
            av[i] = _free(av[i]);
        }
        av = _free(av);
    }
    path_buf = _free(path_buf);
  }
#else
    /* XXX FIXME: better error message is needed. */
    if ((xx = poptReadConfigFiles(optCon, rpmpoptfiles)) != 0)
       rpmlog(RPMLOG_WARNING, "existing POPT configuration file \"%s\" considered INSECURE -- not loaded\n", rpmpoptfiles);
#endif

    /* read standard POPT configuration files */
    /* XXX FIXME: the 2nd arg useEnv flag is UNUSED. */
    (void) poptReadDefaultConfig(optCon, 1);

    poptSetExecPath(optCon, USRLIBRPM, 1);
#endif	/* NOTYET */

    /* Process all options, whine if unknown. */
    while ((rc = poptGetNextOpt(optCon)) > 0) {
	const char * optArg = poptGetOptArg(optCon);
	optArg = _free(optArg);
	switch (rc) {
	default:
	    fprintf(stderr, _("%s: option table misconfigured (%d)\n"),
		__progname, rc);
	    exit(EXIT_FAILURE);

	    break;
        }
    }

    if (rc < -1) {
	fprintf(stderr, "%s: %s: %s\n", __progname,
		poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
		poptStrerror(rc));
	exit(EXIT_FAILURE);
    }

    /* Read rpm configuration (if not already read). */
    rpmioConfigured();

    if (__debug) {
	rpmIncreaseVerbosity();
	rpmIncreaseVerbosity();
    }

    /* Spew process and credentials info. */
    rpmioInitInfo(RPMLOG_DEBUG, NULL);

    return optCon;
}

/**
 * Return text between pl and matching pr characters.
 * @param p		start of text
 * @param pl		left char, i.e. '[', '(', '{', etc.
 * @param pr		right char, i.e. ']', ')', '}', etc.
 * @return		pointer to matching pr char in string (or NULL)
 */
RPM_GNUC_PURE
static char *
matchchar(char * p, char pl, char pr)
{
    int lvl = 0;
    char c;

    while ((c = *p++) != '\0') {
        if (c == '\\') {	/* Ignore escaped chars */
            p++;
            continue;
        }
        if (c == pr) {
            if (--lvl <= 0)     return --p;
        } else if (c == pl)
            lvl++;
    }
    return (char *)NULL;
}

rpmioP rpmioPFree(rpmioP P)
{
    if (P != NULL) {
	P->str = _free(P->str);
	P->av = argvFree(P->av);
	P = _free(P);
    }
    return NULL;
}

rpmRC rpmioParse(rpmioP *Pptr, const char * str)
{
    static char whitespace[] = " \t\n\r";
    rpmioP P;
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    char *b;
    char *be;
    int xx;
    int c;

    if ((P = *Pptr) == NULL)
	*Pptr = P = (rpmioP) xcalloc(1, sizeof(*P));

    if (str != NULL) {
	P->str = _free(P->str);
	P->next = P->str = xstrdup(str);
    }

    /* Ltrim whitespace. Anything left to parse? */
    if ((b = P->next) != NULL)
    while (*b && strchr(whitespace, *b))
	*b++ = '\0';
    if (b == NULL || *b == '\0')
	return RPMRC_NOTFOUND;

    /* Parse next command into an argv. */
    P->ac = 0;
    P->av = argvFree(P->av);
    if ((be = b) != NULL)
  while (1) {
    c = *be++;
    switch (c) {
    default:
	break;
    case '\\':		/* escaped character. */
	if (*be != '\0')
	    be++;
	break;
    case '\0':		/* end-of-command termination. */
    case '\n':
    case '\r':
    case ';':	
	if (be[-1] != '\0')
	    be[-1] = '\0';
	else
	    be--;			/* XXX one too far */
	if ((be - b) > 1) {
	    xx = argvAdd(&P->av, b);
	    P->ac++;
	}
	goto exit;
	break;
    case '[':		/* XPath construct with '[' balancing. */
	if ((be = matchchar(be, '[', ']')) == NULL) {
	    be += strlen(b);	/* XXX unmatched ']' */
	    goto exit;
	}
	be++;
	break;
    case '"':		/* quoted string */
	while (1) {
	    if ((be = strchr(be, '"')) == NULL) {
		be += strlen(b);	/* XXX unmatched '"' */
		goto exit;
	    }
	    be++;
	    if (be[-2] == '\\')	/* escaped quote */
		continue;
	    break;
	}
	break;
    case ' ':		/* argument separator */
    case '\t':
	be[-1] = '\0';
	if ((be - b) > 1) {
	    xx = argvAdd(&P->av, b);
	    P->ac++;
	}
	b = be;
	while (*b && (*b == ' ' || *b == '\t'))
	    *b++ = '\0';
	be = b;
	break;
    }
  }
    rc = RPMRC_OK;

exit:
    P->next = be;
    return rc;
}
