#ifndef H_POPTIO
#define	H_POPTIO

/** \ingroup rpmio
 * \file rpmio/poptIO.h
 */

#include <rpmio.h>
#include <rpmiotypes.h>
#include <rpmcb.h>
#include <rpmmacro.h>
#include <rpmmg.h>
#include <rpmpgp.h>
#include <rpmsw.h>
#include <rpmurl.h>
#include <argv.h>
#include <fts.h>
#include <mire.h>
#include <popt.h>

#ifdef __cplusplus
extern "C" {
#endif

/** \ingroup rpmio
 */

extern int __debug;	/* XXX I know I'm gonna regret __debug ... */

extern int _rpmio_popt_context_flags;	/* XXX POPT_CONTEXT_POSIXMEHARDER */

extern pgpHashAlgo rpmioDigestHashAlgo;

/** \ingroup rpmio
 * Popt option table for options to select digest algorithm.
 */
extern struct poptOption		rpmioDigestPoptTable[];

/** \ingroup rpmio
 * Popt option table for options shared by all modes and executables.
 */
extern struct poptOption		rpmioAllPoptTable[];

extern unsigned rpmioFtsFlags;

/** \ingroup rpmio
 * Popt table(s) for options to set Fts(3)/Open(2)/Mount(2)/Umount(2) flags.
 */
extern struct poptOption		rpmioFtsFlagsTable[];
extern unsigned rpmioOpenFlags;
extern struct poptOption		rpmioOpenFlagsTable[];
extern unsigned int rpmioMountFlags;
extern struct poptOption		rpmioMountFlagsTable[];
extern unsigned int rpmioUmountFlags;
extern struct poptOption		rpmioUmountFlagsTable[];

/** \ingroup rpmio
 * Return names from bits using a popt table (limited to 32 bits by popt).
 * @param optTable		popt table
 * @param optFlags		bits
 * @return			(malloc'd) name string
 */
char * rpmioB2N(struct poptOption * optTable, unsigned long optFlags);

/** \ingroup rpmio
 * Return bits from names using a popt table (limited to 32 bits by popt).
 * @param optTable		popt table
 * @param options		names
 * @return			bits
 */
unsigned long rpmioN2B(struct poptOption * optTable, const char * options);

extern const char * rpmioPipeOutput;

extern const char * rpmioRootDir;

/**
 * Make sure that rpm configuration has been read.
 * @warning Options like --rcfile and --verbose must precede callers option.
 */
void rpmioConfigured(void);

/**
 * Initial process and credentials information.
 * @param lvl			rpmlog lvl
 * @param msg			display message string (or NULL)
 */
void
rpmioInitInfo(int lvl, const char *msg);

/**
 * Initialize time and resource usage information.
 * @param lvl			rpmlog lvl
 * @param msg			display message string (or NULL)
 */
int
rpmioInitUsage(int lvl, const char *msg);

/**
 * Spew time and resource usage information.
 * @param lvl			rpmlog lvl
 * @param msg			display message string (or NULL)
 */
void
rpmioFiniInfo(int lvl, const char *msg);

/**
 * Spew resource usage information.
 * @param lvl			rpmlog lvl
 * @param msg			display message string (or NULL)
 */
int
rpmioFiniUsage(int lvl, const char *msg);

/**
 * Destroy most everything needed by an rpm CLI executable context.
 * @param optCon		popt context
 * @return			NULL always
 */
poptContext
rpmioFini(poptContext optCon);

/**
 * Initialize most everything needed by an rpmio executable context.
 * @param argc			no. of args
 * @param argv			arg array
 * @param optionsTable		popt option table
 * @return			popt context (or NULL)
 */
poptContext
rpmioInit(int argc, char *const argv[], struct poptOption * optionsTable);

#ifdef __cplusplus
}
#endif

#endif	/* H_POPTIO */
