#ifndef	H_RPMBAG
#define	H_RPMBAG

/** \ingroup rpmio
 * \file rpmio/rpmbag.h
 */

/** \ingroup rpmio
 */
extern int _rpmbag_debug;

/** \ingroup rpmio
 */
typedef struct rpmbag_s * rpmbag;
typedef struct rpmsdb_s * rpmsdb;

#if defined(_RPMBAG_INTERNAL)

struct rpmsdb_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    void * _bf;
    int dbmode;
    void * _db;
};

/** \ingroup rpmio
 */
struct rpmbag_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * fn;
    int flags;
    size_t nsdbp;
    rpmsdb * sdbp;
};
#endif	/* _RPMBAG_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a bag wrapper instance.
 * @param bag		bag wrapper
 * @return		NULL on last dereference
 */
rpmbag rpmbagUnlink (rpmbag bag);
#define	rpmbagUnlink(_bag)	\
    ((rpmbag)rpmioUnlinkPoolItem((rpmioItem)(_bag), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a bag wrapper instance.
 * @param bag		bag wrapper
 * @return		new bag wrapper reference
 */
rpmbag rpmbagLink (rpmbag bag);
#define	rpmbagLink(_bag)	\
    ((rpmbag)rpmioLinkPoolItem((rpmioItem)(_bag), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a bag wrapper.
 * @param bag		bag wrapper
 * @return		NULL on last dereference
 */
rpmbag rpmbagFree(rpmbag bag);
#define	rpmbagFree(_bag)	\
    ((rpmbag)rpmioFreePoolItem((rpmioItem)(_bag), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a bag wrapper.
 * @param fn		bag file
 * @param flags		bag flags
 * @return		new bag wrapper
 */
rpmbag rpmbagNew(const char * fn, int flags);

int rpmbagAdd(rpmbag bag, void *_db, int dbmode);

int rpmbagDel(rpmbag bag, int i);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMBAG */
