#ifndef	_H_RPMBZ_
#define	_H_RPMBZ_

#include <bzlib.h>

/**
 */
typedef	struct rpmbz_s * rpmbz;

/**
 */
#if defined(_RPMBZ_INTERNAL)
struct rpmbz_s {
    BZFILE *bzfile;	
    bz_stream strm;
    int bzerr;
    int oflags;			/*!< open flags */
    int omode;			/*!< open mode */
    FILE * fp;			/*!< file pointer */
    int B;			/*!< blockSize100K (default: 9) */
    int S;			/*!< small (default: 0) */
    int V;			/*!< verbosity (default: 0) */
    int W;			/*!< workFactor (default: 30) */
    unsigned int nbytes_in;
    unsigned int nbytes_out;

    unsigned int blocksize;	/*!< compression block size */

    void * zq;			/*!< buffer pools. */
    void * zlog;		/*!< trace logging. */
};

#ifdef __cplusplus
GENfree(rpmbz)
#endif	/* __cplusplus */

static rpmbz rpmbzFini(rpmbz bz)
{
    bz = _free(bz);
    return NULL;
}

static
rpmbz rpmbzInit(int level, int small, int verbosity, int oflags, mode_t omode)
{
    rpmbz bz = (rpmbz) xcalloc(1, sizeof(*bz));
    static int _bzdB = 9;
    static int _bzdS = 0;
    static int _bzdV = 1;
    static int _bzdW = 30;

    bz->B = (level >= 1 && level <= 9) ? level : _bzdB;
    bz->S = (small >= 0) ? small : _bzdS;
    bz->V = (verbosity >= 0 && verbosity <= 4) ? verbosity : _bzdV;
    bz->W = _bzdW;
    bz->oflags = (oflags != -1 ? oflags : O_RDONLY);
    bz->omode = (omode != 0177777 ? omode : 0644);
    return bz;
}
#endif	/* _RPMBZ_INTERNAL */

/**
 */
const char * rpmbzStrerror(rpmbz bz);

/**
 */
void rpmbzClose(rpmbz bz, int abort, const char ** errmsg);

/**
 */
rpmbz rpmbzFree(rpmbz bz, int abort);

/**
 */
rpmbz rpmbzNew(const char * path, const char * fmode, int fdno);

/**
 */
ssize_t rpmbzRead(rpmbz bz, char * buf, size_t count, const char ** errmsg);

/**
 */
ssize_t rpmbzWrite(rpmbz bz, const char * buf, size_t count,
		const char ** errmsg);

#ifdef	NOTYET
/**
 */
int rpmbzSeek(void * _bz, _libio_pos_t pos, int whence);

/**
 */
rpmbz rpmbzOpen(const char * path, const char * fmode);

/**
 */
rpmbz rpmbzFdopen(void * _fdno, const char * fmode);

/**
 */
int rpmbzFlush(void * _bz);

/**
 */
int rpmbzCompressBlock(rpmbz bz, rpmzJob job);

/**
 */
int rpmbzDecompressBlock(rpmbz bz, rpmzJob job);
#endif	/* NOTYET */

#endif	/* _H_RPMBZ_ */
