#ifndef	H_RPMCAP
#define	H_RPMCAP

/** \ingroup rpmio
 * \file rpmio/rpmcap.h
 */

#include <rpmiotypes.h>
#include <rpmio.h>

extern int _rpmcap_debug;

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Return validated capability string.
 * @param s	capability string
 * @return	validated capability string (malloc'd)
 */
char * rpmcapValidate(const char *s);

/**
 * Get file capabity from fdno.
 * @param fdno	file descriptor
 * @return	capability string (malloc'd)
 */
char * rpmcapGetFD(int fdno);

/**
 * Get file capabity from path.
 * @param fn	file name
 * @return	capability string (malloc'd)
 */
char * rpmcapGetFN(const char *fn);

/**
 * Set file capabity on fdno.
 * @param fdno	file descriptor
 * @param s	capability string
 * @return	0 on success
 */
int rpmcapSetFD(int fdno, const char *s);

/**
 * Set file capabity on path.
 * @param fn	file name
 * @param s	capability string
 * @return	0 on success
 */
int rpmcapSetFN(const char *fn, const char *s);

/**
 * Verify file capabity on fdno.
 * @param fdno	file descriptor
 * @param s	capability string
 * @return	0 on success
 */
int rpmcapVerifyFD(int fdno, const char *s);

/**
 * Verify file capabity on path.
 * @param fn	file name
 * @param s	capability string
 * @return	0 on success
 */
int rpmcapVerifyFN(const char *fn, const char *s);

/**
 * Get process capabity from pid.
 * @param pid	process id
 * @return	capability string (malloc'd)
 */
char * rpmcapGetP(pid_t pid);

/**
 * Set process capabity on pid.
 * @param pid	process id
 * @param s	capability string
 * @return	0 on success
 */
int rpmcapSetP(pid_t pid, const char * s);

int rpmcapDump(int lvl, const char * msg);
int rpmcapTest(const char *s);
int rpmcapDrop(const char *s);

int Prctl(int option, unsigned long arg2, unsigned long arg3,
                 unsigned long arg4, unsigned long arg5);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMCAP */
