#ifndef RPMDAV_H
#define RPMDAV_H

/** \ingroup rpmio
 * \file rpmio/rpmdav.h
 */
#include <rpmio.h>
#include <rpmurl.h>

/**
 */
extern int _ftp_debug;
extern int _dav_debug;
extern int rpmioHttpReadTimeoutSecs;
extern int rpmioHttpConnectTimeoutSecs;
extern const char * rpmioHttpAccept;
extern const char * rpmioHttpUserAgent;

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Create an argv directory from an ftp:// URI.
 * @param path		ftp:// directory path
 * @return 		argv DIR
 */
DIR * ftpOpendir(const char * path);

/**
 * Close active neon transfer(s) (if any).
 * @param _u		URL container
 * @return		0 on success
 */
int davDisconnect(void * _u);

/**
 * Free persistent neon session state.
 * @param u		URL container
 * @return		0 on success
 */
int davFree(urlinfo u);

/**
 * Free global neon+openSSL state memory.
 */
void davDestroy(void);

/**
 * Send a http request.
 * @param ctrl		connection stream
 * @param httpCmd	http command
 * @param httpArg	http command argument (NULL if none)
 * @returns		0 on success
 */
int davReq(FD_t ctrl, const char * httpCmd, const char * httpArg);

/**
 * Read a http response.
 * @param u		URL container
 * @param ctrl		connection stream
 * @retval *str		error msg		
 * @returns		0 on success
 */
int davResp(urlinfo u, FD_t ctrl, char *const * str);

/**
 * Open a URL.
 * @param url
 * @param flags
 * @param mode
 * @retval *uret
 * @return
 */
FD_t davOpen(const char * url, int flags, mode_t mode, urlinfo * uret);

FD_t httpOpen(const char * url, int flags, mode_t mode, urlinfo * uret);

/**
 */
ssize_t davRead(void * cookie, char * buf, size_t count);

/**
 */
ssize_t davWrite(void * cookie, const char * buf, size_t count);

/**
 */
int davSeek(void * cookie, _libio_pos_t pos, int whence);

/**
 */
int davClose(void * cookie);

/**
 */
int davMkdir(const char * path, mode_t mode);

/**
 */
int davRmdir(const char * path);

/**
 */
int davRename(const char * oldpath, const char * newpath);

/**
 */
int davUnlink(const char * path);

/**
 * Close a DAV collection.
 * @param dir		argv DIR
 * @return 		0 always
 */
int davClosedir(DIR * dir);

/**
 * Return next entry from a DAV collection.
 * @param dir		argv DIR
 * @return 		next entry
 */
struct dirent * davReaddir(DIR * dir);

/**
 * Create an argv directory from DAV collection.
 * @param path		URL for DAV collection path
 * @return 		argv DIR
 */
DIR * davOpendir(const char * path);

/**
 * stat(2) clone.
 */
int davStat(const char * path, struct stat * st);

/**
 * lstat(2) clone.
 */
int davLstat(const char * path, struct stat * st);

/**
 * realpath(3) clone.
 */
char * davRealpath(const char * path, char * resolved_path);

#ifdef __cplusplus
}
#endif

#endif /* RPMDAV_H */
