#include "system.h"

#if defined(WITH_LIBEV)

#if defined(HAVE_EV_H)
# include <ev.h>
#endif

#else	/* WITH_LIBEV */

/* XXX stub in enough to include rpmev.h */
#define	EV_P_		void * loop,
#define	EV_A_		NULL,
#define	EV_DEFAULT	ev_default_loop(0)
#define	EVFLAG_AUTO	0
#define	EVRUN_ONCE	2

typedef	struct ev_watcher	ev_idle;
typedef	struct ev_watcher	ev_async;

struct ev_watcher {
    int active;
    int pending;
    int priority;
    void * data;
    void (*cb)(EV_P_ void *w, int revents);
};

union ev_any_watcher
{
    struct ev_watcher w;
    struct ev_watcher async;
};

#define	ev_default_loop(_a)	NULL

#define	ev_userdata(_x)		NULL

#define	ev_set_userdata(_x, y)

#define	ev_idle_stop(_w)	(void)_w
#define	ev_async_send(_l, _w)	(void)_w
#define	ev_run(_a, _b)
#define	ev_loop_destroy(_a)

#endif	/* WITH_LIBEV */

#if defined(WITH_LIBEIO)

#if defined(HAVE_EIO_H)
# include <eio.h>
#endif

#else	/* WITH_LIBEIO */

typedef struct eio_req		eio_req;
typedef int (*eio_cb)(eio_req *req);

typedef struct eio_pwd *	eio_wd;
typedef double			eio_tstamp;
typedef int			eio_uid_t;
typedef int			eio_gid_t;
typedef ssize_t			eio_ssize_t;

/* XXX stub in enough to include rpmeio.h */
struct eio_req
{
    eio_req volatile *next;	/* private ETP */
    eio_wd wd;		/* all applicable requests: working directory of pathname, old name; wd_open: return wd */
    eio_ssize_t result;	/* result of syscall, e.g. result = read (... */
    off_t offs;		/* read, write, truncate, readahead, sync_file_range, fallocate: file offset, mknod: dev_t */
    size_t size;	/* read, write, readahead, sendfile, msync, mlock, sync_file_range, fallocate: length */
    void *ptr1;		/* all applicable requests: pathname, old name; readdir: optional eio_dirents */
    void *ptr2;		/* all applicable requests: new name or memory buffer; readdir: name strings */
    eio_tstamp nv1;	/* utime, futime: atime; busy: sleep time */
    eio_tstamp nv2;	/* utime, futime: mtime */

    int int1;		/* all applicable requests: file descriptor; sendfile: output fd; open, msync, mlockall, readdir: flags */
    long int2;		/* chown, fchown: uid; sendfile: input fd; open, chmod, mkdir, mknod: file mode, seek: whence, sync_file_range, fallocate: flags */
    long int3;		/* chown, fchown: gid; rename, link: working directory of new name */
    int errorno;	/* errno value on syscall return */

    signed char type;	/* EIO_xxx constant ETP */

    unsigned char cancelled; /* ETP */

    unsigned char flags;	/* private */
#define	EIO_FLAG_PTR1_FREE	0x01
#define	EIO_FLAG_PTR2_FREE	0x02

    signed char pri;	/* the priority */
#define	EIO_PRI_MIN		-4
#define	EIO_PRI_MAX		4
#define	EIO_PRI_DEFAULT		0

    void *data;
    eio_cb finish;
    void (*destroy)(eio_req *req); /* called when request no longer needed */
    void (*feed)(eio_req *req);    /* only used for group requests */

    eio_req *grp, *grp_prev, *grp_next, *grp_first; /* private ETP */
};

#define	eio_poll()		-1
#define	eio_nthreads()		0
#define	eio_nready()		0
#define	eio_nreqs()		0
#define	eio_npending()		0
#define	eio_init(_a, _b)	-1
#define	eio_cancel(_a)
#define	eio_submit(_a)

#define	eio_set_max_poll_time(max_poll_time);
#define	eio_set_max_poll_reqs(max_poll_reqs);
#define	eio_set_min_parallel(min_parallel);
#define	eio_set_max_parallel(max_parallel);
#define	eio_set_max_idle(max_idle);
#define	eio_set_idle_timeout(idle_timeout);

#define EIO_RESULT(req)		((req)->result)
#define EIO_BUF(req)		((req)->ptr2)
#define EIO_STAT_BUF(req)	(void *)EIO_BUF(req)
#define EIO_STATVFS_BUF(req)	(void *)EIO_BUF(req)
#define EIO_PATH(req)		((char *)(req)->ptr1)

#endif	/* WITH_LIBEIO */

#include <rpmio.h>
#include <rpmzlog.h>
#include <poptIO.h>

#define	_RPMEV_INTERNAL
#include <rpmev.h>

#define	_RPMEIO_INTERNAL
#include <rpmeio.h>

#include "debug.h"

int _rpmeio_debug = -1;

#define SPEW(_fmt, ...) \
    if (_rpmeio_debug || _rpmio_debug) \
	fprintf(stderr, _fmt "\n", __VA_ARGS__)

#define ZSPEW(_fmt, ...) \
    if (zlog) \
	rpmzLogAdd(zlog, _fmt, __VA_ARGS__)

#define REQ_ERRNO(req)		((req)->errorno)

/*==============================================================*/
const char * rpmeioReqType(int reqtype)
{
    switch (reqtype) {
    case RPMEIO_CUSTOM:		return("EIO_CUSTOM");
    case RPMEIO_WD_OPEN:	return("EIO_WD_OPEN");
    case RPMEIO_WD_CLOSE:	return("EIO_WD_CLOSE");
    case RPMEIO_CLOSE:		return("EIO_CLOSE");
    case RPMEIO_DUP2:		return("EIO_DUP2");
    case RPMEIO_SEEK:		return("EIO_SEEK");
    case RPMEIO_READ:		return("EIO_READ");
    case RPMEIO_WRITE:		return("EIO_WRITE");
    case RPMEIO_READAHEAD:	return("EIO_READAHEAD");
    case RPMEIO_SENDFILE:	return("EIO_SENDFILE");
    case RPMEIO_FSTAT:		return("EIO_FSTAT");
    case RPMEIO_FSTATVFS:	return("EIO_FSTATVFS");
    case RPMEIO_FTRUNCATE:	return("EIO_FTRUNCATE");
    case RPMEIO_FUTIME:		return("EIO_FUTIME");
    case RPMEIO_FCHMOD:		return("EIO_FCHMOD");
    case RPMEIO_FCHOWN:		return("EIO_FCHOWN");
    case RPMEIO_SYNC:		return("EIO_SYNC");
    case RPMEIO_FSYNC:		return("EIO_FSYNC");
    case RPMEIO_FDATASYNC:	return("EIO_FDATASYNC");
    case RPMEIO_SYNCFS:		return("EIO_SYNCFS");
    case RPMEIO_MSYNC:		return("EIO_MSYNC");
    case RPMEIO_MTOUCH:		return("EIO_MTOUCH");
    case RPMEIO_SYNC_FILE_RANGE: return("EIO_SYNC_FILE_RANGE");
    case RPMEIO_FALLOCATE:	return("EIO_FALLOCATE");
    case RPMEIO_MLOCK:		return("EIO_MLOCK");
    case RPMEIO_MLOCKALL:	return("EIO_MLOCKALL");
    case RPMEIO_GROUP:		return("EIO_GROUP");
    case RPMEIO_NOP:		return("EIO_NOP");
    case RPMEIO_BUSY:		return("EIO_BUSY");
    case RPMEIO_REALPATH:	return("EIO_REALPATH");
    case RPMEIO_READDIR:	return("EIO_READDIR");
    case RPMEIO_OPEN:		return("EIO_OPEN");
    case RPMEIO_STAT:		return("EIO_STAT");
    case RPMEIO_LSTAT:		return("EIO_LSTAT");
    case RPMEIO_STATVFS:	return("EIO_STATVFS");
    case RPMEIO_TRUNCATE:	return("EIO_TRUNCATE");
    case RPMEIO_UTIME:		return("EIO_UTIME");
    case RPMEIO_CHMOD:		return("EIO_CHMOD");
    case RPMEIO_CHOWN:		return("EIO_CHOWN");
    case RPMEIO_UNLINK:		return("EIO_UNLINK");
    case RPMEIO_RMDIR:		return("EIO_RMDIR");
    case RPMEIO_MKDIR:		return("EIO_MKDIR");
    case RPMEIO_RENAME:		return("EIO_RENAME");
    case RPMEIO_MKNOD:		return("EIO_MKNOD");
    case RPMEIO_LINK:		return("EIO_LINK");
    case RPMEIO_SYMLINK:	return("EIO_SYMLINK");
    case RPMEIO_READLINK:	return("EIO_READLINK");
    }
    return ("?EIO_TYPE?");
}

void rpmeioDumpREQ(const char *msg, eio_req * req, FILE *fp)
{
    if (fp == NULL)	fp = stderr;
    if (msg) fprintf(stderr, "========================== %s(%p)\n", msg, req);

    if (req) {
	rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);

	if (eio->evl)
	    fprintf(fp, "%25s: %p\n", "evl", eio->evl);

#define PRINT(_fmt, _foo) \
    {   fprintf(fp, "%25s: %"#_fmt"\n", #_foo, req->_foo); }
	if (req->next)
	    PRINT(p, next);
	if (req->wd)
	    PRINT(p, wd);

	if (req->result)
	    PRINT(zd, result);
	if (req->offs)
	    PRINT(ld, offs);
	if (req->size)
	    PRINT(zu, size);

#ifdef	DYING
	if (req->ptr1)
	    PRINT(p, ptr1);
	if (req->ptr2)
	    PRINT(p, ptr2);
#else
	if (req->ptr1)
	    fprintf(fp, "%25s: %p %s\n", "ptr1", req->ptr1, (char *)req->ptr1);
	if (req->ptr2)
	    fprintf(fp, "%25s: %p %s\n", "ptr2", req->ptr2, (char *)req->ptr2);
#endif

	if (req->nv1 != 0.0)
	    PRINT(f, nv1);
	if (req->nv2 != 0.0)
	    PRINT(f, nv2);

	if (req->int1)
	    PRINT(d, int1);
	if (req->int2)
	    PRINT(ld, int2);
	if (req->int3)
	    PRINT(ld, int3);

#ifdef	DYING
	PRINT(d, errorno);
#else
	if (req->result < 0 && req->errorno)
	    fprintf(fp, "%25s: %d %s\n", "errorno", REQ_ERRNO(req), strerror(REQ_ERRNO(req)));
#endif

#ifdef	DYING
	PRINT(hhx, type);
#else
	fprintf(fp, "%25s: %d %s data %s\n", "type", req->type, REQ_TYPE(req), (char *)REQ_DATA(req));
#endif

	if (req->cancelled)
	    PRINT(hhx, cancelled);
	if (req->flags)
	    PRINT(hhx, flags);
	if (!(req->pri == 0 || req->pri == 4))
	    PRINT(hhx, pri);

#ifdef	NOTYET
	if (req->data)
	    PRINT(p, data);
#endif
	if (req->finish)
	    PRINT(p, finish);
	if (req->destroy)
	    PRINT(p, destroy);
	if (req->feed)
	    PRINT(p, feed);

	if (req->grp)
	    PRINT(p, grp);
	if (req->grp_prev)
	    PRINT(p, grp_prev);
	if (req->grp_next)
	    PRINT(p, grp_next);
	if (req->grp_first)
	    PRINT(p, grp_first);
#undef	PRINT

    }
}

/*==============================================================*/

struct rpmevl_s {
    unsigned flags;
    int pipefdno[2];
#define	pipein	evl->pipefdno[0]
#define	pipeout	evl->pipefdno[1]
    struct ev_loop *loop;
    rpmev idle_ev;
    rpmev async_ev;
    rpmzLog zlog;		/*!< high precision timestamp'd log */
};
static struct rpmevl_s _evl;

static rpmevl rpmevlGet(void)
{
    if (_evl.flags & RPMEVL_PIPE)
	return &_evl;
    return (rpmevl) ev_userdata(EV_DEFAULT);
}

/* idle watcher callback, only used when eio_poll */
/* didn't handle all results in one call */
static void
repeat (EV_P_ ev_idle *w, int revents)
{
    rpmevl evl = (rpmevl) ev_userdata(loop);
    rpmzLog zlog = evl->zlog;
ZSPEW("--> %s(%p, %p, %d)", __FUNCTION__, loop, w, revents);

    if (eio_poll() != -1) {
	ev_idle_stop (EV_A_ w);
    }
}

/* eio has some results, process them */
static void
ready (EV_P_ ev_async *w, int revents)
{
    rpmevl evl = (rpmevl) ev_userdata(loop);
    rpmzLog zlog = evl->zlog;
ZSPEW("--> %s(%p, %p, %d)", __FUNCTION__, loop, w, revents);

    if (eio_poll() == -1)
	evl->idle_ev = rpmevNew(RPMEV_TYPE_IDLE,      NULL);
}

/* wake up the event loop */
static
void want_poll (void)
{
    rpmevl evl = rpmevlGet();
    rpmzLog zlog = evl->zlog;

    if (evl->flags & RPMEVL_PIPE) {
	char dummy = '\0';
	ssize_t nw = write(pipeout, &dummy, sizeof(dummy));
ZSPEW("<-- %s(%d) nw %zd", "write", pipeout, nw);
    } else {
	ev_async * w = &evl->async_ev->evw.async;
	ev_async_send(evl->loop, w);
ZSPEW("<-- %s()", __FUNCTION__);
    }
}

static
void done_poll(void)
{
    rpmevl evl = rpmevlGet();
    rpmzLog zlog = evl->zlog;

    if (evl->flags & RPMEVL_PIPE) {
	char dummy = '\0';
	ssize_t nr = read(pipein, &dummy, sizeof(dummy));
ZSPEW("<-- %s(%d) nr %zd", " read", pipein, nr);
    }
ZSPEW("<-- %s() nthreads:%u nreqs:%u nready:%u npending:%u", __FUNCTION__, eio_nthreads(), eio_nreqs(), eio_nready(), eio_npending());
}

#ifdef	TODO

#define	EIO_STACKSIZE	4096	/* default? */

eio _req *grp = eio_grp(eio_cb cb, void *data);
    eio_grp_feed(eio_req *grp, void (*feed)(eio_req *req), int limit);
    eio_grp_limit(eio_req *grp, int limit);
    eio_grp_add(eio_req *grp, eio_req *req);
    eio_grp_cancel(eio_req *grp);

    eio_ssize_t eio_sendfile_sync (int ofd, int ifd, off_t offset, size_t count);

#endif	/* TODO */

rpmevl rpmeioStart(double max_poll_time, unsigned max_poll_reqs,
		unsigned min_parallel, unsigned max_parallel,
		unsigned max_idle, unsigned idle_timeout, unsigned flags)
{
    int rc;
    rpmevl evl = NULL;

    if (flags & RPMEVL_PIPE) {
	evl = &_evl;
	evl->pipefdno[0] = -1;
	evl->pipefdno[1] = -1;
	evl->loop = NULL;
	evl->idle_ev = NULL;
	evl->async_ev = NULL;

	rc = pipe(evl->pipefdno);
	if (!rc)
	    rc = eio_init(want_poll, done_poll);

SPEW("<-- %s() evl %p pipe [%d,%d]", __FUNCTION__, evl, evl->pipefdno[0], evl->pipefdno[1]);
    } else {
	evl = xcalloc(1, sizeof(*evl));
	evl->pipefdno[0] = -1;
	evl->pipefdno[1] = -1;

	evl->loop = ev_default_loop(EVFLAG_AUTO);
	ev_set_userdata(evl->loop, evl);
	evl->async_ev = rpmevNew(RPMEV_TYPE_ASYNC,     ready);

	rc = eio_init (want_poll, done_poll);

SPEW("<-- %s() evl %p loop %p", __FUNCTION__, evl, evl->loop);
    }
    if (max_poll_time > 0.0) {
	eio_set_max_poll_time(max_poll_time);
    }
    if (max_poll_reqs) {
	eio_set_max_poll_reqs(max_poll_reqs);
    }
    if (min_parallel) {
	eio_set_min_parallel(min_parallel);
    }
    if (max_parallel) {
	eio_set_max_parallel(max_parallel);
    }
    if (max_idle) {
	eio_set_max_idle(max_idle);
    }
    if (idle_timeout) {
	eio_set_idle_timeout(idle_timeout);
    }
    evl->flags = flags;
    evl->zlog = (_rpmeio_debug || _rpmio_debug) ? rpmzLogNew(NULL) : NULL;

    return evl;
}

void rpmeioLoop(rpmevl evl)
{
    rpmzLog zlog = (evl ? evl->zlog : NULL);
    int npolls = 0;

ZSPEW("==> %s()", __FUNCTION__);

    struct pollfd pfd;
    pfd.fd = pipein;
    pfd.events = POLLIN;

    while (eio_nreqs() > 0) {
ZSPEW("*** %s() nthreads:%u nreqs:%u nready:%u npending:%u", __FUNCTION__, eio_nthreads(), eio_nreqs(), eio_nready(), eio_npending());
	if (evl->flags & RPMEVL_PIPE) {
	    (void) poll(&pfd, 1, -1);
	    npolls = eio_poll();
	} else {
	    ev_run(evl->loop, EVRUN_ONCE);
	    npolls++;
	}
    }

ZSPEW("<== %s() npolls %d", __FUNCTION__, npolls);
}

void rpmeioDrain(rpmevl evl)
{
    while (eio_nreqs() > 0)
	rpmeioLoop(evl);

    if (!(evl->flags & RPMEVL_PIPE)) {
	if (evl->idle_ev)
	    evl->idle_ev = rpmevFree(evl->idle_ev);
	if (evl->async_ev)
	    evl->async_ev = rpmevFree(evl->async_ev);
	ev_loop_destroy(evl->loop);
    }

fprintf(stderr, "***************************************\n");
    if (evl->zlog)
	evl->zlog = rpmzLogDump(evl->zlog, NULL);
fprintf(stderr, "***************************************\n");

    memset(evl, 0, sizeof(*evl));
    if (evl != &_evl)
	evl = _free(evl);

}

void Eio_cancel(eio_req * req)
{
    rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);
SPEW("--> %s(%p) eio %p", __FUNCTION__, req, eio);
    eio_cancel(req);
}

void Eio_destroy(eio_req * req)
{
    rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);
SPEW("--> %s(%p) eio %p", __FUNCTION__, req, eio);

    /* XXX wipe variables to shorten rpmeioDumpREQ() spewage. */
    eio->evl = NULL;

    req->next = NULL;
    req->result = 0;
    req->errorno = 0;

    req->pri = 0;
    req->flags = 0;
    req->ptr1 = NULL;
    req->ptr2 = NULL;

    req->nv1 = 0.0;
    req->nv2 = 0.0;
    req->offs = 0;
    req->size = 0;
    req->int1 = 0;
    req->int2 = 0;
    req->int3 = 0;
    req->finish = NULL;
    req->destroy = NULL;

rpmeioDumpREQ(__FUNCTION__, req, NULL);

    do {
	eio = rpmeioFree(eio);
    } while (0);

}

eio_req * Eio_submit(eio_req * req)
{
    rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);
    rpmevl evl = (rpmevl) eio->evl;
    rpmzLog zlog = (evl ? evl->zlog : NULL);
#ifdef	NOTYET
    eio = rpmeioLink(eio);
#endif
assert(req == &eio->req);
    req = &eio->req;
    eio_submit(req);
ZSPEW("<-- %s(%p) eio %p", __FUNCTION__, req, eio);
    return req;
}

/*==============================================================*/
int rpmeioCB(eio_req * req)
{
    rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);
    rpmevl evl = (rpmevl) eio->evl;
    rpmzLog zlog = (evl ? evl->zlog : NULL);
ZSPEW("--> %s(%s|%p) res %ld eio %p", __FUNCTION__, REQ_TYPE(req), REQ_DATA(req), EIO_RESULT(req), eio);

    if (req->result >= 0)
    switch (req->type) {
    case RPMEIO_READLINK:
    case RPMEIO_REALPATH:
    {
	char * b = (char *) EIO_BUF(req);
	b[req->result] = '\0';
    }   break;
    default:
	break;
    }

rpmeioDumpREQ(__FUNCTION__, req, NULL);

    if (EIO_RESULT(req) < 0) {
	errno = REQ_ERRNO(req);
	perror(REQ_TYPE(req));
    }

#ifdef	NOTYET
    eio = rpmeioFree(eio);
#endif

    return 0;
}

int rpmeioCB_readdir(eio_req * req)
{
    rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);
    rpmevl evl = (rpmevl) eio->evl;
    rpmzLog zlog = (evl ? evl->zlog : NULL);
    char *buf = (char *) EIO_BUF(req);

ZSPEW("--> %s(%s|%p) res %ld eio %p", __FUNCTION__, REQ_TYPE(req), REQ_DATA(req), EIO_RESULT(req), eio);
rpmeioDumpREQ(__FUNCTION__, req, NULL);

    if (EIO_RESULT(req) < 0)
	return 0;

    while (EIO_RESULT(req)--) {
	SPEW("\treaddir = <%s>", buf);
	buf += strlen(buf) + 1;
    }

#ifdef	NOTYET
    eio = rpmeioFree(eio);
#endif

    return 0;
}

int rpmeioCB_stat(eio_req * req)
{
    rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);
    rpmevl evl = (rpmevl) eio->evl;
    rpmzLog zlog = (evl ? evl->zlog : NULL);
    struct stat *buf = EIO_STAT_BUF(req);

ZSPEW("--> %s(%s|%p) res %ld eio %p", __FUNCTION__, REQ_TYPE(req), REQ_DATA(req), EIO_RESULT(req), eio);
rpmeioDumpREQ(__FUNCTION__, req, NULL);

    if (req->type == RPMEIO_FSTAT) {
	SPEW("\tfstat = %ld", EIO_RESULT(req));
    } else {
	SPEW("\tstat(%s) = %ld", EIO_PATH(req), EIO_RESULT(req));
    }

    if (!EIO_RESULT(req))
	SPEW("\tstat size %ld perm 0%o", buf->st_size,
	       buf->st_mode & 0777);

#ifdef	NOTYET
    eio = rpmeioFree(eio);
#endif

    return 0;
}

int rpmeioCB_read(eio_req * req)
{
    rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);
    rpmevl evl = (rpmevl) eio->evl;
    rpmzLog zlog = (evl ? evl->zlog : NULL);
    unsigned char *buf = (unsigned char *) EIO_BUF(req);

ZSPEW("--> %s(%s|%p) res %ld eio %p (%02x%02x%02x%02x %02x%02x%02x%02x)",
	__FUNCTION__, REQ_TYPE(req), REQ_DATA(req), EIO_RESULT(req), eio,
	buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7]);
rpmeioDumpREQ(__FUNCTION__, req, NULL);

#ifdef	NOTYET
    eio = rpmeioFree(eio);
#endif

    return 0;
}

int rpmeioLastFdno;

int rpmeioCB_open(eio_req * req)
{
    rpmeio eio = (rpmeio) (((rpmioItem)req) - 1);
    rpmevl evl = (rpmevl) eio->evl;
    rpmzLog zlog = (evl ? evl->zlog : NULL);
ZSPEW("--> %s(%s|%p) res %ld eio %p", __FUNCTION__, REQ_TYPE(req), REQ_DATA(req), EIO_RESULT(req), eio);
rpmeioDumpREQ(__FUNCTION__, req, NULL);

    rpmeioLastFdno = EIO_RESULT(req);

#ifdef	NOTYET
    eio = rpmeioFree(eio);
#endif

    return 0;
}

/*==============================================================*/
eio_req * Eio_request(int reqtype, ...)
{
    rpmeio eio = NULL;
    rpmevl evl = NULL;
    rpmzLog zlog = NULL;
    eio_req * req = NULL;

    int pri = 0;
    eio_cb cb = rpmeioCB;
    void * data = NULL;

    if (reqtype >= RPMEIO_REQ_TYPE_NUM)
	goto exit;

    eio = rpmeioNew(reqtype);
    evl = eio->evl;
    zlog = (evl ? evl->zlog : NULL);
    req = &eio->req;

    va_list ap;
    va_start(ap, reqtype);

    switch (reqtype) {
    case RPMEIO_CUSTOM:
	req->feed = va_arg(ap, void (*) (eio_req *));
 	cb = rpmeioCB;		/* XXX */
	data = "custom";
	break;
    case RPMEIO_WD_OPEN:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
 	cb = rpmeioCB;		/* XXX */
	data = "wd_open";
	break;
    case RPMEIO_WD_CLOSE:
	req->wd = va_arg(ap, eio_wd);
 	cb = rpmeioCB;		/* XXX */
	data = "wd_close";
	break;
    case RPMEIO_CLOSE:
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB;
	data = "close";
	break;
    case RPMEIO_DUP2:
	req->int1 = va_arg(ap, int);
	req->int2 = va_arg(ap, int);
 	cb = rpmeioCB;
	data = "dup2";
	break;
    case RPMEIO_SEEK:
	req->int1 = va_arg(ap, int);
	req->offs = va_arg(ap, off_t);
	req->int2 = va_arg(ap, int);
 	cb = rpmeioCB;		/* XXX */
	data = "seek";
	break;
    case RPMEIO_READ:
	req->int1 = va_arg(ap, int);
	req->ptr2 = va_arg(ap, void *);
	req->size = va_arg(ap, size_t);
	req->offs = va_arg(ap, off_t);
 	cb = rpmeioCB_read;
	data = "read";
	break;
    case RPMEIO_WRITE:
	req->int1 = va_arg(ap, int);
	req->ptr2 = va_arg(ap, void *);
	req->size = va_arg(ap, size_t);
	req->offs = va_arg(ap, off_t);
 	cb = rpmeioCB;
	data = "write";
	break;
    case RPMEIO_READAHEAD:
	req->int1 = va_arg(ap, int);
	req->offs = va_arg(ap, off_t);
	req->size = va_arg(ap, size_t);
 	cb = rpmeioCB;
	data = "readahead";
	break;
    case RPMEIO_SENDFILE:
	req->int1 = va_arg(ap, int);
	req->int2 = va_arg(ap, int);
	req->offs = va_arg(ap, off_t);
	req->size = va_arg(ap, size_t);
 	cb = rpmeioCB;
	data = "sendfile";
	break;
    case RPMEIO_FSTAT:
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB_stat;
	data = "fstat";
	break;
    case RPMEIO_FSTATVFS:
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB;		/* XXX */
	data = "statvfs";
	break;
    case RPMEIO_FTRUNCATE:
	req->int1 = va_arg(ap, int);
	req->offs = va_arg(ap, off_t);
 	cb = rpmeioCB;
	data = "ftruncate";
	break;
    case RPMEIO_FUTIME:
	req->int1 = va_arg(ap, int);
	req->nv1 = va_arg(ap, eio_tstamp);
	req->nv2 = va_arg(ap, eio_tstamp);
 	cb = rpmeioCB;
	data = "futime";
	break;
    case RPMEIO_FCHMOD:
	req->int1 = va_arg(ap, int);
	req->int2 = va_arg(ap, mode_t);
 	cb = rpmeioCB;
	data = "fchmod";
	break;
    case RPMEIO_FCHOWN:
	req->int1 = va_arg(ap, int);
	req->int2 = va_arg(ap, uid_t);
	req->int3 = va_arg(ap, gid_t);
 	cb = rpmeioCB;
	data = "fchown";
	break;
    case RPMEIO_SYNC:
 	cb = rpmeioCB;
	data = "sync";
	break;
    case RPMEIO_FSYNC:
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB;
	data = "fsync";
	break;
    case RPMEIO_FDATASYNC:
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB;
	data = "fdatasync";
	break;
    case RPMEIO_SYNCFS:
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB;		/* XXX */
	data = "syncfs";
	break;
    case RPMEIO_MSYNC:
	req->ptr2 = va_arg(ap, void *);
	req->size = va_arg(ap, size_t);
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB;		/* XXX */
	data = "msync";
	break;
    case RPMEIO_MTOUCH:
	req->ptr2 = va_arg(ap, void *);
	req->size = va_arg(ap, size_t);
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB;		/* XXX */
	data = "mtouch";
	break;
    case RPMEIO_SYNC_FILE_RANGE:
	req->int1 = va_arg(ap, int);
	req->offs = va_arg(ap, off_t);
	req->size = va_arg(ap, size_t);
	req->int2 = va_arg(ap, int);
 	cb = rpmeioCB;		/* XXX */
	data = "sync_file_range";
	break;
    case RPMEIO_FALLOCATE:
	req->int1 = va_arg(ap, int);
	req->int2 = va_arg(ap, mode_t);
	req->offs = va_arg(ap, off_t);
	req->size = va_arg(ap, size_t);
 	cb = rpmeioCB;		/* XXX */
	data = "fallocate";
	break;
    case RPMEIO_MLOCK:
	req->ptr2 = va_arg(ap, void *);
	req->size = va_arg(ap, size_t);
 	cb = rpmeioCB;		/* XXX */
	data = "mlock";
	break;
    case RPMEIO_MLOCKALL:
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB;		/* XXX */
	data = "mlockall";
	break;
    case RPMEIO_GROUP:
 	cb = rpmeioCB;		/* XXX */
	data = "group";
	pri = EIO_PRI_MAX;	/* XXX */
	break;
    case RPMEIO_NOP:
 	cb = rpmeioCB;
	data = "nop";
	break;
    case RPMEIO_BUSY:
	req->nv1 = va_arg(ap, eio_tstamp);
 	cb = rpmeioCB;
	data = "busy";
	break;
    case RPMEIO_REALPATH:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
 	cb = rpmeioCB;		/* XXX */
	data = "realpath";
	break;
    case RPMEIO_READDIR:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->int1 = va_arg(ap, int);
 	cb = rpmeioCB_readdir;
	data = "readdir";
	break;
    case RPMEIO_OPEN:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->int1 = va_arg(ap, int);
	req->int2 = va_arg(ap, mode_t);
 	cb = rpmeioCB_open;
	data = "open";
	break;
    case RPMEIO_STAT:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
 	cb = rpmeioCB_stat;
	data = "stat";
	break;
    case RPMEIO_LSTAT:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
 	cb = rpmeioCB_stat;
	data = "lstat";
	break;
    case RPMEIO_STATVFS:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
 	cb = rpmeioCB;		/* XXX */
	data = "statvfs";
	break;
    case RPMEIO_TRUNCATE:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->offs = va_arg(ap, off_t);
 	cb = rpmeioCB;
	data = "truncate";
	break;
    case RPMEIO_UTIME:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->nv1 = va_arg(ap, eio_tstamp);
	req->nv2 = va_arg(ap, eio_tstamp);
 	cb = rpmeioCB;
	data = "utime";
	break;
    case RPMEIO_CHMOD:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->int2 = va_arg(ap, mode_t);
 	cb = rpmeioCB;		/* XXX */
	data = "chmod";
	break;
    case RPMEIO_CHOWN:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->int2 = va_arg(ap, uid_t);
	req->int3 = va_arg(ap, gid_t);
 	cb = rpmeioCB;
	data = "chown";
	break;
    case RPMEIO_UNLINK:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
 	cb = rpmeioCB;
	data = "unlink";
	break;
    case RPMEIO_RMDIR:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
 	cb = rpmeioCB;
	data = "rmdir";
	break;
    case RPMEIO_MKDIR:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->int2 = va_arg(ap, mode_t);
 	cb = rpmeioCB;
	data = "mkdir";
	break;
    case RPMEIO_RENAME:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->ptr2 = va_arg(ap, char *);
	req->ptr2 = xstrdup((char *)req->ptr2);
	req->flags |= EIO_FLAG_PTR2_FREE;
 	cb = rpmeioCB;
	data = "rename";
	break;
    case RPMEIO_MKNOD:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->int2 = va_arg(ap, mode_t);
	req->offs = va_arg(ap, dev_t);
 	cb = rpmeioCB;
	data = "mknod";
	break;
    case RPMEIO_LINK:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->ptr2 = va_arg(ap, char *);
	req->ptr2 = xstrdup((char *)req->ptr2);
	req->flags |= EIO_FLAG_PTR2_FREE;
 	cb = rpmeioCB;
	data = "link";
	break;
    case RPMEIO_SYMLINK:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
	req->ptr2 = va_arg(ap, char *);
	req->ptr2 = xstrdup((char *)req->ptr2);
	req->flags |= EIO_FLAG_PTR2_FREE;
 	cb = rpmeioCB;
	data = "symlink";
	break;
    case RPMEIO_READLINK:
	req->ptr1 = va_arg(ap, char *);
	req->ptr1 = xstrdup((char *)req->ptr1);
	req->flags |= EIO_FLAG_PTR1_FREE;
 	cb = rpmeioCB;
	data = "readlink";
	break;
    default:
assert(0);
	break;
    }

    va_end(ap);

    eio->req.pri = pri;
    eio->req.data = data;
    eio->req.finish = cb;

rpmeioDumpREQ(__FUNCTION__, req, NULL);

exit:
ZSPEW("<-- %s(%s|%p) req %p", __FUNCTION__, REQ_TYPE(req), REQ_DATA(req), req);

    return req;
}

/*==============================================================*/
eio_req *Eio_wd_open   (const char *path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_WD_OPEN, path, pri, cb, data) );
}

eio_req *Eio_wd_close  (eio_wd wd, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_WD_CLOSE, wd, pri, cb, data) );
}

eio_req *Eio_nop       (int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_NOP, pri, cb, data) );
}

eio_req *Eio_busy      (eio_tstamp delay, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_BUSY, delay, pri, cb, data) );
}

eio_req *Eio_sync      (int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_SYNC, pri, cb, data) );
}

eio_req *Eio_fsync     (int fd, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FSYNC, fd, pri, cb, data) );
}

eio_req *Eio_fdatasync (int fd, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FDATASYNC, fd, pri, cb, data) );
}

eio_req *Eio_syncfs    (int fd, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_SYNC, fd, pri, cb, data) );
}

eio_req *Eio_msync     (void *addr, size_t length, int flags, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_MSYNC, addr, length, flags, pri, cb, data) );
}

eio_req *Eio_mtouch    (void *addr, size_t length, int flags, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_MTOUCH, addr, length, flags, pri, cb, data) );
}

eio_req *Eio_mlock     (void *addr, size_t length, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_MLOCK, addr, length, pri, cb, data) );
}

eio_req *Eio_mlockall  (int flags, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_MLOCKALL, flags, pri, cb, data) );
}

eio_req *Eio_sync_file_range (int fd, off_t offset, size_t nbytes, unsigned int flags, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_SYNC_FILE_RANGE, fd, offset, nbytes, flags, pri, cb, data) );
}

eio_req *Eio_fallocate (int fd, int mode, off_t offset, size_t len, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FALLOCATE, fd, mode, offset, len, pri, cb, data) );
}

eio_req *Eio_close     (int fd, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_CLOSE, fd, pri, cb, data) );
}

eio_req *Eio_readahead (int fd, off_t offset, size_t length, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_READAHEAD, fd, offset, length, pri, cb, data) );
}

eio_req *Eio_seek      (int fd, off_t offset, int whence, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_SEEK, fd, offset, whence, pri, cb, data) );
}

eio_req *Eio_read      (int fd, void *buf, size_t length, off_t offset, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_READ, fd, buf, length, offset, pri, cb, data) );
}

eio_req *Eio_write     (int fd, void *buf, size_t length, off_t offset, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_WRITE, fd, buf, length, offset, pri, cb, data) );
}

eio_req *Eio_fstat     (int fd, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FSTAT, fd, pri, cb, data) );
}

eio_req *Eio_fstatvfs  (int fd, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FSTATVFS, fd, pri, cb, data) );
}

eio_req *Eio_futime    (int fd, eio_tstamp atime, eio_tstamp mtime, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FUTIME, fd, atime, mtime, pri, cb, data) );
}

eio_req *Eio_ftruncate (int fd, off_t offset, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FTRUNCATE, fd, offset, pri, cb, data) );
}

eio_req *Eio_fchmod    (int fd, mode_t mode, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FCHMOD, fd, mode, pri, cb, data) );
}

eio_req *Eio_fchown    (int fd, eio_uid_t uid, eio_gid_t gid, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_FCHOWN, fd, uid, gid, pri, cb, data) );
}

eio_req *Eio_dup2      (int fd, int fd2, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_DUP2, fd, fd2, pri, cb, data) );
}

eio_req *Eio_sendfile  (int out_fd, int in_fd, off_t in_offset, size_t length, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_SENDFILE, out_fd, in_fd, in_offset, length, pri, cb, data) );
}

eio_req *Eio_open      (const char *path, int flags, mode_t mode, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_OPEN, path, flags, mode, pri, cb, data) );
}

eio_req *Eio_utime     (const char *path, eio_tstamp atime, eio_tstamp mtime, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_UTIME, path, atime, mtime, pri, cb, data) );
}

eio_req *Eio_truncate  (const char *path, off_t offset, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_TRUNCATE, path, offset, pri, cb, data) );
}

eio_req *Eio_chown     (const char *path, eio_uid_t uid, eio_gid_t gid, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_CHOWN, path, uid, gid, pri, cb, data) );
}

eio_req *Eio_chmod     (const char *path, mode_t mode, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_CHMOD, path, mode, pri, cb, data) );
}

eio_req *Eio_mkdir     (const char *path, mode_t mode, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_MKDIR, path, mode, pri, cb, data) );
}

eio_req *Eio_readdir   (const char *path, int flags, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_READDIR, path, flags, pri, cb, data) );
}

eio_req *Eio_rmdir     (const char *path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_RMDIR, path, pri, cb, data) );
}

eio_req *Eio_unlink    (const char *path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_UNLINK, path, pri, cb, data) );
}

eio_req *Eio_readlink  (const char *path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_READLINK, path, pri, cb, data) );
}

eio_req *Eio_realpath  (const char *path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_REALPATH, path, pri, cb, data) );
}

eio_req *Eio_stat      (const char *path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_STAT, path, pri, cb, data) );
}

eio_req *Eio_lstat     (const char *path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_LSTAT, path, pri, cb, data) );
}

eio_req *Eio_statvfs   (const char *path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_STATVFS, path, pri, cb, data) );
}

eio_req *Eio_mknod     (const char *path, mode_t mode, dev_t dev, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_MKNOD, path, mode, dev, pri, cb, data) );
}

eio_req *Eio_link      (const char *path, const char *new_path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_LINK, path, new_path, pri, cb, data) );
}

eio_req *Eio_symlink   (const char *path, const char *new_path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_SYMLINK, path, new_path, pri, cb, data) );
}

eio_req *Eio_rename    (const char *path, const char *new_path, int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_RENAME, path, new_path, pri, cb, data) );
}

eio_req *Eio_custom    (void (*execute)(eio_req *), int pri, eio_cb cb, void *data)
{
    return Eio_submit( Eio_request(RPMEIO_CUSTOM, execute, pri, cb, data) );
}

/*==============================================================*/
static void rpmeioFini(void *_eio)
{
    rpmeio eio = (rpmeio) _eio;
    if (eio) {

#ifdef	NOTYET
	if (eio->req.flags & EIO_FLAG_PTR1_FREE)
	    eio->req.ptr1 = _free(eio->req.ptr1);
	if (eio->req.flags & EIO_FLAG_PTR2_FREE)
	    eio->req.ptr2 = _free(eio->req.ptr2);
#endif
	eio->req.ptr1 = NULL;
	eio->req.ptr2 = NULL;

	eio->req.next = NULL;
	eio->req.wd = NULL;
	eio->req.result = 0;
	eio->req.offs = 0;
	eio->req.size = 0;
	eio->req.ptr1 = NULL;
	eio->req.ptr2 = NULL;
	eio->req.nv1 = 0.0;
	eio->req.nv2 = 0.0;
	eio->req.int1 = 0;
	eio->req.int2 = 0;
	eio->req.int3 = 0;
	eio->req.errorno = 0;

	eio->req.type = 0;
	eio->req.cancelled = 0;
	eio->req.flags = 0;
	eio->req.pri = 0;

	eio->req.data = NULL;
	eio->req.finish = NULL;
	eio->req.destroy = NULL;

	eio->req.feed = NULL;
	eio->req.grp = NULL;
	eio->req.grp_prev = NULL;
	eio->req.grp_next = NULL;
	eio->req.grp_first = NULL;

	eio->evl = NULL;

    }
}

#define	rpmeioDbug	NULL
#define	rpmeioInit	NULL
RPMIOPOOL_MODULE(eio)

rpmeio rpmeioNew(rpmeioType reqtype, ...)
{
    rpmeio eio = rpmeioGetPool(_rpmeioPool);

	eio->req.next = NULL;
	eio->req.wd = NULL;
	eio->req.result = 0;
	eio->req.offs = 0;
	eio->req.size = 0;
	eio->req.ptr1 = NULL;
	eio->req.ptr2 = NULL;
	eio->req.nv1 = 0.0;
	eio->req.nv2 = 0.0;
	eio->req.int1 = 0;
	eio->req.int2 = 0;
	eio->req.int3 = 0;
	eio->req.errorno = 0;

    eio->req.type = reqtype;
	eio->req.cancelled = 0;
	eio->req.flags = 0;
    eio->req.pri = 0;

    eio->req.data = NULL;
    eio->req.finish = NULL;
    eio->req.destroy = Eio_destroy;

	eio->req.feed = NULL;
	eio->req.grp = NULL;
	eio->req.grp_prev = NULL;
	eio->req.grp_next = NULL;
	eio->req.grp_first = NULL;

    eio->evl = (void *) rpmevlGet();

    return rpmeioLink(eio);
}
