#ifndef	_H_RPMEIO_
#define	_H_RPMEIO_

/**
 */
extern int _rpmeio_debug;

typedef	struct rpmeio_s * rpmeio;

typedef struct rpmevl_s * rpmevl;

/**
 */
typedef enum rpmeioType_e {
    RPMEIO_CUSTOM	=  0,	/* EIO_CUSTOM */
    RPMEIO_WD_OPEN	=  1,	/* EIO_WD_OPEN */
    RPMEIO_WD_CLOSE	=  2,	/* EIO_WD_CLOSE */
    RPMEIO_CLOSE	=  3,	/* EIO_CLOSE */
    RPMEIO_DUP2		=  4,	/* EIO_DUP2 */
    RPMEIO_SEEK		=  5,	/* EIO_SEEK */
    RPMEIO_READ		=  6,	/* EIO_READ */
    RPMEIO_WRITE	=  7,	/* EIO_WRITE */
    RPMEIO_READAHEAD	=  8,	/* EIO_READAHEAD */
    RPMEIO_SENDFILE	=  9,	/* EIO_SENDFILE */
    RPMEIO_FSTAT	= 10,	/* EIO_FSTAT */
    RPMEIO_FSTATVFS	= 11,	/* EIO_FSTATVFS */
    RPMEIO_FTRUNCATE	= 12,	/* EIO_FTRUNCATE */
    RPMEIO_FUTIME	= 13,	/* EIO_FUTIME */
    RPMEIO_FCHMOD	= 14,	/* EIO_FCHMOD */
    RPMEIO_FCHOWN	= 15,	/* EIO_FCHOWN */
    RPMEIO_SYNC		= 16,	/* EIO_SYNC */
    RPMEIO_FSYNC	= 17,	/* EIO_FSYNC */
    RPMEIO_FDATASYNC	= 18,	/* EIO_FDATASYNC */
    RPMEIO_SYNCFS	= 19,	/* EIO_SYNCFS */
    RPMEIO_MSYNC	= 20,	/* EIO_MSYNC */
    RPMEIO_MTOUCH	= 21,	/* EIO_MTOUCH */
    RPMEIO_SYNC_FILE_RANGE = 22,	/* EIO_SYNC_FILE_RANGE */
    RPMEIO_FALLOCATE	= 23,	/* EIO_FALLOCATE */
    RPMEIO_MLOCK	= 24,	/* EIO_MLOCK */
    RPMEIO_MLOCKALL	= 25,	/* EIO_MLOCKALL */
    RPMEIO_GROUP	= 26,	/* EIO_GROUP */
    RPMEIO_NOP		= 27,	/* EIO_NOP */
    RPMEIO_BUSY		= 28,	/* EIO_BUSY */
    RPMEIO_REALPATH	= 29,	/* EIO_REALPATH */
    RPMEIO_READDIR	= 30,	/* EIO_READDIR */
    RPMEIO_OPEN		= 31,	/* EIO_OPEN */
    RPMEIO_STAT		= 32,	/* EIO_STAT */
    RPMEIO_LSTAT	= 33,	/* EIO_LSTAT */
    RPMEIO_STATVFS	= 34,	/* EIO_STATVFS */
    RPMEIO_TRUNCATE	= 35,	/* EIO_TRUNCATE */
    RPMEIO_UTIME	= 36,	/* EIO_UTIME */
    RPMEIO_CHMOD	= 37,	/* EIO_CHMOD */
    RPMEIO_CHOWN	= 38,	/* EIO_CHOWN */
    RPMEIO_UNLINK	= 39,	/* EIO_UNLINK */
    RPMEIO_RMDIR	= 40,	/* EIO_RMDIR */
    RPMEIO_MKDIR	= 41,	/* EIO_MKDIR */
    RPMEIO_RENAME	= 42,	/* EIO_RENAME */
    RPMEIO_MKNOD	= 43,	/* EIO_MKNOD */
    RPMEIO_LINK		= 44,	/* EIO_LINK */
    RPMEIO_SYMLINK	= 45,	/* EIO_SYMLINK */
    RPMEIO_READLINK	= 46,	/* EIO_READLINK */
    RPMEIO_REQ_TYPE_NUM	= 47,	/* EIO_REQ_TYPE_NUM */
} rpmeioType;
#define	RPMEIO_REQ_TYPE_MASK	0xff

/**
 */
#define _KFB(n) (1U << (n))
#define _MFB(n) (_KFB(n))
#define	RPMEIO_FLAGS_TYPE_SHIFT	23
typedef enum rpmeioFlags_e {
    RPMEIO_FLAGS_NONE		= 0,
	/* bits 1-31 reserved */
} rpmeioFlags;
#undef  _MFB
#undef  _KFB

#ifdef __cplusplus
extern "C" {
#endif

/**
 */
#if defined(_RPMEIO_INTERNAL)

struct rpmeio_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    eio_req req;
    void * evl;
};

RPM_GNUC_CONST
extern const char * rpmeioReqType(int reqtype);
#define REQ_DATA(_req)  (_req->data)
#define REQ_TYPE(_req)  (rpmeioReqType(_req->type))

extern void rpmeioDumpREQ(const char *msg, eio_req * req, FILE *fp);

rpmevl rpmeioStart(double max_poll_time, unsigned max_poll_reqs,
		unsigned min_parallel, unsigned max_parallel,
		unsigned max_idle, unsigned idle_timeout, unsigned flags);
#define	RPMEVL_PIPE	0x1

void rpmeioLoop(rpmevl evl);
void rpmeioDrain(rpmevl evl);

void Eio_cancel(eio_req * req);
void Eio_destroy(eio_req * req);

eio_req * Eio_request(int reqtype, ...);
eio_req * Eio_submit(eio_req * req);

int rpmeioCB(eio_req * req);
int rpmeioCB_readdir(eio_req * req);
int rpmeioCB_stat(eio_req * req);
int rpmeioCB_read(eio_req * req);

int rpmeioLastFdno;
int rpmeioCB_open(eio_req * req);

eio_req *Eio_wd_open   (const char *path, int pri, eio_cb cb, void *data); /* result=wd */

eio_req *Eio_wd_close  (eio_wd wd, int pri, eio_cb cb, void *data);

eio_req *Eio_nop       (int pri, eio_cb cb, void *data); /* does nothing except go through the whole process */

eio_req *Eio_busy      (eio_tstamp delay, int pri, eio_cb cb, void *data); /* ties a thread for this long, simulating busyness */

eio_req *Eio_sync      (int pri, eio_cb cb, void *data);

eio_req *Eio_fsync     (int fd, int pri, eio_cb cb, void *data);

eio_req *Eio_fdatasync (int fd, int pri, eio_cb cb, void *data);

eio_req *Eio_syncfs    (int fd, int pri, eio_cb cb, void *data);

eio_req *Eio_msync     (void *addr, size_t length, int flags, int pri, eio_cb cb, void *data);

eio_req *Eio_mtouch    (void *addr, size_t length, int flags, int pri, eio_cb cb, void *data);

eio_req *Eio_mlock     (void *addr, size_t length, int pri, eio_cb cb, void *data);

eio_req *Eio_mlockall  (int flags, int pri, eio_cb cb, void *data);

eio_req *Eio_sync_file_range (int fd, off_t offset, size_t nbytes, unsigned int flags, int pri, eio_cb cb, void *data);

eio_req *Eio_fallocate (int fd, int mode, off_t offset, size_t len, int pri, eio_cb cb, void *data);

eio_req *Eio_close     (int fd, int pri, eio_cb cb, void *data);

eio_req *Eio_readahead (int fd, off_t offset, size_t length, int pri, eio_cb cb, void *data);

eio_req *Eio_seek      (int fd, off_t offset, int whence, int pri, eio_cb cb, void *data);

eio_req *Eio_read      (int fd, void *buf, size_t length, off_t offset, int pri, eio_cb cb, void *data);

eio_req *Eio_write     (int fd, void *buf, size_t length, off_t offset, int pri, eio_cb cb, void *data);

eio_req *Eio_fstat     (int fd, int pri, eio_cb cb, void *data); /* stat buffer=ptr2 allocated dynamically */

eio_req *Eio_fstatvfs  (int fd, int pri, eio_cb cb, void *data); /* stat buffer=ptr2 allocated dynamically */

eio_req *Eio_futime    (int fd, eio_tstamp atime, eio_tstamp mtime, int pri, eio_cb cb, void *data);

eio_req *Eio_ftruncate (int fd, off_t offset, int pri, eio_cb cb, void *data);

eio_req *Eio_fchmod    (int fd, mode_t mode, int pri, eio_cb cb, void *data);

eio_req *Eio_fchown    (int fd, eio_uid_t uid, eio_gid_t gid, int pri, eio_cb cb, void *data);

eio_req *Eio_dup2      (int fd, int fd2, int pri, eio_cb cb, void *data);

eio_req *Eio_sendfile  (int out_fd, int in_fd, off_t in_offset, size_t length, int pri, eio_cb cb, void *data);

eio_req *Eio_open      (const char *path, int flags, mode_t mode, int pri, eio_cb cb, void *data);

eio_req *Eio_utime     (const char *path, eio_tstamp atime, eio_tstamp mtime, int pri, eio_cb cb, void *data);

eio_req *Eio_truncate  (const char *path, off_t offset, int pri, eio_cb cb, void *data);

eio_req *Eio_chown     (const char *path, eio_uid_t uid, eio_gid_t gid, int pri, eio_cb cb, void *data);

eio_req *Eio_chmod     (const char *path, mode_t mode, int pri, eio_cb cb, void *data);

eio_req *Eio_mkdir     (const char *path, mode_t mode, int pri, eio_cb cb, void *data);

eio_req *Eio_readdir   (const char *path, int flags, int pri, eio_cb cb, void *data); /* result=ptr2 allocated dynamically */

eio_req *Eio_rmdir     (const char *path, int pri, eio_cb cb, void *data);

eio_req *Eio_unlink    (const char *path, int pri, eio_cb cb, void *data);

eio_req *Eio_readlink  (const char *path, int pri, eio_cb cb, void *data); /* result=ptr2 allocated dynamically */

eio_req *Eio_realpath  (const char *path, int pri, eio_cb cb, void *data); /* result=ptr2 allocated dynamically */

eio_req *Eio_stat      (const char *path, int pri, eio_cb cb, void *data); /* stat buffer=ptr2 allocated dynamically */

eio_req *Eio_lstat     (const char *path, int pri, eio_cb cb, void *data); /* stat buffer=ptr2 allocated dynamically */

eio_req *Eio_statvfs   (const char *path, int pri, eio_cb cb, void *data); /* stat buffer=ptr2 allocated dynamically */

eio_req *Eio_mknod     (const char *path, mode_t mode, dev_t dev, int pri, eio_cb cb, void *data);

eio_req *Eio_link      (const char *path, const char *new_path, int pri, eio_cb cb, void *data);

eio_req *Eio_symlink   (const char *path, const char *new_path, int pri, eio_cb cb, void *data);

eio_req *Eio_rename    (const char *path, const char *new_path, int pri, eio_cb cb, void *data);

#ifdef	NOTYET
eio_req *Eio_custom    (void (*execute)(eio_req *), int pri, eio_cb cb, void *data);
#endif	/* NOTYET */

#endif	/* _RPMEIO_INTERNAL */

/**
 * Unreference a eio wrapper instance.
 * @param eio		eio wrapper
 * @return		NULL on last dereference
 */
rpmeio rpmeioUnlink (rpmeio eio);
#define	rpmeioUnlink(_eio)	\
    ((rpmeio)rpmioUnlinkPoolItem((rpmioItem)(_eio), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a eio wrapper instance.
 * @param eio		eio wrapper
 * @return		new eio wrapper reference
 */
rpmeio rpmeioLink (rpmeio eio);
#define	rpmeioLink(_eio)	\
    ((rpmeio)rpmioLinkPoolItem((rpmioItem)(_eio), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a eio wrapper.
 * @param eio		eio wrapper
 * @return		NULL on last dereference
 */
rpmeio rpmeioFree(rpmeio eio);
#define	rpmeioFree(_eio)	\
    ((rpmeio)rpmioFreePoolItem((rpmioItem)(_eio), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create a eio wrapper.
 * @return		eio wrapper
 */
rpmeio rpmeioNew(rpmeioType reqtype, ...);

#ifdef __cplusplus
}
#endif

#endif	/* _H_RPMEIO_ */
