#include "system.h"
#include <stdarg.h>

#if defined(WITH_LIBEV)

#if defined(HAVE_EV_H)
# include <ev.h>
#endif

#else	/* WITH_LIBEV */

/* XXX stub in enough to include rpmev.h */
#define	EV_P_	void * loop,
struct ev_watcher {
    int active;
    int pending;
    int priority;
    void * data;
    void (*cb)(EV_P_ void *w, int revents);
};

union ev_any_watcher
{
    struct ev_watcher w;
};

#endif	/* WITH_LIBEV */

#include <rpmio.h>
#include <poptIO.h>

#define	_RPMEV_INTERNAL
#include <rpmev.h>

#include "debug.h"

int _rpmev_debug = -1;

#define SPEW(_fmt, ...) \
    if (_rpmev_debug || _rpmio_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

/*==============================================================*/
unsigned rpmioEV = 0;
struct poptOption rpmioEVTable[] = {
#if defined(WITH_LIBEV)
 { "read", '\0', POPT_BIT_SET,		&rpmioEV, EV_READ,
        N_("read"), N_("EV_READ") },
 { "write", '\0', POPT_BIT_SET,		&rpmioEV, EV_WRITE,
        N_("write"), N_("EV_WRITE") },
 { "internal", '\0', POPT_BIT_SET,	&rpmioEV, EV__IOFDSET,
        N_("internal"), N_("EV__IOFDSET") },
 { "timer", '\0', POPT_BIT_SET,		&rpmioEV, EV_TIMER,
        N_("timer"), N_("EV_TIMER") },
 { "periodic", '\0', POPT_BIT_SET,	&rpmioEV, EV_PERIODIC,
        N_("periodic"), N_("EV_PERIODIC") },
 { "signal", '\0', POPT_BIT_SET,	&rpmioEV, EV_SIGNAL,
        N_("signal"), N_("EV_SIGNAL") },
 { "child", '\0', POPT_BIT_SET,		&rpmioEV, EV_CHILD,
        N_("child"), N_("EV_CHILD") },
 { "stat", '\0', POPT_BIT_SET,		&rpmioEV, EV_STAT,
        N_("stat"), N_("EV_STAT") },
 { "idle", '\0', POPT_BIT_SET,		&rpmioEV, EV_IDLE,
        N_("idle"), N_("EV_IDLE") },
 { "prepare", '\0', POPT_BIT_SET,	&rpmioEV, EV_PREPARE,
        N_("prepare"), N_("EV_PREPARE") },
 { "check", '\0', POPT_BIT_SET,		&rpmioEV, EV_CHECK,
        N_("check"), N_("EV_CHECK") },
 { "embed", '\0', POPT_BIT_SET,		&rpmioEV, EV_EMBED,
        N_("embed"), N_("EV_EMBED") },
 { "fork", '\0', POPT_BIT_SET,		&rpmioEV, EV_FORK,
        N_("fork"), N_("EV_FORK") },
 { "cleanup", '\0', POPT_BIT_SET,	&rpmioEV, EV_CLEANUP,
        N_("cleanup"), N_("EV_CLEANUP") },
 { "async", '\0', POPT_BIT_SET,		&rpmioEV, EV_ASYNC,
        N_("async"), N_("EV_ASYNC") },
 { "custom", '\0', POPT_BIT_SET,	&rpmioEV, EV_CUSTOM,
        N_("custom"), N_("EV_CUSTOM") },
 { "error", '\0', POPT_BIT_SET,		&rpmioEV, EV_ERROR,
        N_("error"), N_("EV_ERROR") },
#endif	/* WITH_LIBEV */
   POPT_TABLEEND
};

unsigned rpmioEVFLAG = 0;
struct poptOption rpmioEVFLAGTable[] = {
#if defined(WITH_LIBEV)
 { "env", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE,	&rpmioEVFLAG, EVFLAG_NOENV,
        N_("env"), N_("EVFLAG_NOENV") },
 { "forkcheck", '\0', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE,	&rpmioEVFLAG, EVFLAG_FORKCHECK,
        N_("forkcheck"), N_("EVFLAG_FORKCHECK") },
 { "inotify", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE, &rpmioEVFLAG, EVFLAG_NOINOTIFY,
        N_("inotify"), N_("EVFLAG_NOINOTIFY") },
 { "signalfd", '\0', POPT_BIT_SET|POPT_ARGFLAG_TOGGLE, &rpmioEVFLAG, EVFLAG_SIGNALFD,
        N_("signalfd"), N_("EVFLAG_SIGNALFD") },
 { "sigmask", '\0', POPT_BIT_CLR|POPT_ARGFLAG_TOGGLE, &rpmioEVFLAG, EVFLAG_NOSIGMASK,
        N_("sigmask"), N_("EVFLAG_NOSIGMASK") },
#endif	/* WITH_LIBEV */
   POPT_TABLEEND
};

unsigned rpmioEVBACKEND = 0;
struct poptOption rpmioEVBACKENDTable[] = {
#if defined(WITH_LIBEV)
 { "select", '\0', POPT_BIT_SET,	&rpmioEVBACKEND, EVBACKEND_SELECT,
        N_("select"), N_("EVBACKEND_SELECT") },
 { "poll", '\0', POPT_BIT_SET,		&rpmioEVBACKEND, EVBACKEND_POLL,
        N_("poll"), N_("EVBACKEND_POLL") },
 { "epoll", '\0', POPT_BIT_SET,		&rpmioEVBACKEND, EVBACKEND_EPOLL,
        N_("epoll"), N_("EVBACKEND_EPOLL") },
 { "kqueue", '\0', POPT_BIT_SET,	&rpmioEVBACKEND, EVBACKEND_KQUEUE,
        N_("kqueue"), N_("EVBACKEND_KQUEUE") },
 { "devpoll", '\0', POPT_BIT_SET,	&rpmioEVBACKEND, EVBACKEND_DEVPOLL,
        N_("devpoll"), N_("EVBACKEND_DEVPOLL") },
 { "port", '\0', POPT_BIT_SET,		&rpmioEVBACKEND, EVBACKEND_PORT,
        N_("port"), N_("EVBACKEND_PORT") },
#endif	/* WITH_LIBEV */
   POPT_TABLEEND
};

/*==============================================================*/
/* every watcher type has its own typedef'd struct */
/* with the name ev_TYPE */

static rpmioItem rpmevqHead = NULL;
static rpmioItem * rpmevqTail = &rpmevqHead;

rpmev rpmevqGetHead(void)
{
    rpmioItem item;
    RPM_GNUC_TM_ATOMIC {
        if ((item = rpmevqHead) != NULL) {
            rpmevqHead = item->next;
            item->next = NULL;
            if (rpmevqHead == NULL)
                rpmevqTail = &rpmevqHead;
        }
    }
if (item)
SPEW("<--\t%s() rc %p\n", __FUNCTION__, item);
    return (rpmev) item;
}

void rpmevqPutTail(rpmev ev)
{
SPEW("-->\t%s(%p)\n", __FUNCTION__, ev);
    if (ev == NULL)
	return;
    rpmioItem item = &ev->_item;
    item->next = NULL;
    RPM_GNUC_TM_ATOMIC {
	*rpmevqTail = item;
        rpmevqTail = (rpmioItem *) &item->next;
    }
}

void rpmevqPutHead(rpmev ev)
{
SPEW("-->\t%s(%p)\n", __FUNCTION__, ev);
    if (ev == NULL)
	return;
    rpmioItem item = &ev->_item;
    RPM_GNUC_TM_ATOMIC {
	item->next = rpmevqHead->next;
	if (item->next == NULL)
	    rpmevqTail = (rpmioItem *) &item->next;
	rpmevqHead->next = item;
    }
}

static void rpmevqRemove(rpmev ev)
{
SPEW("-->\t%s(%p)\n", __FUNCTION__, ev);
    if (ev == NULL)
	return;
    rpmioItem item = &ev->_item;
    rpmioItem * prev = &rpmevqHead;
    while (prev && (*prev)->next != NULL) {
	if ((*prev)->next == item) {
	    (*prev)->next = item->next;
	    item->next = NULL;
	    if (rpmevqHead == NULL)
		rpmevqTail = &rpmevqHead;
	    break;
	}
	prev = (*prev)->next;
    }
}

/* all watcher callbacks have a similar signature */

/* --- */
static
void rpmevChild_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_child * w = (ev_child *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_CHILD);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s) pid %u status %u (0x%x)\n", __FUNCTION__, w, rstr, w->rpid, __WEXITSTATUS(w->rstatus), w->rstatus);
    rstr = _free(rstr);
  }
    ev_child_stop (EV_A_ w);
#ifdef	BUGGY
    ev = rpmevFree(ev);		/* XXX deallocate w rpmevRemove(ev)? */
#endif
#endif	/* WITH_LIBEV */
}

static
int rpmevStart(void)
{
    int rc;
SPEW("--> %s()\n", __FUNCTION__);
    pid_t pid = fork();
    if (pid < 0) {
	rc = -1;
	exit(rc);
    } else if (pid == 0) {
	sleep(2);
	rc = 42;
	exit(rc);
    } else {
	rpmev ev = rpmevNew(RPMEV_TYPE_CHILD, NULL, pid, 0);
	(void)ev;
#ifdef	DYING
	ev_child * w = &ev->evw.child;
	w->pid = 0;
	w->flags = 0;	/** XXX 0=terminate, 1=stop/continue */
	ev_child_start (EV_DEFAULT_ w);
#endif
	rc = 0;
    }
SPEW("<-- %s() rc %d\n", __FUNCTION__, rc);
    return rc;
}

/* --- */
static void rpmevIo_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_io * w = (ev_io *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_IO);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s) fd %d events 0x%x\n", __FUNCTION__, w, rstr, w->fd, w->events);
    rstr = _free(rstr);
  }
    /* for one-shot events, one must manually stop the watcher */
    /* with its corresponding stop function. */
    ev_io_stop(EV_A_ w);

    /* this causes all nested ev_loop's to stop iterating */
    ev_unloop(EV_A_ EVUNLOOP_ALL);
#endif	/* WITH_LIBEV */
}
/* --- */
static void rpmevTimer_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_timer * w = (ev_timer *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_TIMER);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s) repeat %f\n", __FUNCTION__, w, rstr, w->repeat);
    rstr = _free(rstr);
  }
#ifdef	DYING
    /* this causes the innermost ev_loop to stop iterating */
    ev_unloop(EV_A_ EVUNLOOP_ONE);
#endif
    w->repeat = 5.5;
    ev_timer_again(EV_A_ w);
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevPeriodic_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_periodic * w = (ev_periodic *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_PERIODIC);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s) offset %f interval %f reschedule_cb %p\n", __FUNCTION__, w, rstr, w->offset, w->interval, w->reschedule_cb);
    rstr = _free(rstr);
  }
    rpmevStart();
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevSignal_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_signal * w = (ev_signal *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_SIGNAL);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s) signum %d\n", __FUNCTION__, w, rstr, w->signum);
    rstr = _free(rstr);
  }
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevStat_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_stat * w = (ev_stat *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_STAT);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s) path %s wd %d\n", __FUNCTION__, w, rstr, w->path, w->wd);
    rstr = _free(rstr);
  }
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevIdle_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_idle * w = (ev_idle *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_IDLE);
#ifdef	NOISY
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s)\n", __FUNCTION__, w, rstr);
    rstr = _free(rstr);
  }
#endif
    struct timespec ts = { 0, 10*1000*1000 };
    nanosleep(&ts, NULL);
    sched_yield();
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevPrepare_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_prepare * w = (ev_prepare *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_PREPARE);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s)\n", __FUNCTION__, w, rstr);
    rstr = _free(rstr);
  }
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevCheck_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_check * w = (ev_check *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_CHECK);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s)\n", __FUNCTION__, w, rstr);
    rstr = _free(rstr);
  }
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevFork_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_fork * w = (ev_fork *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_FORK);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s)\n", __FUNCTION__, w, rstr);
    rstr = _free(rstr);
  }
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevCleanup_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_cleanup * w = (ev_cleanup *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_CLEANUP);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s)\n", __FUNCTION__, w, rstr);
    rstr = _free(rstr);
  }
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevEmbed_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_embed * w = (ev_embed *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_EMBED);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s)\n", __FUNCTION__, w, rstr);
    rstr = _free(rstr);
  }
#endif	/* WITH_LIBEV */
}

/* --- */
static void rpmevAsync_cb(EV_P_ void * _w, int revents)
{
#if defined(WITH_LIBEV)
    ev_async * w = (ev_async *) _w;
    rpmev ev = (rpmev) w->data;
assert(ev && ev->evtype == RPMEV_TYPE_ASYNC);
  { char *rstr = rpmioB2N(rpmioEVTable, revents);
    SPEW("--> %s(%p, %s)\n", __FUNCTION__, w, rstr);
    rstr = _free(rstr);
  }
#endif	/* WITH_LIBEV */
}

void rpmevDump(const char *msg, rpmev ev, FILE *fp)
{
    if (fp == NULL)	fp = stderr;
    if (msg) fprintf(stderr, "========================= %s(%p)\n", msg, ev);
    if (ev) {
#define PRINT(_fmt, _foo) \
    {   fprintf(fp, "%25s: %"#_fmt"\n", #_foo, ev->_foo); }
	PRINT(u, evtype);
	PRINT(d, evw.w.active);
	PRINT(d, evw.w.pending);
	PRINT(p, evw.w.data);
	PRINT(p, evw.w.cb);
	PRINT(p, _item.next);
fprintf(stderr, "%25s: &%p next %p\n", "head", &rpmevqHead, (rpmevqHead ? rpmevqHead->next : NULL));
fprintf(stderr, "%25s:  %p next %p\n", "tail", rpmevqTail, (rpmevqTail ? *rpmevqTail : NULL));
#undef	PRINT

    }
}

static void rpmevFini(void *_ev)
{
    rpmev ev = (rpmev) _ev;
    if (ev) {

#ifdef	BUGGY
	/* If still linked into thge allocation chain, unlink. */
	if (ev->_item.next)
	    rpmevqRemove(ev);
#endif

#if defined(WITH_LIBEV)
	/* XXX ev->loop private loops? */
	struct ev_loop *loop = ev_default_loop_uc_();

	if (loop)
	switch (ev->evtype) {
	default:
	case RPMEV_TYPE_UNKNOWN:
	{   ev_watcher * w = &ev->evw.w;
	    (void *)w;
	}   break;
	case RPMEV_TYPE_IO:
	{   ev_io * w = &ev->evw.io;
	    ev_io_stop(loop, w);
	}   break;
	case RPMEV_TYPE_TIMER:
	{   ev_timer * w = &ev->evw.timer;
	    ev_timer_stop(loop, w);
	}   break;
	case RPMEV_TYPE_PERIODIC:
	{   ev_periodic * w = &ev->evw.periodic;
	    ev_periodic_stop(loop, w);
	}   break;
	case RPMEV_TYPE_SIGNAL:
	{   ev_signal * w = &ev->evw.signal;
	    ev_signal_stop(loop, w);
	}   break;
	case RPMEV_TYPE_CHILD:
	{   ev_child * w = &ev->evw.child;
	    ev_child_stop(loop, w);
	}   break;
	case RPMEV_TYPE_STAT:
	{   ev_stat * w = &ev->evw.stat;
	    ev_stat_stop(loop, w);
	    w->path = _free(w->path);
	}   break;
	case RPMEV_TYPE_IDLE:
	{   ev_idle * w = &ev->evw.idle;
	    ev_idle_stop(loop, w);
	}   break;
	case RPMEV_TYPE_PREPARE:
	{   ev_prepare * w = &ev->evw.prepare;
	    ev_prepare_stop(loop, w);
	}   break;
	case RPMEV_TYPE_CHECK:
	{   ev_check * w = &ev->evw.check;
	    ev_check_stop(loop, w);
	}   break;
	case RPMEV_TYPE_FORK:
	{   ev_fork * w = &ev->evw.fork;
	    ev_fork_stop(loop, w);
	}   break;
	case RPMEV_TYPE_CLEANUP:
	{   ev_cleanup * w = &ev->evw.cleanup;
	    ev_cleanup_stop(loop, w);
	}   break;
	case RPMEV_TYPE_EMBED:
	{   ev_embed * w = &ev->evw.embed;
	    ev_embed_stop(loop, w);
	}   break;
	case RPMEV_TYPE_ASYNC:
	{   ev_async * w = &ev->evw.async;
	    ev_async_stop(loop, w);
	}   break;
	}
#endif	/* WITH_LIBEV */

	/* XXX stop/destroyiterate loop? */
	ev->evtype = 0;
    }
}

#define	rpmevDbug	NULL
#define	rpmevInit	NULL
RPMIOPOOL_MODULE(ev)

rpmev rpmevNew(rpmevType evtype, void * _evcb, ...)
{
    rpmev ev = rpmevGetPool(_rpmevPool);

    ev->evtype = evtype;

#if defined(WITH_LIBEV)
    struct ev_loop *loop = ev_default_loop(EVFLAG_AUTO);
    va_list ap;

    /* Initialize the base class. */
    ev_watcher * watcher = &ev->evw.w;
    ev_init(watcher, _evcb);
    watcher->data = ev;

    /* Process args and start the event. */
    va_start(ap, _evcb);

    switch (evtype) {
    default:
    case RPMEV_TYPE_UNKNOWN:
    {   ev_watcher * w = &ev->evw.w;
	(void *)w;	/* XXX NOTYET */
    }	break;
    case RPMEV_TYPE_IO:
    {	ev_io * w = &ev->evw.io;
	w->cb = (_evcb ? _evcb : rpmevIo_cb);
	w->fd = va_arg(ap, int);
	w->events = va_arg(ap, int);
	ev_io_start(loop, w);
    }	break;
    case RPMEV_TYPE_TIMER:
    {	ev_timer * w = &ev->evw.timer;
	w->cb = (_evcb ? _evcb : rpmevTimer_cb);
	w->at = va_arg(ap, ev_tstamp);
	w->repeat = va_arg(ap, ev_tstamp);
	ev_timer_start(loop, w);
    }	break;
    case RPMEV_TYPE_PERIODIC:
    {	ev_periodic * w = &ev->evw.periodic;
	w->cb = (_evcb ? _evcb : rpmevPeriodic_cb);
	w->offset = va_arg(ap, ev_tstamp);
	w->interval = va_arg(ap, ev_tstamp);
	w->reschedule_cb = va_arg(ap, void *);
	ev_periodic_start(loop, w);
    }	break;
    case RPMEV_TYPE_SIGNAL:
    {	ev_signal * w = &ev->evw.signal;
	w->cb = (_evcb ? _evcb : rpmevSignal_cb);
	w->signum = va_arg(ap, int);
	ev_signal_start(loop, w);
    }	break;
    case RPMEV_TYPE_CHILD:
    {	ev_child * w = &ev->evw.child;
	w->cb = (_evcb ? _evcb : rpmevChild_cb);
	w->flags = va_arg(ap, int);
	w->pid = va_arg(ap, int);
	ev_child_start (EV_DEFAULT_ w);		/* XXX only main */
    }	break;
    case RPMEV_TYPE_STAT:
    {	ev_stat * w = &ev->evw.stat;
	w->cb = (_evcb ? _evcb : rpmevStat_cb);
	const char *path = va_arg(ap, const char *);
	w->path = (path ? xstrdup(path) : NULL);
	w->interval = va_arg(ap, ev_tstamp);
	w->wd = -2;
	ev_stat_start(loop, w);
    }	break;
    case RPMEV_TYPE_IDLE:
    {	ev_idle * w = &ev->evw.idle;
	w->cb = (_evcb ? _evcb : rpmevIdle_cb);
rpmevDump("IDLE", ev, NULL);
	ev_idle_start(loop, w);
    }	break;
    case RPMEV_TYPE_PREPARE:
    {	ev_prepare * w = &ev->evw.prepare;
	w->cb = (_evcb ? _evcb : rpmevPrepare_cb);
	ev_prepare_start(loop, w);
    }	break;
    case RPMEV_TYPE_CHECK:
    {	ev_check * w = &ev->evw.check;
	w->cb = (_evcb ? _evcb : rpmevCheck_cb);
	ev_check_start(loop, w);
    }	break;
    case RPMEV_TYPE_FORK:
    {	ev_fork * w = &ev->evw.fork;
	w->cb = (_evcb ? _evcb : rpmevFork_cb);
	ev_fork_start(loop, w);
    }	break;
    case RPMEV_TYPE_CLEANUP:
    {	ev_cleanup * w = &ev->evw.cleanup;
	w->cb = (_evcb ? _evcb : rpmevCleanup_cb);
	ev_cleanup_start(loop, w);
    }	break;
    case RPMEV_TYPE_EMBED:
    {	ev_embed * w = &ev->evw.embed;
	w->cb = (_evcb ? _evcb : rpmevEmbed_cb);
	w->other = va_arg(ap, void *);
	ev_embed_start(loop, w);
    }	break;
    case RPMEV_TYPE_ASYNC:
    {	ev_async * w = &ev->evw.async;
	w->cb = (_evcb ? _evcb : rpmevAsync_cb);
rpmevDump("ASYNC", ev, NULL);
	ev_async_start(loop, w);
    }	break;
    }

    va_end(ap);

    /* XXX ev_start? */

    /* Add to allocated event chain. */
    rpmevqPutTail(ev);
#endif	/* WITH_LIBEV */

    return rpmevLink(ev);
}

