#ifndef	H_RPMIO_INTERNAL
#define	H_RPMIO_INTERNAL

/** \ingroup rpmio
 * \file rpmio/rpmio_internal.h
 */

#include <rpmiotypes.h>
#include <rpmlog.h>
#include <rpmio.h>
#include <rpmurl.h>

#define	_RPMPGP_INTERNAL
#include <rpmpgp.h>

#include <rpmxar.h>

/** \ingroup rpmio
 */
typedef struct _FDSTACK_s {
    FDIO_t		io;
    void *		fp;
    int			fdno;
} FDSTACK_t;

/** \ingroup rpmio
 * Identify per-desciptor I/O operation statistics.
 */
typedef enum fdOpX_e {
    FDSTAT_READ		=  0,	/*!< read statistics index. */
    FDSTAT_WRITE	=  1,	/*!< write statistics index. */
    FDSTAT_SEEK		=  2,	/*!< seek statistics index. */
    FDSTAT_CLOSE	=  3,	/*!< close statistics index */
    FDSTAT_FDIGEST	=  4,	/*!< digest statistics index. */
    FDSTAT_FOPEN	=  5,	/*!< open statistics index. */
    FDSTAT_FCNTL	=  6,	/*!< fcntl statistics index. */
    FDSTAT_FLOCK	=  7,	/*!< flock statistics index. */
    FDSTAT_FCHDIR	=  8,	/*!< fchdir statistics index. */
    FDSTAT_FCHMOD	=  9,	/*!< fchmod statistics index. */
    FDSTAT_FCHOWN	= 10,	/*!< fchown statistics index. */
    FDSTAT_FUTIMES	= 11,	/*!< futimes statistics index. */
    FDSTAT_FUTIMENS	= 12,	/*!< futimens statistics index. */
    FDSTAT_FGETXATTR	= 13,	/*!< fgetxattr statistics index. */
    FDSTAT_FSETXATTR	= 14,	/*!< fsetxattr statistics index. */
    FDSTAT_FLISTXATTR	= 15,	/*!< flistxattr statistics index. */
    FDSTAT_FREMOVEXATTR	= 16,	/*!< fremovexattr statistics index. */
    FDSTAT_FCHFLAGS	= 17,	/*!< fchflags statistics index. */
    FDSTAT_FALLOCATE	= 18,	/*!< fallocate statistics index. */
    FDSTAT_FTRUNCATE	= 19,	/*!< ftruncate statistics index. */
    FDSTAT_FDATASYNC	= 20,	/*!< fdatasync statistics index. */
    FDSTAT_FADVISE	= 21,	/*!< fadvise statistics index. */
    FDSTAT_FSYNC	= 22,	/*!< fsync statistics index. */
    FDSTAT_SYNCFS	= 23,	/*!< syncfs statistics index. */
    FDSTAT_FINCORE	= 24,	/*!< fincore statistics index. */

    FDSTAT_MAX
} fdOpX;

/** \ingroup rpmio
 * Per-descriptor flag bits.
 */
typedef enum fdFlags_e {
    RPMFD_FLAG_NONE		= 0,
	/* bits 0-22 reserved  for linux O_* flags */
    RPMFD_FLAG_MEMFD		= (1 << 23), /* XXX dupes RPMMSQ_FLAGS_TYPE */
    RPMFD_FLAG_FTRUNCATE	= (1 << 24), /* XXX dupes RPMMSQ_FLAGS_TYPE */
    RPMFD_FLAG_SYNCFS		= (1 << 25), /* XXX dupes RPMMSQ_FLAGS_TYPE */
    RPMFD_FLAG_FSYNC		= (1 << 26), /* XXX dupes RPMMSQ_FLAGS_TYPE */
    RPMFD_FLAG_FDATASYNC	= (1 << 27),
    RPMFD_FLAG_FADVISE		= (1 << 28),
    RPMFD_FLAG_FALLOCATE	= (1 << 29),
    RPMFD_FLAG_DEBUGREFS	= (1 << 30),
    RPMFD_FLAG_DEBUGIO		= (1 << 31),
} fdFlags;
#define	RPMFD_SET(_flags, _a)	((_flags) |= RPMFD_FLAG_##_a)
#define	RPMFD_CLR(_flags, _a)	((_flags) &= ~RPMFD_FLAG_##_a)
#define	RPMFD_ISSET(_fd, _a)	((_fd)->flags &  RPMFD_FLAG_##_a)

/** \ingroup rpmio
 * Cumulative statistics for a descriptor.
 */
struct _FDSTAT_s {
    struct rpmop_s	ops[FDSTAT_MAX];	/*!< Cumulative statistics. */
};

/** \ingroup rpmio
 */
typedef struct _FDDIGEST_s {
    DIGEST_CTX		hashctx;
} * FDDIGEST_t;

/** \ingroup rpmio
 * The FD_t File Handle data structure.
 */
struct _FD_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    fdFlags	flags;		/* file handle flags */
    int		magic;
#define	FDMAGIC			0x04463138
    int		nfps;		/* no. of stacked file handles */
    FDSTACK_t	fps[8];		/* stacked file handles */

    void *	u;		/* ufdio: URL info */
    int		ut;		/* ufdio: URL type */
    void *	req;		/* ufdio: HTTP request */

    int		rd_timeoutsecs;	/* ufdRead: per FD_t timer */
    ssize_t	bytesRemain;	/* ufdio: */
    ssize_t	contentLength;	/* ufdio: */
    int		persist;	/* ufdio: */
    int		wr_chunked;	/* ufdio: */

    int		syserrno;	/* last system errno encountered */
    const void *errcookie;	/* gzdio/bzdio/ufdio: */

    const char *opath;		/* open(2) args. */
    int		oflags;
    mode_t	omode;

    size_t	nio;		/* stdio: buffer size. */
    void *	io;		/* stdio: buffer. */

    rpmxar	xar;		/* xar archive wrapper */
    pgpDig	dig;		/* signature parameters */

    FDSTAT_t	stats;		/* I/O statistics */

    size_t	ndigests;	/* no. of digests. */
    DIGEST_CTX *digests;	/* digest contexts. */

    const char *contentType;	/* ufdio: (HTTP) */
    const char *contentDisposition;	/* ufdio: (HTTP) */
    time_t	lastModified;	/* ufdio: (HTTP) */
    int		ftpFileDoneNeeded; /* ufdio: (FTP) */
    unsigned long long	fd_cpioPos;	/* cpio: */
};

#define	FDSANE(fd)	assert(fd != NULL && fd->magic == FDMAGIC)

#define DBG(_fd, _flag, _fmt) \
    if (_rpmio_debug || ((_fd) ? RPMFD_ISSET(((FD_t)(_fd)), _flag) : 0)) \
	fprintf _fmt

#define DBGIO(_fd, _fmt)   DBG((_fd), DEBUGIO, _fmt)
#define DBGREFS(_fd, _fmt) DBG((_fd), DEBUGREFS, _fmt)

#ifdef __cplusplus
extern "C" {
#endif

/** \ingroup rpmio
 */
const char * fdbg(FD_t fd);

/** \ingroup rpmio
 */
int fdFgets(FD_t fd, char * buf, size_t len);

/** \ingroup rpmio
 */
FD_t ftpOpen(const char *url, int flags, mode_t mode, urlinfo *uret);

/** \ingroup rpmio
 */
int ftpReq(FD_t data, const char * ftpCmd, const char * ftpArg);

/** \ingroup rpmio
 */
int ftpCmd(const char * cmd, const char * url, const char * arg2);

/** \ingroup rpmio
 */
int ufdClose(void * cookie);

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
const char * fdGetOPath(FD_t fd)
{
    FDSANE(fd);
    return fd->opath;
}
static inline
void fdSetOPath(FD_t fd, const char *opath)
{
    FDSANE(fd);
    if (opath) {
	if (fd->opath) {
	    free((void *)fd->opath);
	    fd->opath = NULL;
	}
	fd->opath = xstrdup(opath);
    }
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
int fdGetFlags(FD_t fd)
{
    FDSANE(fd);
    return fd->flags;
}
static inline
int fdSetFlags(FD_t fd, fdFlags flags)
{
    FDSANE(fd);
    int old_flags = fd->flags;
    fd->flags = flags;
    return old_flags;
}

/** \ingroup rpmio
 */
static inline
int fdGetOFlags(FD_t fd)
{
    FDSANE(fd);
    return fd->oflags;
}
static inline
int fdSetOFlags(FD_t fd, int oflags)
{
    FDSANE(fd);
    int old_oflags = fd->oflags;
    fd->oflags = oflags;
    return old_oflags;
}

/** \ingroup rpmio
 */
static inline
mode_t fdGetOMode(FD_t fd)
{
    FDSANE(fd);
    return fd->omode;
}
static inline
mode_t fdSetOMode(FD_t fd, mode_t omode)
{
    FDSANE(fd);
    mode_t old_omode = fd->omode;
    fd->omode = omode;
    return old_omode;
}

/** \ingroup rpmio
 */
static inline
void fdSetOpen(FD_t fd, const char * opath, int oflags, mode_t omode)
{
    FDSANE(fd);
    fdSetOPath(fd, opath);
    fdSetOFlags(fd, oflags);
    fdSetOMode(fd, omode);
}

/** \ingroup rpmio
 */
static inline
void fdSetDig(FD_t fd, pgpDig dig)
{
    FDSANE(fd);
    fd->dig = pgpDigFree(fd->dig);
    fd->dig = pgpDigLink(dig);
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
pgpDig fdGetDig(FD_t fd)
{
    FDSANE(fd);
    return fd->dig;
}

/** \ingroup rpmio
 */
static inline
void fdSetXAR(FD_t fd, rpmxar xar)
{
    FDSANE(fd);
    fd->xar = rpmxarLink(xar, "fdSetXAR");
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
rpmxar fdGetXAR(FD_t fd)
{
    FDSANE(fd);
    return fd->xar;
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
FDIO_t fdGetIo(FD_t fd)
{
    FDSANE(fd);
    return fd->fps[fd->nfps].io;
}

/** \ingroup rpmio
 */
static inline
void fdSetIo(FD_t fd, FDIO_t io)
{
    FDSANE(fd);
    fd->fps[fd->nfps].io = io;
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
FILE * fdGetFILE(FD_t fd)
{
    FDSANE(fd);
    return ((FILE *)fd->fps[fd->nfps].fp);
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
void * fdGetFp(FD_t fd)
{
    FDSANE(fd);
    return fd->fps[fd->nfps].fp;
}

/** \ingroup rpmio
 */
static inline
void fdSetFp(FD_t fd, void * fp)
{
    FDSANE(fd);
    fd->fps[fd->nfps].fp = fp;
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
int fdGetFdno(FD_t fd)
{
    FDSANE(fd);
    return fd->fps[fd->nfps].fdno;
}

/** \ingroup rpmio
 */
static inline
int fdSetFdno(FD_t fd, int fdno)
{
    int ofdno;
    FDSANE(fd);
    ofdno = fd->fps[fd->nfps].fdno;
    fd->fps[fd->nfps].fdno = fdno;
    return ofdno;
}

/** \ingroup rpmio
 */
static inline
void fdSetContentLength(FD_t fd, ssize_t contentLength)
{
    FDSANE(fd);
    fd->contentLength = fd->bytesRemain = contentLength;
}

/** \ingroup rpmio
 */
static inline
void fdPush(FD_t fd, FDIO_t io, void * fp, int fdno)
{
    FDSANE(fd);
    if (fd->nfps >= (int)(sizeof(fd->fps)/sizeof(fd->fps[0]) - 1))
	return;
    fd->nfps++;
    fdSetIo(fd, io);
    fdSetFp(fd, fp);
    fdSetFdno(fd, fdno);
}

/** \ingroup rpmio
 */
static inline
void fdPop(FD_t fd)
{
    FDSANE(fd);
    if (fd->nfps < 0) return;
    fdSetIo(fd, NULL);
    fdSetFp(fd, NULL);
    fdSetFdno(fd, -1);
    fd->nfps--;
}

/** \ingroup rpmio
 */
static inline
rpmop fdstat_op(FD_t fd, fdOpX opx)
{
    rpmop op = NULL;

    if (fd != NULL && fd->stats != NULL && (int)opx >= 0 && opx < FDSTAT_MAX)
	op = fd->stats->ops + opx;
    return op;
}

/** \ingroup rpmio
 */
static inline
void fdstat_enter(FD_t fd, int opx)
{
    if (fd == NULL) return;
    if (fd->stats != NULL)
	(void) rpmswEnter(fdstat_op(fd, (fdOpX) opx), 0);
}

/** \ingroup rpmio
 */
static inline
void fdstat_exit(FD_t fd, int opx, ssize_t rc)
{
    if (fd == NULL) return;
    if (rc == -1)
	fd->syserrno = errno;
    else if (rc > 0 && fd->bytesRemain > 0)
	switch (opx) {
	case FDSTAT_READ:
	case FDSTAT_WRITE:
	    fd->bytesRemain -= rc;
	    break;
	default:
	    break;
	}
    if (fd->stats != NULL)
	(void) rpmswExit(fdstat_op(fd, (fdOpX) opx), rc);
}

/** \ingroup rpmio
 */
static inline
void fdstat_sum(FD_t fd, FDSTAT_t sum)
{
    if (fd == NULL || fd->stats == NULL || sum == NULL) return;
    for (int opx = 0; opx < FDSTAT_MAX; opx++) {
	rpmop sop = &fd->stats->ops[opx];
	rpmop dop = &sum->ops[opx];
	(void) rpmswAdd(dop, sop);
    }
}

/** \ingroup rpmio
 */
static inline
void fdstat_print(FD_t fd, const char * msg, FILE * fp)
{
    static int usec_scale = (1000*1000);

    if (fd == NULL || fd->stats == NULL) return;
    for (int opx = 0; opx < FDSTAT_MAX; opx++) {
	static const char *const names[] = {
	    [FDSTAT_READ]		= "        read",
	    [FDSTAT_WRITE]		= "       write",
	    [FDSTAT_SEEK]		= "        seek",
#ifdef	NOISY
	    [FDSTAT_CLOSE]		= "       close",
#endif
	    [FDSTAT_FDIGEST]		= "     fdigest",
	    [FDSTAT_FOPEN]		= "       fopen",
	    [FDSTAT_FCNTL]		= "       fcntl",
	    [FDSTAT_FLOCK]		= "       flock",
	    [FDSTAT_FCHDIR]		= "      fchdir",
	    [FDSTAT_FCHMOD]		= "      fchmod",
	    [FDSTAT_FCHOWN]		= "      fchown",
	    [FDSTAT_FUTIMES]		= "     futimes",
	    [FDSTAT_FUTIMENS]		= "    futimens",
	    [FDSTAT_FGETXATTR]		= "   fgetxattr",
	    [FDSTAT_FSETXATTR]		= "   fsetxattr",
	    [FDSTAT_FLISTXATTR]		= "  flistxattr",
	    [FDSTAT_FREMOVEXATTR]	= "fremovexattr",
	    [FDSTAT_FCHFLAGS]		= "    fchflags",
	    [FDSTAT_FALLOCATE]		= "   fallocate",
	    [FDSTAT_FTRUNCATE]		= "   ftruncate",
#ifdef	NOISY
	    [FDSTAT_FADVISE]		= "     fadvise",
#endif
	    [FDSTAT_FDATASYNC]		= "   fdatasync",
	    [FDSTAT_FSYNC]		= "       fsync",
	    [FDSTAT_SYNCFS]		= "      syncfs",

	    [FDSTAT_FINCORE]		= "     fincore",
	};
	rpmop op = &fd->stats->ops[opx];

	if (names[opx] == NULL || *names[opx] == '\0') continue;
	if (op->count <= 0) continue;

	if (msg != NULL) fprintf(fp, "%s:", msg);
	fprintf(fp, "%8d %s(s), %8lu total bytes in %d.%06d secs\n",
		op->count, names[opx], (unsigned long)op->bytes,
		(int)(op->usecs/usec_scale), (int)(op->usecs%usec_scale));

    }
}

/** \ingroup rpmio
 */
static inline
void fdSetSyserrno(FD_t fd, int syserrno, const void * errcookie)
{
    FDSANE(fd);
    fd->syserrno = syserrno;
    fd->errcookie = errcookie;
}

/** \ingroup rpmio
 */
static inline
int fdGetRdTimeoutSecs(FD_t fd)
{
    FDSANE(fd);
    return fd->rd_timeoutsecs;
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
unsigned long long fdGetCpioPos(FD_t fd)
{
    FDSANE(fd);
    return fd->fd_cpioPos;
}

/** \ingroup rpmio
 */
static inline
void fdSetCpioPos(FD_t fd, long int cpioPos)
{
    FDSANE(fd);
    fd->fd_cpioPos = cpioPos;
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
FD_t c2f(void * cookie)
{
    FD_t fd = (FD_t) cookie;
    FDSANE(fd);
    return fd;
}

/** \ingroup rpmio
 * Attach digest to fd.
 */
static inline
void fdInitDigest(FD_t fd, pgpHashAlgo hashalgo, int _flags)
{
    static const int opx = FDSTAT_FDIGEST;
    rpmDigestFlags flags = (rpmDigestFlags) _flags;
    fd->digests = (DIGEST_CTX *) realloc(fd->digests,
			(fd->ndigests + 1) * sizeof(*fd->digests));
    fdstat_enter(fd, opx);
    fd->digests[fd->ndigests++] = rpmDigestInit(hashalgo, flags);
    fdstat_exit(fd, opx, 0);
}

/** \ingroup rpmio
 * Attach digest to fd.
 */
static inline
void fdInitHmac(FD_t fd, const void * key, size_t keylen)
{
    static const int opx = FDSTAT_FDIGEST;
    fdstat_enter(fd, opx);
    if (fd->digests != NULL && fd->ndigests > 0 && key != NULL)
	(void) rpmHmacInit(fd->digests[fd->ndigests-1], key, keylen);
    fdstat_exit(fd, opx, 0);
}

/** \ingroup rpmio
 * Update digest(s) attached to fd.
 */
static inline
void fdUpdateDigests(FD_t fd, const unsigned char * buf, ssize_t buflen)
{
    static const int opx = FDSTAT_FDIGEST;

    fdstat_enter(fd, opx);
    if (fd->ndigests > 0 && buf != NULL && buflen > 0)
#if defined(_OPENMP)
#pragma omp parallel for if (fd->ndigests > 1)
#endif
    for (int i = fd->ndigests - 1; i >= 0; i--) {
	DIGEST_CTX ctx = fd->digests[i];
	if (ctx == NULL)
	    continue;
	(void) rpmDigestUpdate(ctx, buf, buflen);
    }
    fdstat_exit(fd, opx, buflen);
}

/** \ingroup rpmio
 */
static inline
void fdFiniDigest(FD_t fd, pgpHashAlgo hashalgo,
		void * datap, size_t * lenp, int asAscii)
{
    static const int opx = FDSTAT_FDIGEST;
    int i = -1;

    fdstat_enter(fd, opx);
    if (fd->ndigests > 0)
    for (i = fd->ndigests - 1; i >= 0; i--) {
	DIGEST_CTX ctx = fd->digests[i];
	if (ctx == NULL)
	    continue;
	if (rpmDigestAlgo(ctx) != hashalgo)
	    continue;
	fd->digests[i] = NULL;
	(void) rpmDigestFinal(ctx, datap, lenp, asAscii);
	break;
    }
    if (i < 0) {
	if (datap != NULL) *(void **)datap = NULL;
	if (lenp != NULL) *lenp = 0;
    }
    fdstat_exit(fd, opx, 0);
}

/** \ingroup rpmio
 */
static inline
void fdStealDigest(FD_t fd, pgpDig dig)
{
    static const int opx = FDSTAT_FDIGEST;

    fdstat_enter(fd, opx);
    if (fd->ndigests > 0)
    for (int i = fd->ndigests - 1; i >= 0; i--) {
	DIGEST_CTX ctx = fd->digests[i];
	if (ctx != NULL)
	switch (rpmDigestAlgo(ctx)) {
	case PGPHASHALGO_MD5:
assert(dig->md5ctx == NULL);
	    dig->md5ctx = ctx;
	    fd->digests[i] = NULL;
	    break;
	case PGPHASHALGO_SHA1:
	case PGPHASHALGO_RIPEMD160:
	case PGPHASHALGO_SHA256:
	case PGPHASHALGO_SHA384:
	case PGPHASHALGO_SHA512:
assert(dig->hsha == NULL);
	    dig->hsha = ctx;
	    fd->digests[i] = NULL;
	    break;
	default:
	    break;
	}
    }
    fdstat_exit(fd, opx, 0);
}

/** \ingroup rpmio
 */
RPM_GNUC_PURE
static inline
int fdFileno(void * cookie)
{
    FD_t fd;
    if (cookie == NULL) return -2;
    fd = c2f(cookie);
    return fd->fps[0].fdno;
}

/** \ingroup rpmio
 * Common prologue(s) for checking wrapped system functions.
 * @return		file descriptor
 */
static inline
int fdSyscallPreFdno(const char * sysname, FD_t fd, int local, const char ** lpathp)
{
    int fdno = Fileno(fd);
    const char * path = fdGetOPath(fd);
    const char * lpath;
    int ut = urlPath(path, &lpath);

    FDSANE(fd);
DBGIO(fd, (stderr, "--> %s(%p) fdno %d path %s\n", sysname, fd, fdno, path));

    if (fdno < 0) {
	errno = EBADF;
	goto exit;
    }

    if (local)			/* XXX Restrict to local file paths? */
    switch (ut) {
    case URL_IS_PATH:
    case URL_IS_UNKNOWN:
	break;
    default:
	fdno = -1;
	errno = EINVAL;
	goto exit;
	break;
    }

exit:
    if (lpathp)
	*lpathp = lpath;
    return fdno;
}

static inline
const char * fdSyscallPrePath(const char * sysname, const char *path, int local)
{
    const char * lpath = NULL;
    int ut = urlPath(path, &lpath);

if (_rpmio_debug)
fprintf(stderr, "--> %s(%s)\n", sysname, path);

    if (path == NULL) {
	path = NULL;
	errno = EFAULT;
	goto exit;
    }

    if (local)			/* XXX Restrict to local file paths? */
    switch (ut) {
    case URL_IS_PATH:		/* XXX Drop file:/// prefix. */
	path = lpath;
	/*@fallthrough@*/
    case URL_IS_UNKNOWN:
	break;
    default:
	path = NULL;
	errno = EINVAL;
	goto exit;
	break;
    }

exit:
    return path;
}

/** \ingroup rpmio
 * Common epilogue for checking wrapped system functions.
 * @return		exit code
 */
static inline
int fdSyscallPostFdno(const char * sysname, FD_t fd, int fdno, int rc)
{
    if (rc < 0)
	rpmlog(RPMLOG_DEBUG, _("%s(%d) failed: rc %d %m\n"),
		__FUNCTION__, fdno, rc);
DBGIO(fd, (stderr, "<-- %s(%p) fdno %d rc %d\n", sysname, fd, fdno, rc));
    return rc;
}

static inline
int fdSyscallPostPath(const char * sysname, const char *lpath, int rc)
{
    if (rc < 0)
	rpmlog(RPMLOG_DEBUG, _("%s(%s) failed: rc %d %m\n"),
		__FUNCTION__, lpath, rc);
if (_rpmio_debug)
fprintf(stderr, "<-- %s(%s) rc %d\n", sysname, lpath, rc);
    return rc;
}

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMIO_INTERNAL */
