#ifndef RPMJNI_H
#define RPMJNI_H

/** \ingroup rpmio
 * \file rpmio/rpmjni.h
 */

#include <rpmiotypes.h>
#include <rpmio.h>

typedef struct rpmjni_s * rpmjni;

extern int _rpmjni_debug;

extern rpmjni _rpmjniI;

#if defined(_RPMJNI_INTERNAL)
struct rpmjni_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    void * ENV;			/* JNIEnv* */
    void * VM;			/* JavaVM* */
    void * C;			/* jclass FindClass("bsh/Interpreter") */
    void * mid_main;	/* ("main", "([Ljava/lang/String;)V") */
    void * mid_init;	/* ("<init>", "()V") */
    void * I;			/* jobject NewObject(C, mid_init) */
    void * IC;			/* jclass GetObjectClass(I) */
    void * mid_eval;	/* ("eval", "(Ljava/lang/String;)Ljava/lang/Object;") */
    void * String;		/* jclass */
    void * Integer;			/* jclass */
    void * Float;		/* jclass */
    void * Double;		/* jclass */
    void * mid_intValue;	/* jmethodID */
    void * mid_floatValue;	/* jmethodID */
    void * mid_doubleValue;	/* jmethodID */
};
#endif /* _RPMJNI_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a jni interpreter instance.
 * @param jni	jni interpreter
 * @return		NULL on last dereference
 */
rpmjni rpmjniUnlink (rpmjni jni);
#define	rpmjniUnlink(_jni)	\
    ((rpmjni)rpmioUnlinkPoolItem((rpmioItem)(_jni), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a jni interpreter instance.
 * @param jni	jni interpreter
 * @return		new jni interpreter reference
 */
rpmjni rpmjniLink (rpmjni jni);
#define	rpmjniLink(_jni)	\
    ((rpmjni)rpmioLinkPoolItem((rpmioItem)(_jni), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a jni interpreter.
 * @param jni	jni interpreter
 * @return		NULL on last dereference
 */
rpmjni rpmjniFree(rpmjni jni);
#define	rpmjniFree(_jni)	\
    ((rpmjni)rpmioFreePoolItem((rpmioItem)(_jni), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a jni interpreter.
 * @param av		jni interpreter args (or NULL)
 * @param flags		jni interpreter flags ((1<<31): use global interpreter)
 * @return		new jni interpreter
 */
rpmjni rpmjniNew(char ** av, uint32_t flags);

/**
 * Execute jni string.
 * @param jni	jni interpreter (NULL uses global interpreter)
 * @param str		jni string to execute (NULL returns RPMRC_FAIL)
 * @param resultp	*resultp jni exec result
 * @return		RPMRC_OK on success
 */
rpmRC rpmjniRun(rpmjni jni, const char * str, const char ** resultp);

#ifdef __cplusplus
}
#endif

#endif /* RPMJNI_H */
