#ifndef RPMJSS_H
#define RPMJSS_H

/** \ingroup rpmio
 * \file rpmio/rpmjss.h
 */

#include <rpmiotypes.h>
#include <rpmio.h>
#include <argv.h>

typedef struct rpmjss_s * rpmjss;

extern int _rpmjss_debug;

extern rpmjss _rpmjssI;

#if defined(_RPMJSS_INTERNAL)
/**
 * Interpreter flags.
 */
typedef enum rpmjssFlags_e {
    RPMJSS_FLAGS_NONE		= 0,
    RPMJSS_FLAGS_STRICT		= (1<< 0),	/* JSOPTION_STRICT */
    RPMJSS_FLAGS_WERROR		= (1<< 1),	/* JSOPTION_WERROR */
    RPMJSS_FLAGS_VAROBJFIX	= (1<< 2),	/* JSOPTION_VAROBJFIX */
    RPMJSS_FLAGS_PRIVATE_IS_NSISUPPORTS = (1<< 3), /* JSOPTION_PRIVATE_IS_NSISUPPORTS */
    RPMJSS_FLAGS_COMPILE_N_GO	= (1<< 4),	/* JSOPTION_COMPILE_N_GO */
    RPMJSS_FLAGS_ATLINE		= (1<< 5),	/* JSOPTION_ATLINE */
    RPMJSS_FLAGS_XML		= (1<< 6),	/* JSOPTION_XML */
	/* bit 7 unused */
    RPMJSS_FLAGS_DONT_REPORT_UNCAUGHT = (1<< 8),	/* JSOPTION_DONT_REPORT_UNCAUGHT */
    RPMJSS_FLAGS_RELIMIT	= (1<< 9),	/* JSOPTION_RELIMIT */
    RPMJSS_FLAGS_ANONFUNFIX	= (1<<10),	/* JSOPTION_ANONFUNFIX */
    RPMJSS_FLAGS_JIT		= (1<<11),	/* JSOPTION_JIT */
    RPMJSS_FLAGS_NO_SCRIPT_RVAL	= (1<<12),	/* JSOPTION_NO_SCRIPT_RVAL */
    RPMJSS_FLAGS_UNROOTED_GLOBAL= (1<<13),	/* JSOPTION_UNROOTED_GLOBAL */
	/* bits 14-15 unused */
    RPMJSS_FLAGS_NOEXEC		= (1<<16),	/*!< -n */
    RPMJSS_FLAGS_SKIPSHEBANG	= (1<<17),	/*!< -F */
    RPMJSS_FLAGS_LOADRC		= (1<<18),	/*!< -R */
    RPMJSS_FLAGS_NOUTF8		= (1<<19),	/*!< -U */
    RPMJSS_FLAGS_NOCACHE	= (1<<20),	/*!< -C */
    RPMJSS_FLAGS_NOWARN		= (1<<21),	/*!< -W */
    RPMJSS_FLAGS_ALLOW		= (1<<22),	/*!< -a */
	/* bits 23-30 unused */
    RPMJSS_FLAGS_GLOBAL		= (1<<31),
} rpmjssFlags;

typedef enum rpmjssIonFlags_e {
    ION_SHARED_STUBS		= (1 <<  0),
    ION_SCALAR_REPLACEMENT	= (1 <<  1),
    ION_GVN			= (1 <<  2),
    ION_LICM			= (1 <<  3),
    ION_EDGECASE_ANALYSIS	= (1 <<  4),
    ION_PGO			= (1 <<  5),
    ION_RANGE_ANALYSIS		= (1 <<  6),
    ION_SINCOS			= (1 <<  7),
    ION_SINK			= (1 <<  8),
    ION_LOOP_UNROLLING		= (1 <<  9),
    ION_INSTRUCTION_REORDERING	= (1 << 10),
    ION_CHECK_RANGE_ANALYSIS	= (1 << 11),
    ION_EXTRA_CHECKS		= (1 << 12),
    ION_INLINING		= (1 << 13),
    ION_OSR			= (1 << 14),
    ION_LIMIT_SCRIPT_SIZE	= (1 << 15),
    ION_EAGER			= (1 << 16),
    ION_OFFTHREAD_COMPILE	= (1 << 17),
    ION_PARALLEL_COMPILE	= (1 << 18),
} rpmjssIonFlags;

typedef enum rpmjssXFlags_e {
    XF_INTERACTIVE		= (1 <<  0),
    XF_COMPILE_ONLY		= (1 <<  1),
    XF_WARNINGS			= (1 <<  2),
    XF_STRICT			= (1 <<  3),
    XF_DUMP_BYTECODE		= (1 <<  4),
    XF_PRINT_TIMING		= (1 <<  5),
    XF_CACHE_PER_PROCESS	= (1 <<  6),
    XF_CODE_COVERAGE		= (1 <<  7),
    XF_PRINT_ALLOC		= (1 <<  8),
    XF_ION			= (1 <<  9),
    XF_ASMJS			= (1 << 10),
    XF_NATIVE_REGEXP		= (1 << 11),
    XF_UNBOXED_OBJECTS		= (1 << 12),
    XF_UNBOXED_ARRAYS		= (1 << 13),
    XF_BASELINE			= (1 << 14),
    XF_BASELINE_EAGER		= (1 << 15),
    XF_NONWRITABLE_JITCODE	= (1 << 16),
    XF_FPU			= (1 << 17),
    XF_SSE3			= (1 << 18),
    XF_SSE4			= (1 << 19),
    XF_AVX			= (1 << 20),
    XF_FUZZING_SAFE		= (1 << 21),
    XF_DISABLE_OOM		= (1 << 22),
    XF_THREADS			= (1 << 23),
    XF_DUMP_ENTRAINED_VARIABLES	= (1 << 24),
    XF_GGC			= (1 << 25),
    XF_CGC			= (1 << 26),
    XF_INCREMENTAL_GC		= (1 << 27),
    XF_ARM_SIM_ICACHE_CHECKS	= (1 << 28),
    XF_MIPS_SIM_ICACHE_CHECKS	= (1 << 29),
} rpmjssXFlags;

typedef	struct JSIO_s * JSIO_t;
struct JSIO_s {
    void (*mozFini)	(rpmjss jss);
    void * (*mozInit)	(rpmjss jss);
    int  (*mozRun)	(rpmjss jss, const char ** resultp,
		const char * script, const char * fn, unsigned ln);
};

extern JSIO_t mozjs185;
extern JSIO_t mozjs17;
extern JSIO_t mozjs24;
extern JSIO_t mozjs31;
extern JSIO_t mozjs38;
extern JSIO_t mozjs45;

struct rpmjss_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */

    rpmjssFlags flags;		/*!< JSOPTION_FOO in 0xffff bits */
    ARGV_t av;
    int ac;

    JSIO_t jsio;
    void * I;			/*!< JS interpreter {rt, cx, globalObj} */

    const char *Ifile;
    const char *Imodule;
    const char *Icode;
    const char *cachedir;
    const char *moduledir;

    const char *script_path;
    const char *script_args;

    const char * ion_shared_stubs;
    const char * ion_scalar_replacement;
    const char * ion_gvn;
    const char * ion_licm;
    const char * ion_edgecase_analysis;
    const char * ion_pgo;
    const char * ion_range_analysis;
    const char * ion_sincos;
    const char * ion_sink;
    const char * ion_loop_unrolling;
    const char * ion_instruction_reordering;

    const char * ion_inlining;
    const char * ion_osr;
    const char * ion_limit_script_size;
    const char * ion_offthread_compile;

    const char * ion_regalloc;

    const char * arm_hwcap;

    const char * gc_zeal;

    int thread_count;

    int ion_warmup_threshold;
    int baseline_warmup_threshold;

    unsigned available_memory;
    int arm_asm_nop_fill;
    int asm_pool_max_offset;
    int arm_sim_stop_at;
    int mips_sim_stop_at;
    unsigned nursery_size;

    rpmjssXFlags xf;
    rpmjssIonFlags ionf;
};

struct rpmjss_s _rpmjss;

#endif /* _RPMJSS_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a js shell interpreter instance.
 * @param jss		js shell interpreter
 * @return		NULL on last dereference
 */
rpmjss rpmjssUnlink (rpmjss jss);
#define	rpmjssUnlink(_jss)	\
    ((rpmjss)rpmioUnlinkPoolItem((rpmioItem)(_jss), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a js shell interpreter instance.
 * @param jss		js shell interpreter
 * @return		new js shell interpreter reference
 */
rpmjss rpmjssLink (rpmjss jss);
#define	rpmjssLink(_jss)	\
    ((rpmjss)rpmioLinkPoolItem((rpmioItem)(_jss), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a js shell interpreter.
 * @param jss		js shell interpreter
 * @return		NULL on last dereference
 */
rpmjss rpmjssFree(rpmjss jss);
#define	rpmjssFree(_jss)	\
    ((rpmjss)rpmioFreePoolItem((rpmioItem)(_jss), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a js shell interpreter.
 * @param av		js shell interpreter args (or NULL)
 * @param flags		js shell interpreter flags ((1<<31): use global interpreter)
 * @return		new js interpreter
 */
rpmjss rpmjssNew(char ** av, uint32_t flags);

/**
 * Execute js from a file.
 * @param js		js interpreter (NULL uses global interpreter)
 * @param fn		js file to run (NULL returns RPMRC_FAIL)
 * @param Iargv		js script argv
 * @param resultp	*resultp js exec result
 * @return		RPMRC_OK on success
 */
rpmRC rpmjssRunFile(rpmjss jss, const char * fn,
		char *const * Iargv, const char ** resultp);

/**
 * Execute js string.
 * @param js		js interpreter (NULL uses global interpreter)
 * @param str		js string to execute (NULL returns RPMRC_FAIL)
 * @param resultp	*resultp js exec result
 * @return		RPMRC_OK on success
 */
rpmRC rpmjssRun(rpmjss jss, const char * str, const char ** resultp);

#ifdef __cplusplus
}
#endif

#endif /* RPMJSS_H */
