#ifndef RPMLUA_H
#define RPMLUA_H

typedef enum rpmluavType_e {
    RPMLUAV_NIL		= 0,
    RPMLUAV_STRING	= 1,
    RPMLUAV_NUMBER	= 2
} rpmluavType;

#if defined(_RPMLUA_INTERNAL)

#include <stdarg.h>
#include <lua.h>

struct rpmlua_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    lua_State *L;
    int pushsize;
    int storeprint;
    size_t printbufsize;
    size_t printbufused;
    char *printbuf;
};

struct rpmluav_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    rpmluavType keyType;
    rpmluavType valueType;
    union {
	const char *str;
	const void *ptr;
	double num;
    } key;
    union {
	const char *str;
	const void *ptr;
	double num;
    } value;
    int listmode;
};

#endif /* _RPMLUA_INTERNAL */

typedef struct rpmlua_s * rpmlua;
typedef struct rpmluav_s * rpmluav;

#ifdef __cplusplus
extern "C" {
#endif

extern const char * rpmluaFiles;

extern const char * rpmLUA_PATH;

extern const char * rpmLUA_CPATH;

rpmlua rpmluaGetGlobalState(void)
	RPM_GNUC_PURE;

rpmlua rpmluaNew(void);

void *rpmluaFree(rpmlua lua);

int rpmluaCheckScript(rpmlua _lua, const char *script, const char *name);
int rpmluaRunScript(rpmlua _lua, const char *script, const char *name);
int rpmluaRunScriptFile(rpmlua _lua, const char *filename);
void rpmluaInteractive(rpmlua _lua);

void *rpmluaGetData(rpmlua _lua, const char *key);
void rpmluaSetData(rpmlua _lua, const char *key, const void *data);

const char *rpmluaGetPrintBuffer(rpmlua _lua);
void rpmluaSetPrintBuffer(rpmlua _lua, int flag);

void rpmluaGetVar(rpmlua _lua, rpmluav var);
void rpmluaSetVar(rpmlua _lua, rpmluav var);
void rpmluaDelVar(rpmlua _lua, const char *key, ...);
int rpmluaVarExists(rpmlua _lua, const char *key, ...);
void rpmluaPushTable(rpmlua _lua, const char *key, ...);
void rpmluaPop(rpmlua _lua);

rpmluav rpmluavNew(void);
void * rpmluavFree(rpmluav var);
void rpmluavSetListMode(rpmluav var, int flag);
void rpmluavSetKey(rpmluav var, rpmluavType type, const void *value);
void rpmluavSetValue(rpmluav var, rpmluavType type, const void *value);
void rpmluavGetKey(rpmluav var, rpmluavType *type, void **value);
void rpmluavGetValue(rpmluav var, rpmluavType *type, void **value);

/* Optional helpers for numbers. */
void rpmluavSetKeyNum(rpmluav var, double value);
void rpmluavSetValueNum(rpmluav var, double value);
double rpmluavGetKeyNum(rpmluav var);
double rpmluavGetValueNum(rpmluav var);
int rpmluavKeyIsNum(rpmluav var)
	RPM_GNUC_PURE;
int rpmluavValueIsNum(rpmluav var)
	RPM_GNUC_PURE;

#ifdef __cplusplus
}
#endif

#endif /* RPMLUA_H */
