#ifndef	H_RPMMG
#define	H_RPMMG

/** \ingroup rpmio
 * \file rpmio/rpmmg.h
 */

#include <stddef.h>

/** \ingroup rpmio
 */
extern int _rpmmg_debug;

/** \ingroup rpmio
 */
typedef struct rpmmg_s * rpmmg;

#if defined(_RPMMG_INTERNAL)
/** \ingroup rpmio
 */
struct rpmmg_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * fn;
    int flags;
    struct magic_set * ms;
};
#endif	/* _RPMMG_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a magic wrapper instance.
 * @param mg		magic wrapper
 * @return		NULL on last dereference
 */
rpmmg rpmmgUnlink (rpmmg mg);
#define	rpmmgUnlink(_mg)	\
    ((rpmmg)rpmioUnlinkPoolItem((rpmioItem)(_mg), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a magic wrapper instance.
 * @param mg		magic wrapper
 * @return		new magic wrapper reference
 */
rpmmg rpmmgLink (rpmmg mg);
#define	rpmmgLink(_mg)	\
    ((rpmmg)rpmioLinkPoolItem((rpmioItem)(_mg), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a magic wrapper.
 * @param mg		magic wrapper
 * @return		NULL on last dereference
 */
rpmmg rpmmgFree(rpmmg mg);
#define	rpmmgFree(_mg)	\
    ((rpmmg)rpmioFreePoolItem((rpmioItem)(_mg), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a magic wrapper.
 * @param fn		magic file
 * @param flags		magic flags
 * @return		new magic wrapper
 */
rpmmg rpmmgNew(const char * fn, int flags);

/**
 * Return magic string for a file.
 * @param mg		magic wrapper
 * @param fn		file path
 * @return		file magic string
 */
const char * rpmmgFile(rpmmg mg, const char *fn);

/**
 * Return magic string for a buffer.
 * @param mg		magic wrapper
 * @param b		buffer
 * @param nb		no. bytes in buffer
 * @return		buffer magic string
 */
const char * rpmmgBuffer(rpmmg mg, const char * b, size_t nb);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMMG */
