#ifndef	H_RPMMQTT
#define	H_RPMMQTT

/** \ingroup rpmio
 * \file rpmio/rpmmqtt.h
 */

extern int _rpmmqtt_debug;

typedef struct rpmmqtt_s	* rpmmqtt;
typedef struct pahoState_s	* pahoState_t;
typedef	struct mqttVec_s	* mqttVec_t;

#define _KFB(n) (1U << (n))
#define _MFB(n) (_KFB(n) | 0x40000000)
typedef enum mqttFlags_e {
    MQTT_FLAGS_NONE		= 0,

    MQTT_FLAGS_CLEAN		= _MFB( 0),	/*!< (sub) -c */
    MQTT_FLAGS_EOL		= _MFB( 1),	/*!< (sub) -N */
    MQTT_FLAGS_NOSTALE		= _MFB( 2),	/*!< (sub) -R */

    MQTT_FLAGS_LINES		= _MFB( 3),	/*!< -l */
    MQTT_FLAGS_EMPTY		= _MFB( 4),	/*!< -n */
    MQTT_FLAGS_DNSSRV		= _MFB( 5),	/*!< -S */
    MQTT_FLAGS_INSECURE		= _MFB( 6),	/*!< --insecure */

    MQTT_FLAGS_RETAIN		= _MFB( 8),	/*!< -r */
    MQTT_FLAGS_WILL_RETAIN	= _MFB( 9),	/*!< --will-retain */
    MQTT_FLAGS_BUFFER		= _MFB(10),	/*!< -b */
	/* unused 11-15 */
    MQTT_FLAGS_DURABLE		= _MFB(16),	/*!<    --durable */
    MQTT_FLAGS_EXCLUSIVE	= _MFB(17),	/*!< -x,--exclusive */
    MQTT_FLAGS_NOACK		= _MFB(18),	/*!<    --no-ack */
    MQTT_FLAGS_IFEMPTY		= _MFB(19),	/*!< -e,--if-empty */
    MQTT_FLAGS_IFUNUSED		= _MFB(20),	/*!< -u,--if-unused */

    MQTT_FLAGS_PASSIVE		= _MFB(21),	/*!< XXX W2DO? */
    MQTT_FLAGS_AUTODELETE	= _MFB(22),	/*!< XXX W2DO? */
    MQTT_FLAGS_NOLOCAL		= _MFB(23),	/*!< XXX W2DO? */

} mqttFlags;
#define	MQTT_FLAGS_DEFAULT	((mqttFlags)(MQTT_FLAGS_CLEAN|MQTT_FLAGS_EOL))
#undef	_MFB
#undef	_KFB

typedef enum mqttInput_e {
    MQTT_INPUT_UNKNOWN		= 0,
    MQTT_INPUT_OPTIONS		= 1,		/*!< -m str */
    MQTT_INPUT_FILES_LINES	= 2,		/*!< -l -f fn */
    MQTT_INPUT_FILES		= 3,		/*!< -f fn */
} mqttInput;

typedef enum mqttOutput_e {
    MQTT_OUTPUT_UNKNOWN		= 0,
    MQTT_OUTPUT_STDOUT		= (1 << 0),
    MQTT_OUTPUT_FILE		= (1 << 1),
    MQTT_OUTPUT_CALLBACK	= (1 << 2),
} mqttOutput;

#ifdef __cplusplus
extern "C" {
#endif

#if defined(_RPMMQTT_INTERNAL)

#define SPEW(_list) \
	if ((mqtt && mqtt->debug) || _rpmmqtt_debug < 0) fprintf _list

struct rpmmqtt_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    MacroContext mc;
    char ** av;
    int ac;

    const char *_progname;
    const char *_progmode;
    mqttFlags flags;
    mqttInput msg_input;
    mqttOutput msg_output;

    rpmiob iob;					/*   -b */
    int debug;					/*!< -d */
    const char **ifn;				/*!< -f */
    FD_t ifd;
    const char **msgs;				/*!< -m */
    const char * ofn;				/*!< -o */
    FD_t ofd;

    /* -- Connect options: */
    const char * host;				/*!< -h */
    int port;					/*!< -p */
    const char * user;				/*!< -u */
    const char * password;			/*!< -P */
    const char *_address;			/*!< -A */
    const char * cacertfile;			/*!< --cafile */
    const char *_capath;			/*!< --capath */
    const char * certfile;			/*!< --cert */
    const char * keyfile;			/*!< --key */
    const char * ciphers;			/*!< --ciphers */
          char *_tls_version;			/*!< --tls-version */
    const char *_psk_key;			/*!< --psk */
    const char *_psk_identity;			/*!< --psk-identity */
    const char *_proxy;				/*!< --proxy */
    const char *_uri;				/*!< -U */
          char * uri;
    urltype ut;			/* "tcp://localhost:1833" */
    urlinfo u;

    /* -- MQTT protocol options: */
          char *_clientid;			/*!< -i */
    const char * clientid;
          char * idprefix;			/*!< -I */
    int keepalive;				/*!< -k */
    int max_inflight;				/*!< -M */
    unsigned qos;				/*<! -q */
    const char **topics;			/*!< -t */
    const char **subtopics;
          char * topic;
          char * protocol_version;		/*!< -V */
          char * will_message;			/*!< --will-payload */
    int will_qos;				/*!< --will-qos */
    const char * will_topic;			/*!< --will-topic */
    unsigned timeout;		/* msecs */
    unsigned persist_type;
    void * persist_ctx;		/* MQTTClient_persistence */
    const char * persist_path;
    const char * cachedn;
    int trace;

    /* -- MQTT subscribe options: */
    int msg_count;
    int max_msg_count;				/*!< (sub) -C */
    const char **_filter_out;			/*!< -T */

    /* -- AMQP options: */
    const char *vhost;			/*    --vhost */
    const char *queue;			/* -Q,--queue */
    const char *exchange;		/* -X,--exchange */
    const char *_routing_key;		/* -R,--routing-key */
    const char *_reply_to;		/* -t,--reply-to */
    const char *_content_type;		/* -C,--content-type */
    const char *_content_encoding;	/* -E,--content-encoding */
    int _delivery;			/*    --persistent */
    int _prefetch_count;		/*    --prefetch-count */

    int _verify;		/* XXX ~MQTT_FLAGS_INSECURE */
    int _fail_if_no_peer_cert;;
    const char * _sasl_mechanism;
    int _connection_timeout;	/* XXX mqtt->timeout */
    int _channel_max;

    /* -- internals */
    void * I;			/* MQTTClient */
    void * M;			/* pn_message_t * */
    volatile int connected;	/* XXX mosquitto/rabbitmq */
    volatile int finished;
    volatile unsigned token;
    char * serverURI;
    int MQTTVersion;
    int sessionPresent;
    pahoState_t state;
    mqttVec_t vec;
};

struct mqttVec_s {
    const char * name;
    const char * uri;
    uint16_t port;
    uint16_t sport;
    const char * prefix;
    int all;
    void * errs;
    size_t nerrs;
    rpmRC (*destroy)		(rpmmqtt mqtt);
    rpmRC (*create)		(rpmmqtt mqtt);
    rpmRC (*disconnect)		(rpmmqtt mqtt);
    rpmRC (*connect)		(rpmmqtt mqtt);
    rpmRC (*isconnected)	(rpmmqtt mqtt);
    rpmRC (*unsubscribe)	(rpmmqtt mqtt, const char *topic);
    rpmRC (*subscribe)		(rpmmqtt mqtt, const char *topic, int qos);
    rpmRC (*unsubscribeMany)	(rpmmqtt mqtt, int ac, char ** av);
    rpmRC (*subscribeMany)	(rpmmqtt mqtt, int ac, char ** av);
    rpmRC (*sendMessage)	(rpmmqtt mqtt, const char * topic,
					const char *s, size_t ns);
};

static inline
rpmRC rpmmqttDestroy(rpmmqtt mqtt)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->destroy
	? (mqtt->vec->destroy) (mqtt)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p) rc %d\n", __FUNCTION__, mqtt, rc));
    return rc;
}

static inline
rpmRC rpmmqttCreate(rpmmqtt mqtt)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->create
	? (mqtt->vec->create) (mqtt)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p) rc %d\n", __FUNCTION__, mqtt, rc));
    return rc;
}

static inline
rpmRC rpmmqttDisconnect(rpmmqtt mqtt)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->disconnect
	? (mqtt->vec->disconnect) (mqtt)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p) rc %d\n", __FUNCTION__, mqtt, rc));
    return rc;
}

static inline
rpmRC rpmmqttConnect(rpmmqtt mqtt)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->connect
	? (mqtt->vec->connect) (mqtt)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p) rc %d\n", __FUNCTION__, mqtt, rc));
    return rc;
}

static inline
rpmRC rpmmqttIsConnected(rpmmqtt mqtt)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->isconnected
	? (mqtt->vec->isconnected) (mqtt)
	: RPMRC_NOTFOUND;	/* XXX */
SPEW((stderr, "<-- %s(%p) rc %d\n", __FUNCTION__, mqtt, rc));
    return rc;
}

static inline
rpmRC rpmmqttUnsubscribe(rpmmqtt mqtt, const char *topic)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->unsubscribe
	? (mqtt->vec->unsubscribe) (mqtt, topic)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p,\"%s\") rc %d\n", __FUNCTION__, mqtt, topic, rc));
    return rc;
}

static inline
rpmRC rpmmqttSubscribe(rpmmqtt mqtt, const char *topic, int qos)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->subscribe
	? (mqtt->vec->subscribe) (mqtt, topic, qos)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p,%p[%d]) rc %d\n", __FUNCTION__, mqtt, topic, qos, rc));
    return rc;
}

static inline
rpmRC rpmmqttUnsubscribeMany(rpmmqtt mqtt, int ac, char ** av)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->unsubscribeMany
	? (mqtt->vec->unsubscribeMany) (mqtt, ac, av)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p,%p[%d]) rc %d\n", __FUNCTION__, mqtt, av, ac, rc));
    return rc;
}

static inline
rpmRC rpmmqttSubscribeMany(rpmmqtt mqtt, int ac, char ** av)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->subscribeMany
	? (mqtt->vec->subscribeMany) (mqtt, ac, av)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p,%p[%d]) rc %d\n", __FUNCTION__, mqtt, av, ac, rc));
    return rc;
}

static inline
rpmRC rpmmqttSendMessage(rpmmqtt mqtt, const char * topic,
		const char *s, size_t ns)
{
    rpmRC rc = mqtt && mqtt->vec && mqtt->vec->sendMessage
	? (mqtt->vec->sendMessage) (mqtt, topic, s, ns)
	: RPMRC_FAIL;
SPEW((stderr, "<-- %s(%p,%p[%u]) rc %d\n", __FUNCTION__, mqtt, s, (unsigned)ns, rc));
    return rc;
}

#endif	/* _RPMMQTT_INTERNAL */

/**
 * Unreference a mqtt wrapper instance.
 * @param mqtt		mqtt wrapper
 * @return		NULL on last dereference
 */
rpmmqtt rpmmqttUnlink (rpmmqtt mqtt);
#define	rpmmqttUnlink(_mqtt)	\
    ((rpmmqtt)rpmioUnlinkPoolItem((rpmioItem)(_mqtt), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a mqtt wrapper instance.
 * @param mqtt		mqtt wrapper
 * @return		new mqtt wrapper reference
 */
rpmmqtt rpmmqttLink (rpmmqtt mqtt);
#define	rpmmqttLink(_mqtt)	\
    ((rpmmqtt)rpmioLinkPoolItem((rpmioItem)(_mqtt), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a mqtt wrapper.
 * @param mqtt		mqtt wrapper
 * @return		NULL on last dereference
 */
rpmmqtt rpmmqttFree(rpmmqtt mqtt);
#define	rpmmqttFree(_mqtt)	\
    ((rpmmqtt)rpmioFreePoolItem((rpmioItem)(_mqtt), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a mqtt wrapper.
 * @param fn		mqtt args (or NULL)
 * @param flags		mqtt flags
 * @return		new mqtt wrapper
 */
rpmmqtt rpmmqttNew(char ** av, uint32_t flags);

void rpmmqttFreeMacros(rpmmqtt mqtt);
int rpmmqttDefineMacro(rpmmqtt mqtt, const char * macro, int level);
char * rpmmqttExpand(rpmmqtt mqtt, char **lhs, const char *arg, ...)
	RPM_GNUC_NULL_TERMINATED;
int rpmmqttExpandNumeric(rpmmqtt mqtt, const char *arg);

rpmRC rpmmqttPub(rpmmqtt mqtt, const char * topic, const char *s, size_t ns);

rpmRC rpmmqttSub(rpmmqtt mqtt, const char *s, size_t ns);

rpmRC rpmmqttRun(rpmmqtt mqtt, const char * str, const char ** resultp);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMMQTT */
