#include "system.h"

#include <rpmiotypes.h>
#include <argv.h>
#include <rpmlog.h>

#if defined(WITH_MRBEMBED)
#include <mruby.h>
#include <mruby/array.h>
#include <mruby/compile.h>
#include <mruby/dump.h>
#include <mruby/variable.h>
#endif

#define _RPMMRB_INTERNAL 1
#include "rpmmrb.h"

#include "debug.h"

/**
 * Finalizes a MRuby interpreter instance/pool item
 */
static void rpmmrbFini(void *_mrb)
{
    rpmmrb mrb = (rpmmrb) _mrb;

#if defined(WITH_MRBEMBED)
    mrb_state * I = (mrb_state *) mrb->I;
    mrbc_context * C = (mrbc_context *) mrb->C;
    if (C)
	mrbc_context_free(I, C);
    if (I)
	mrb_close(I);
#endif
    mrb->I = NULL;
}

int _rpmmrb_debug = 0;

#define	rpmmrbDbug	NULL
#define	rpmmrbInit	NULL
RPMIOPOOL_INTERP_MODULE(mrb)

#if defined(WITH_MRBEMBED)
#ifdef	NOTYET
/** Initializes MRuby's StringIO for storing output in a string. */
static const char * rpmmrbInitStringIO = "\
require 'stringio'\n\
$stdout = StringIO.new($result, \"w+\")\n\
";
#endif
#endif	/* WITH_MRBEMBED */

#if defined(WITH_MRBEMBED)
static void rpmmrbInfo(void)
{
    int _lvl = RPMLOG_DEBUG;

    rpmlog(_lvl, "==================== %s\n", MRUBY_DESCRIPTION);
}
#endif	/* WITH_MRBEMBED */

rpmmrb rpmmrbNew(char **av, uint32_t flags)
{
    static const char *_av[] = { "rpmmrb", NULL };
    rpmmrb mrb = (flags & 0x80000000)
	? rpmmrbI() : rpmmrbGetPool(_rpmmrbPool);

    if (av == NULL) av = (char **) _av;

#if defined(WITH_MRBEMBED)
    if (mrb->I == NULL) {
	static int oneshot;
	mrb_state * I = mrb_open();
	int ac = argvCount((ARGV_t)av);

	if (!oneshot) {
	    rpmmrbInfo();
	    oneshot++;
	}
assert(I);
	mrb->I = (void *) I;

	mrb_value ARGV = mrb_ary_new_capa(I, ac);
	for (int i = 0; i < ac; i++)
	    mrb_ary_push(I, ARGV, mrb_str_new_cstr(I, av[i]));
	mrb_define_global_const(I, "ARGV", ARGV);

	mrbc_context * C = mrbc_context_new(I);
assert(C);
	mrb->C = (void *) C;

				/* XXX -v and --version */
				/* XXX --copyright */
	C->dump_result = FALSE;	/* XXX -v and --verbose */
	C->no_exec = FALSE;	/* XXX -c --check */

	/* Set $0 */
	mrb_sym zero_sym = mrb_intern_lit(I, "$0");
#ifdef	NOTYET
	if (args.rfp) {		/* $0 arg is a file or stdin */
	    char * cmdline = args.cmdline ? args.cmdline : "-";
	    mrbc_filename(I, C, cmdline);
	    mrb_gv_set(I, zero_sym, mrb_str_new_cstr(I, cmdline));
	} else			/* $0 arg is -e */
#endif
	{
	    mrbc_filename(I, C, "-e");
	    mrb_gv_set(I, zero_sym, mrb_str_new_lit(I, "-e"));
	}

#ifdef	NOTYET
	(void) rpmmrbRun(mrb, rpmmrbInitStringIO, NULL);
#endif
    }
#endif	/* WITH_MRBEMBED */

    return rpmmrbLink(mrb);
}

rpmRC rpmmrbRun(rpmmrb mrb, const char *str, const char **resultp)
{
    rpmRC rc = RPMRC_FAIL;

if (_rpmmrb_debug)
fprintf(stderr, "==> %s(%p,%s,%p)\n", __FUNCTION__, mrb, str, resultp);

    if (mrb == NULL)
        mrb = rpmmrbI();

#if defined(WITH_MRBEMBED)
    if (str) {
	mrb_state * I = (mrb_state *) mrb->I;
	mrbc_context * C = (mrbc_context *) mrb->C;
	mrb_value v;

	/* Load program */
#ifdef	NOTYET
	if (args.mrbfile)	/* XXX -b uses binary irep */
	    v = mrb_load_irep_file_cxt(I, args.rfp, C);
	else if (args.rfp)	/* XXX arg is a file */
	    v = mrb_load_file_cxt(I, args.rfp, C);
	else
#endif
	    v = mrb_load_string_cxt(I, str, C);
	fflush(stdout);		/* XXX FIXME */

	if (I->exc) {
	    if (!mrb_undef_p(v))
		mrb_print_error(I);
	    rc = RPMRC_FAIL;
	} else if (C->no_exec) {
	    fprintf(stderr, "Syntax OK\n");
	    rc = RPMRC_OK;
	} else
	    rc = RPMRC_OK;
    }
#endif

    return rc;
}
