#ifndef	_H_RPMMSQ_
#define	_H_RPMMSQ_

/**
 */
extern int _rpmmsq_debug;

typedef	struct rpmmsq_s * rpmmsq;

/**
 */
typedef enum rpmmsqType_e {
    RPMMSQ_TYPE_DEFAULT		= 0,
    RPMMSQ_TYPE_POSIX		= 1,
    RPMMSQ_TYPE_SYSV		= 2,
} rpmmsqType;
#define	RPMMSQ_TYPE_MASK	0x7	/* XXX grow to 0x0F possible. */

/**
 */
#define _KFB(n) (1U << (n))
#define _MFB(n) (_KFB(n))
#define	RPMMSQ_FLAGS_TYPE_SHIFT	23
typedef enum rpmmsqFlags_e {
    RPMMSQ_FLAGS_NONE		= 0,
	/* bits 0-22 reserved  for linux O_* flags */
    RPMMSQ_FLAGS_POSIX	= (RPMMSQ_TYPE_POSIX << RPMMSQ_FLAGS_TYPE_SHIFT),
    RPMMSQ_FLAGS_SYSV	= (RPMMSQ_TYPE_SYSV  << RPMMSQ_FLAGS_TYPE_SHIFT),
	/* bits 23-26 reserved for (up to 16) MQ types. */
#define	RPMMSQ_FLAGS_TYPE(_f)	(((_f) >> RPMMSQ_FLAGS_TYPE_SHIFT) & RPMMSQ_TYPE_MASK)
    RPMMSQ_FLAGS_INFO		= _MFB(27),
    RPMMSQ_FLAGS_RESET		= _MFB(28),
    RPMMSQ_FLAGS_DELETE		= _MFB(29),
    RPMMSQ_FLAGS_DEBUGREFS	= _MFB(30),
    RPMMSQ_FLAGS_DEBUGIO	= _MFB(31),
} rpmmsqFlags;
#undef  _MFB
#undef  _KFB

#ifdef __cplusplus
extern "C" {
#endif

/**
 */
#if defined(_RPMMSQ_INTERNAL)
struct rpmmsq_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    struct AIOQ_s aioq;		/*!< message inflight queue */
    const char * qname;		/*!> message queue path. */
    const char * fmode;
    int fdno;
    rpmmsqFlags flags;		/*!< message queue option bits. */
    int qid;			/*!< message queue id. */
    int oflags;			/*!< message queue open flags. */
    int omode;			/*!< message queue open mode. */

    key_t key;			/*!< SysV: ftok(3) or IPC_PRIVATE. */
    long mtype;			/*!< SysV: message type. */

    int msgmax;			/*!< max. inflight messages. */
    int msgsize;		/*!< max. message size. */

    int nqueued;		/*!< no. messages queued. */
    int nsent;			/*!< no. messages sent. */
    int nrecv;			/*!< no. messages received. */
    int ntimeout;		/*!< no. of receive timeouts. */
    int nagain;			/*!< no. of waits (O_NONBLOCK). */

    struct mq_attr oattrs;	/*!< POSIX: mq_open attributes. */
    pthread_attr_t attr;	/*!< POSIX: mq_notify thread attr */
    struct sigevent sigev;	/*!< POSIX: mq_notify sigevent_t */

    struct stat *st;		/*!< POSIX: /dev/mqueue/qname */
    void *ds;			/*!< SysV: struct msqid_ds */
    void *mi;			/*!< SysV: struct msginfo */

    rpmzLog zlog;		/*!< high precision timestamp'd log */
};

rpmmsq rpmmsqOpen(const char * path, const char * fmode);
rpmmsq rpmmsqFdopen(void * _fdno, const char * fmode);
int rpmmsqFlush(void * _msq);
int rpmmsqCtl(void * _msq, int cmd, void *buf);
int rpmmsqNotify(rpmmsq msq, const void *_sevp);
void rpmmsqReader(union sigval sv);
int rpmmsqDump(const char *msg, void * _msq, FILE *fp);

#endif	/* _RPMMSQ_INTERNAL */

/**
 * Unreference a msq wrapper instance.
 * @param msq		msq wrapper
 * @return		NULL on last dereference
 */
rpmmsq rpmmsqUnlink (rpmmsq msq);
#define	rpmmsqUnlink(_msq)	\
    ((rpmmsq)rpmioUnlinkPoolItem((rpmioItem)(_msq), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a msq wrapper instance.
 * @param msq		msq wrapper
 * @return		new msq wrapper reference
 */
rpmmsq rpmmsqLink (rpmmsq msq);
#define	rpmmsqLink(_msq)	\
    ((rpmmsq)rpmioLinkPoolItem((rpmioItem)(_msq), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a msq wrapper.
 * @param msq		msq wrapper
 * @return		NULL on last dereference
 */
rpmmsq rpmmsqFree(rpmmsq msq);
#define	rpmmsqFree(_msq)	\
    ((rpmmsq)rpmioFreePoolItem((rpmioItem)(_msq), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create a msq wrapper.
 * @return		msq wrapper
 */
rpmmsq rpmmsqNew(const char * path, const char * fmode, int fdno, unsigned flags);

ssize_t rpmmsqRecv(rpmmsq msq, char * buf, size_t count, unsigned long *priop);

ssize_t rpmmsqSend(rpmmsq msq, const char * buf, size_t count, unsigned long prio);

int rpmmsqClose(rpmmsq msq, int delete);

#ifdef __cplusplus
}
#endif

#endif	/* _H_RPMMSQ_ */
