#ifndef RPMPERL_H
#define RPMPERL_H

/** \ingroup rpmio
 * \file rpmio/rpmperl.h
 */

#include <rpmiotypes.h>
#include <rpmio.h>

typedef struct rpmperl_s * rpmperl;

extern int _rpmperl_debug;

extern rpmperl _rpmperlI;

#if defined(_RPMPERL_INTERNAL)
struct rpmperl_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    void * I;
};
#endif /* _RPMPERL_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a perl interpreter instance.
 * @param perl		perl interpreter
 * @return		NULL on last dereference
 */
rpmperl rpmperlUnlink (rpmperl perl);
#define	rpmperlUnlink(_perl)	\
    ((rpmperl)rpmioUnlinkPoolItem((rpmioItem)(_perl), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a perl interpreter instance.
 * @param perl		perl interpreter
 * @return		new perl interpreter reference
 */
rpmperl rpmperlLink (rpmperl perl);
#define	rpmperlLink(_perl)	\
    ((rpmperl)rpmioLinkPoolItem((rpmioItem)(_perl), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a perl interpreter.
 * @param perl		perl interpreter
 * @return		NULL on last dereference
 */
rpmperl rpmperlFree(rpmperl perl);
#define	rpmperlFree(_perl)	\
    ((rpmperl)rpmioFreePoolItem((rpmioItem)(_perl), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a perl interpreter.
 * @param av		perl interpreter args (or NULL)
 * @param flags		perl interpreter flags ((1<<31): use global interpreter)
 * @return		new perl interpreter
 */
rpmperl rpmperlNew(char ** av, uint32_t flags);

/**
 * Execute perl string.
 * @param perl		perl interpreter (NULL uses global interpreter)
 * @param str		perl string to execute (NULL returns RPMRC_FAIL)
 * @param resultp	*resultp perl exec result
 * @return		RPMRC_OK on success
 */
rpmRC rpmperlRun(rpmperl perl, const char * str, const char ** resultp);

#ifdef __cplusplus
}
#endif

#endif /* RPMPERL_H */
