#ifndef RPMPYTHON_H
#define RPMPYTHON_H

/** \ingroup rpmio
 * \file rpmio/rpmpython.h
 */

#include <rpmiotypes.h>
#include <rpmio.h>

typedef struct rpmpython_s * rpmpython;

extern int _rpmpython_debug;

extern rpmpython _rpmpythonI;

#if defined(_RPMPYTHON_INTERNAL)
struct rpmpython_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    void * I;			/* (unused) */
};
#endif /* _RPMPYTHON_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a python interpreter instance.
 * @param python	python interpreter
 * @return		NULL on last dereference
 */
rpmpython rpmpythonUnlink (rpmpython python);
#define	rpmpythonUnlink(_python)	\
    ((rpmpython)rpmioUnlinkPoolItem((rpmioItem)(_python), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a python interpreter instance.
 * @param python	python interpreter
 * @return		new python interpreter reference
 */
rpmpython rpmpythonLink (rpmpython python);
#define	rpmpythonLink(_python)	\
    ((rpmpython)rpmioLinkPoolItem((rpmioItem)(_python), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a python interpreter.
 * @param python	python interpreter
 * @return		NULL on last dereference
 */
rpmpython rpmpythonFree(rpmpython python);
#define	rpmpythonFree(_python)	\
    ((rpmpython)rpmioFreePoolItem((rpmioItem)(_python), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a python interpreter.
 * @param av		python interpreter args (or NULL)
 * @param flags		python interpreter flags ((1<<31): use global interpreter)
 * @return		new python interpreter
 */
rpmpython rpmpythonNew(char **av, uint32_t flags);

/**
 * Execute python from a file.
 * @param python	python interpreter (NULL uses global interpreter)
 * @param fn		python file to run (NULL returns RPMRC_FAIL)
 * @param resultp	*resultp python exec result
 * @return		RPMRC_OK on success
 */
rpmRC rpmpythonRunFile(rpmpython python, const char *fn, const char **resultp);

/**
 * Execute python string.
 * @param python	python interpreter (NULL uses global interpreter)
 * @param str		python string to execute (NULL returns RPMRC_FAIL)
 * @param resultp	*resultp python exec result
 * @return		RPMRC_OK on success
 */
rpmRC rpmpythonRun(rpmpython python, const char *str, const char **resultp);

#ifdef __cplusplus
}
#endif

#endif /* RPMPYTHON_H */
