#include "system.h"

#include <rpmiotypes.h>
#include <rpmlog.h>
#include <argv.h>

#if defined(WITH_RUBYEMBED)

/* Make sure Ruby's fun stuff has its own xmalloc & Co functions available */
#undef xmalloc
#undef xcalloc
#undef xrealloc

#undef	PACKAGE_NAME
#undef	PACKAGE_TARNAME
#undef	PACKAGE_VERSION
#undef	PACKAGE_STRING
#undef	PACKAGE_BUGREPORT

#pragma GCC diagnostic ignored "-Wstrict-prototypes"
#define	HAVE_SETPROCTITLE	1	/* XXX ruby-2.1.5 prototype clash */
#include <ruby.h>
#pragma GCC diagnostic warning "-Wstrict-prototypes"

#endif

#define _RPMRUBY_INTERNAL 1
#include "rpmruby.h"

#include "debug.h"

/**
 * Finalizes a Ruby interpreter instance/pool item
 */
static void rpmrubyFini(void *_ruby)
{
    rpmruby ruby = (rpmruby) _ruby;

#if defined(WITH_RUBYEMBED)
    ruby_cleanup(0);
#endif
    ruby->I = NULL;
}

int _rpmruby_debug = 0;

#define	rpmrubyDbug	NULL
#define	rpmrubyInit	NULL
RPMIOPOOL_INTERP_MODULE(ruby)

#if defined(WITH_RUBYEMBED)
/** Initializes Ruby's StringIO for storing output in a string. */
static const char * rpmrubyInitStringIO = "\
require 'stringio'\n\
$stdout = StringIO.new($result, \"w+\")\n\
";
#endif

#if defined(WITH_RUBYEMBED)
static const char *rubyEnvVars[] = {
    "RUBYLIB",
    "RUBYLIB_PREFIX",
    "RUBYOPT",
    "RUBYPATH",
    "RUBYSHELL",
    NULL,
};
static int nRubyEnvVars = sizeof(rubyEnvVars) / sizeof(rubyEnvVars[0]);

#include <ruby/version.h>
static void rpmrubyInfo(void)
{
    const char **av = rubyEnvVars;
    int ac = nRubyEnvVars;
    char *t;
    int _lvl = RPMLOG_DEBUG;

    for (int i = 0; i < ac && av[i] != NULL; i++) {
	if ((t = getenv(av[i])) == NULL)
	    continue;
	rpmlog(_lvl, "    %s=\"%s\"\n", av[i], t);
    }

    rpmlog(_lvl, "==================== %s\n", ruby_description);
}
#endif  /* WITH_RUBYEMBED */

rpmruby rpmrubyNew(char **av, uint32_t flags)
{
    static const char *_av[] = { "rpmruby", NULL };
    rpmruby ruby;

    /* XXX FIXME: recurse like every other embedded interpreter. */
    if (_rpmrubyI)
	return _rpmrubyI;

    if (av == NULL) av = (char **) _av;

    ruby = rpmrubyGetPool(_rpmrubyPool);

#if defined(WITH_RUBYEMBED)
    static int oneshot;

    if (!oneshot) {
	rpmrubyInfo();
	oneshot++;
    }

    RUBY_INIT_STACK;
    ruby_init();
    ruby_init_loadpath();

    ruby_script((char *)av[0]);
    rb_gv_set("$result", rb_str_new2(""));
    (void) rpmrubyRun(ruby, rpmrubyInitStringIO, NULL);
#endif

    return rpmrubyLink(ruby);
}


rpmRC rpmrubyRun(rpmruby ruby, const char *str, const char **resultp)
{
    rpmRC rc = RPMRC_FAIL;

if (_rpmruby_debug)
fprintf(stderr, "==> %s(%p,%s,%p)\n", __FUNCTION__, ruby, str, resultp);

    if (ruby == NULL)
        ruby = rpmrubyI();

#if defined(WITH_RUBYEMBED)
    if (str) {
	int state = -1;
	ruby->state = rb_eval_string_protect(str, &state);

	/* Check whether the evaluation was successful or not */

	if (state == 0) {
	    rc = RPMRC_OK;
	    if (resultp)
		*resultp = RSTRING_PTR(rb_gv_get("$result"));
	}
    }
#endif

    return rc;
}

