#ifndef	H_RPMSP
#define	H_RPMSP

/** \ingroup rpmio
 * \file rpmio/rpmsp.h
 */

#include <rpmiotypes.h>
#include <rpmio.h>

typedef struct rpmsp_s * rpmsp;

extern int _rpmsp_debug;

extern rpmsp _rpmspI;

#if defined(_RPMSP_INTERNAL)
/** \ingroup rpmio
 */
struct rpmsp_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * fn;
    unsigned int flags;
    struct sepol_handle * I;		/*!< sepol_handle_t */
    struct sepol_policydb * DB;		/*!< sepol_policydb_t */
    struct sepol_policy_file * F;	/*!< sepol_policy_file_t */
    struct sepol_context * C;		/*!< sepol_context_t */
    struct sepol_module_package * P;	/*!< sepol_module_package_t */
};
#endif	/* _RPMSP_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a sepol wrapper instance.
 * @param sp		sepol wrapper
 * @return		NULL on last dereference
 */
rpmsp rpmspUnlink (rpmsp sp);
#define	rpmspUnlink(_sp)	\
    ((rpmsp)rpmioUnlinkPoolItem((rpmioItem)(_sp), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a sepol wrapper instance.
 * @param sp		sepol wrapper
 * @return		new sepol wrapper reference
 */
rpmsp rpmspLink (rpmsp sp);
#define	rpmspLink(_sp)	\
    ((rpmsp)rpmioLinkPoolItem((rpmioItem)(_sp), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a sepol wrapper.
 * @param sp		sepol wrapper
 * @return		NULL on last dereference
 */
rpmsp rpmspFree(rpmsp sp);
#define	rpmspFree(_sp)	\
    ((rpmsp)rpmioFreePoolItem((rpmioItem)(_sp), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a sepol wrapper.
 * @param fn		sepol file (unused).
 * @param flags		sepol flags
 * @return		new sepol wrapper
 */
rpmsp rpmspNew(const char * fn, unsigned int flags);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMSP */
