#ifndef RPMSQUIRREL_H
#define RPMSQUIRREL_H

/** \ingroup rpmio
 * \file rpmio/rpmsquirrel.h
 */

#include <rpmiotypes.h>
#include <rpmio.h>

typedef struct rpmsquirrel_s * rpmsquirrel;

extern int _rpmsquirrel_debug;

extern rpmsquirrel _rpmsquirrelI;

#if defined(_RPMSQUIRREL_INTERNAL)
struct rpmsquirrel_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    void * I;			/* HSQUIRRELVM */
    rpmiob iob;
};
#endif /* _RPMSQUIRREL_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a squirrel interpreter instance.
 * @param squirrel		squirrel interpreter
 * @return		NULL on last dereference
 */
rpmsquirrel rpmsquirrelUnlink (rpmsquirrel squirrel);
#define	rpmsquirrelUnlink(_squirrel)	\
    ((rpmsquirrel)rpmioUnlinkPoolItem((rpmioItem)(_squirrel), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a squirrel interpreter instance.
 * @param squirrel		squirrel interpreter
 * @return		new squirrel interpreter reference
 */
rpmsquirrel rpmsquirrelLink (rpmsquirrel squirrel);
#define	rpmsquirrelLink(_squirrel)	\
    ((rpmsquirrel)rpmioLinkPoolItem((rpmioItem)(_squirrel), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a squirrel interpreter.
 * @param squirrel		squirrel interpreter
 * @return		NULL on last dereference
 */
rpmsquirrel rpmsquirrelFree(rpmsquirrel squirrel);
#define	rpmsquirrelFree(_squirrel)	\
    ((rpmsquirrel)rpmioFreePoolItem((rpmioItem)(_squirrel), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a squirrel interpreter.
 * @param av		squirrel interpreter args (or NULL)
 * @param flags		squirrel interpreter flags ((1<<31): use global interpreter)
 * @return		new squirrel interpreter
 */
rpmsquirrel rpmsquirrelNew(char ** av, unsigned flags);

/**
 * Execute squirrel from a file.
 * @param squirrel		squirrel interpreter (NULL uses global interpreter)
 * @param fn		squirrel file to run (NULL returns RPMRC_FAIL)
 * @param resultp	*resultp squirrel exec result
 * @return		RPMRC_OK on success
 */
rpmRC rpmsquirrelRunFile(rpmsquirrel squirrel, const char * fn,
		const char ** resultp);

/**
 * Execute squirrel string.
 * @param squirrel		squirrel interpreter (NULL uses global interpreter)
 * @param str		squirrel string to execute (NULL returns RPMRC_FAIL)
 * @param resultp	*resultp squirrel exec result
 * @return		RPMRC_OK on success
 */
rpmRC rpmsquirrelRun(rpmsquirrel squirrel, const char * str,
		const char ** resultp);

#ifdef __cplusplus
}
#endif

#endif /* RPMSQUIRREL_H */
