/** \ingroup rpmio
 * \file rpmio/rpmsvn.c
 */

#include "system.h"

#include <rpmiotypes.h>
#include <rpmio.h>	/* for *Pool methods */
#include <rpmlog.h>
#include <rpmurl.h>

#if defined(WITH_SUBVERSION)
#if __clang__
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdocumentation"
#endif
#include "svn_pools.h"
#include "svn_client.h"
#include "svn_repos.h"
#include "svn_subst.h"
#if __clang__
#pragma clang diagnostic pop
#endif
#endif

#undef	HAVE_APR_1_APR_H
#define _RPMSVN_INTERNAL
#include <rpmsvn.h>

#include "debug.h"

/*==============================================================*/

/*==============================================================*/

#if defined(WITH_SUBVERSION)
static int npools;
#endif

static void rpmsvnFini(void * _svn)
{
    rpmsvn svn = (rpmsvn) _svn;

#if defined(WITH_SUBVERSION)
    if (svn->pool) {
	svn_pool_destroy(svn->pool);
	svn->pool = NULL;
    }
    if (--npools <= 0)
	apr_terminate();
    svn->allocator = NULL;
#endif

    svn->fn = _free(svn->fn);

}

int _rpmsvn_debug = 0;

#define	rpmsvnDbug	NULL
#define	rpmsvnInit	NULL
RPMIOPOOL_MODULE(svn)

rpmsvn rpmsvnNew(const char * fn, int flags)
{
    rpmsvn svn = rpmsvnGetPool(_rpmsvnPool);

#if defined(WITH_SUBVERSION)
    int xx;
    if (npools++ <= 0) {
	xx = apr_initialize();
assert(xx == APR_SUCCESS);
    }
    xx = apr_allocator_create((apr_allocator_t **)&svn->allocator);
assert(xx == 0);
    apr_allocator_max_free_set(svn->allocator, SVN_ALLOCATOR_RECOMMENDED_MAX_FREE);
    svn->pool = svn_pool_create_ex(NULL, svn->allocator);
    apr_allocator_owner_set(svn->allocator, svn->pool);

#endif

    if (fn)
	svn->fn = xstrdup(fn);


    return rpmsvnLink(svn);
}
