#ifndef	H_RPMSX
#define	H_RPMSX

/** \ingroup rpmio
 * \file rpmio/rpmsx.h
 */

#include <rpmiotypes.h>
#include <rpmio.h>

typedef struct rpmsx_s * rpmsx;

extern int _rpmsx_debug;

extern rpmsx _rpmsxI;

#if defined(_RPMSX_INTERNAL)
/** \ingroup rpmio
 */
struct rpmsx_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * fn;		/*!< File context path (NULL disables). */
    unsigned int flags;
};
#endif	/* _RPMSX_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a SELinux wrapper instance.
 * @param sx		SELinux wrapper
 * @return		NULL on last dereference
 */
rpmsx rpmsxUnlink (rpmsx sx);
#define	rpmsxUnlink(_sx)	\
    ((rpmsx)rpmioUnlinkPoolItem((rpmioItem)(_sx), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a SELinux wrapper instance.
 * @param sx		SELinux wrapper
 * @return		new SELinux wrapper reference
 */
rpmsx rpmsxLink (rpmsx sx);
#define	rpmsxLink(_sx)	\
    ((rpmsx)rpmioLinkPoolItem((rpmioItem)(_sx), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a SELinux wrapper.
 * @param sx		SELinux wrapper
 * @return		NULL on last dereference
 */
rpmsx rpmsxFree(rpmsx sx);
#define	rpmsxFree(_sx)	\
    ((rpmsx)rpmioFreePoolItem((rpmioItem)(_sx), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create and load a SELinux wrapper.
 * @param fn		SELinux file (NULL uses selinux_file_context_path()).
 * @param flags		SELinux set_matchpathcon_flags() flags
 * @return		new SELinux wrapper
 */
rpmsx rpmsxNew(const char * fn, unsigned int flags);

/**
 * Return SELinux enabled state.
 * @param sx		SELinux wrapper (NULL uses active context)
 * @return		SELinux enabled state
 */
int rpmsxEnabled(rpmsx sx);

/**
 * Return security context for a file.
 * @param sx		SELinux wrapper (NULL uses active context)
 * @param fn		file path
 * @param mode		file mode
 * @return		file security context
 */
const char * rpmsxMatch(rpmsx sx, const char * fn, mode_t mode);

/**
 * Get security context for a file.
 * @param sx		SELinux wrapper (NULL uses active context)
 * @param fn		file path
 * @return		file security context (NULL on error/disabled)
 */
const char * rpmsxGetfilecon(rpmsx sx, const char *fn);
const char * rpmsxLgetfilecon(rpmsx sx, const char *fn);

/**
 * Set security context for a file.
 * @param sx		SELinux wrapper (NULL uses active context)
 * @param fn		file path
 * @param mode		file mode
 * @param scon		file security context (NULL calls matchpathcon())
 * @return		0 on success
 */
int rpmsxSetfilecon(rpmsx sx, const char *fn, mode_t mode, const char * scon);
int rpmsxLsetfilecon(rpmsx sx, const char *fn, mode_t mode, const char * scon);

/**
 * Execute a package scriptlet within SELinux context.
 * @param sx		SELinux wrapper
 * @param verified	Scriptlet came from signature verified header? (unused)
 * @param argv		scriptlet helper
 * @return		0 on success
 */
int rpmsxExec(rpmsx sx, int verified, const char ** argv);

#ifdef __cplusplus
}
#endif

#endif	/* H_RPMSX */
