#ifndef H_RPMSYCK
#define H_RPMSYCK

typedef struct rpmsyck_s * rpmsyck;

#ifdef	_RPMSYCK_INTERNAL
#include <rpmhash.h>
typedef enum {
    T_END,
    T_STR,
    T_SEQ,
    T_MAP
} syck_type_t;

typedef struct rpmsyck_node_s * rpmsyck_node;
struct rpmsyck_node_s {
    syck_type_t type;
    char *tag;
    union {
	char *key;
    	rpmsyck_node seq;
	hashTable map;
    } value;
};

struct rpmsyck_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    rpmsyck_node firstNode;
    void * syms;	/* XXX st_table * */
};
#endif

/**
 * Create rpmsyck YAML document.
 * @return		pointer to initialized YAML document
 */
rpmsyck rpmsyckNew(void);

/**
 * Load rpmsyck YAML document from text.
 * @param yaml		string variable holding the YAML document as text.
 * @return		pointer to initialized YAML document
 */
rpmsyck rpmsyckLoad(char *yaml);

/**
 * Unreference a rpmsyck YAML document instance.
 * @param syck		YAML document
 * @return		NULL on last dereference
 */
rpmsyck rpmsyckUnlink (rpmsyck syck);
#define	rpmsyckUnlink(_syck)	\
    ((rpmsyck)rpmioUnlinkPoolItem((rpmioItem)(_syck), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a rpmsyck YAML document instance.
 * @param syck		YAML document
 * @return		new YAML document reference
 */
rpmsyck rpmsyckLink (rpmsyck syck);
#define	rpmsyckLink(_syck)	\
    ((rpmsyck)rpmioLinkPoolItem((rpmioItem)(_syck), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy rpmsyck YAML document.
 * @param syck          pointer to YAML document
 * @return		NULL on last dereference
 */
rpmsyck rpmsyckFree(rpmsyck syck);
#define	rpmsyckFree(_syck)	\
    ((rpmsyck)rpmioFreePoolItem((rpmioItem)(_syck), __FUNCTION__, __FILE__, __LINE__))

#endif /* H_RPMSYCK */
