#include "system.h"

#include <rpmiotypes.h>
#include <rpmlog.h>
#include <argv.h>

#ifdef	WITH_TCL
#include <tcl.h>
#endif
#define _RPMTCL_INTERNAL
#include "rpmtcl.h"

#include "debug.h"

static void rpmtclFini(void * _tcl)
{
    rpmtcl tcl = (rpmtcl) _tcl;

#if defined(WITH_TCL)
    Tcl_DeleteInterp((Tcl_Interp *)tcl->I);
#endif
    tcl->I = NULL;
    (void)rpmiobFree(tcl->iob);
    tcl->iob = NULL;
}

int _rpmtcl_debug = 0;

#define	rpmtclDbug	NULL
#define	rpmtclInit	NULL
RPMIOPOOL_INTERP_MODULE(tcl)

#if defined(WITH_TCL)
static int rpmtclIOclose(ClientData CD, Tcl_Interp *I)
{
if (_rpmtcl_debug)
fprintf(stderr, "==> %s(%p, %p)\n", __FUNCTION__, CD, I);
    return 0;
}

static int rpmtclIOread(ClientData CD, char *b, int nb, int *errnop)
{
if (_rpmtcl_debug)
fprintf(stderr, "==> %s(%p, %p[%d], %p)\n", __FUNCTION__, CD, b, nb, errnop);
    *errnop = EINVAL;
    return -1;
}

static int rpmtclIOwrite(ClientData CD, const char *b, int nb, int *errnop)
{
    rpmtcl tcl = (rpmtcl) CD;
if (_rpmtcl_debug)
fprintf(stderr, "==> %s(%p, %p[%d], %p)\n", __FUNCTION__, CD, b, nb, errnop);
    if (nb > 0) {
	char * t = (char *)b;
	int c = t[nb];
	if (c) t[nb] = '\0';
	(void) rpmiobAppend(tcl->iob, b, 0);
	if (c) t[nb] = c;
    }
    return nb;
}

static int rpmtclIOseek(ClientData CD, long off, int mode, int *errnop)
{
if (_rpmtcl_debug)
fprintf(stderr, "==> %s(%p, %ld, %d, %p)\n", __FUNCTION__, CD, off, mode, errnop);
    *errnop = EINVAL;
    return -1;
}

static Tcl_ChannelType rpmtclIO = {
    (char *)"rpmtclIO",		/* Type name */
    TCL_CHANNEL_VERSION_2,	/* Tcl_ChannelTypeVersion */
    rpmtclIOclose,		/* Tcl_DriverCloseProc */
    rpmtclIOread,		/* Tcl_DriverInputProc */
    rpmtclIOwrite,		/* Tcl_DriverOutputProc */
    rpmtclIOseek,		/* Tcl_DriverSeekProc */
    NULL,			/* Tcl_DriverSetOptionProc */
    NULL,			/* Tcl_DriverGetOptionProc */
    NULL,			/* Tcl_DriverWatchProc */
    NULL,			/* Tcl_DriverGetHandleProc */
    NULL,			/* Tcl_DriverClose2Proc */
    NULL,			/* Tcl_DriverBlockModeProc */
    NULL,			/* Tcl_DriverFlushProc */
    NULL,			/* Tcl_DriverHandlerProc */
    NULL,			/* Tcl_DriverWideSeekProc */
    NULL,			/* Tcl_DriverThreadActionProc */
#if TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION > 4
    NULL,			/* Tcl_DriverTruncateProc */
#endif
};
#endif

#if defined(WITH_TCL)
static void rpmtclInfo(void)
{
    int _lvl = RPMLOG_DEBUG;

    rpmlog(_lvl, "==================== tcl-%s\n", TCL_PATCH_LEVEL);
}
#endif	/* WITH_TCL */

rpmtcl rpmtclNew(char ** av, uint32_t flags)
{
    static const char * _av[] = { "rpmtcl", NULL };
    rpmtcl tcl = (flags & 0x80000000)
	? rpmtclI() : rpmtclGetPool(_rpmtclPool);

    if (av == NULL) av = (char **) _av;
    if (tcl->iob == NULL)
	tcl->iob = rpmiobNew(0);

#if defined(WITH_TCL)
    if (tcl->I == NULL) {
	static int oneshot;
	Tcl_Interp * tclI = Tcl_CreateInterp();
	int ac = argvCount((ARGV_t)av);
	char b[32];

	if (!oneshot) {
	    rpmtclInfo();
	    oneshot++;
	}

	Tcl_SetVar(tclI, "argv",
		Tcl_Merge(ac-1, (const char *const *)av+1), TCL_GLOBAL_ONLY);
	(void)sprintf(b, "%d", ac-1);
	Tcl_SetVar(tclI, "argc", b, TCL_GLOBAL_ONLY);
	Tcl_SetVar(tclI, "argv0", av[0], TCL_GLOBAL_ONLY);
	Tcl_SetVar(tclI, "tcl_interactive", "0", TCL_GLOBAL_ONLY);

	{   Tcl_Channel tclout = Tcl_GetStdChannel(TCL_STDOUT);
	    Tcl_SetChannelOption(tclI, tclout, "-translation", "auto");
	    Tcl_StackChannel(tclI, &rpmtclIO, tcl, TCL_WRITABLE, tclout);
	    tcl->tclout = (void *) tclout;
	}
	tcl->I = tclI;
    }
#endif
    return rpmtclLink(tcl);
}

rpmRC rpmtclRunFile(rpmtcl tcl, const char * fn, const char ** resultp)
{
    rpmRC rc = RPMRC_FAIL;

if (_rpmtcl_debug)
fprintf(stderr, "==> %s(%p,%s)\n", __FUNCTION__, tcl, fn);

    if (tcl == NULL) tcl = rpmtclI();

#if defined(WITH_TCL)
    if (fn != NULL && Tcl_EvalFile((Tcl_Interp *)tcl->I, fn) == TCL_OK) {
	rc = RPMRC_OK;
	if (resultp)
	    *resultp = rpmiobStr(tcl->iob);
    }
#endif
    return rc;
}

rpmRC rpmtclRun(rpmtcl tcl, const char * str, const char ** resultp)
{
    rpmRC rc = RPMRC_FAIL;

if (_rpmtcl_debug)
fprintf(stderr, "==> %s(%p,%s)\n", __FUNCTION__, tcl, str);

    if (tcl == NULL) tcl = rpmtclI();

#if defined(WITH_TCL)
    if (str != NULL && Tcl_Eval((Tcl_Interp *)tcl->I, str) == TCL_OK) {
	rc = RPMRC_OK;
	if (resultp) {
	    *resultp = rpmiobStr(tcl->iob);
	    {	/* XXX tcl needs an rtrim */
		size_t len = strlen(*resultp);
		char *be = (char *) (*resultp + len);
		if (be > *resultp) be--;
		while (strchr("\r\n", *be) != NULL) {
		    *be-- = '\0';
		    len--;
		}
	    }
	}
    }
#endif
    return rc;
}
