#include "system.h"

#include <rpmio.h>
#include <rpmmacro.h>

#define	_RPMW_INTERNAL
#include <rpmw.h>

#include "debug.h"

int _rpmw_debug = 0;

#define SPEW(_fmt, ...) \
    if (_rpmw_debug) \
	fprintf(stderr, _fmt, __VA_ARGS__)

/*==============================================================*/
void rpmwDump(const char * msg, rpmw w, FILE * fp)
{
    if (fp == NULL)	fp = stderr;
    if (msg)	fprintf(fp, "========================== %s(%p)\n", msg, w);
    if (w) {
#define PRINT(_fmt, _foo) \
    {   fprintf(fp, "%25s: %"#_fmt"\n", #_foo, w->_foo); }

	if (w->nints)		PRINT(u, nints);
	if (w->nptrs)		PRINT(u, nptrs);
	if (w->noffs)		PRINT(u, noffs);
	if (w->nsizes)		PRINT(u, nsizes);
	if (w->ndoubles)	PRINT(u, ndoubles);
	if (w->nlongs)		PRINT(u, nlongs);
	if (w->nlonglongs)	PRINT(u, nlonglongs);
	if (w->nother)		PRINT(u, nother);

	rpmDumpMacroTable(w->mc, NULL);

#undef	PRINT
    }
}

/*==============================================================*/
static void rpmwFini(void *_w)
{
    rpmw w = (rpmw) _w;

    if (w) {
	w->mc = rpmmcFree(w->mc);

#ifdef	DYING
	w->subtype = _free(w->subtype);
	w->subname = _free(w->subname);
#endif

    }
}

#define	rpmwDbug	NULL
#define	rpmwInit	NULL
RPMIOPOOL_MODULE(w)

rpmw rpmwNew(void)
{
    rpmw w = rpmwGetPool(_rpmwPool);
    w->mc = rpmmcNew(NULL, 0);
    return rpmwLink(w);
}

