#ifndef H_RPMXAR
#define H_RPMXAR

/**
 * \file rpmio/rpmxar.h
 * Structure(s)and methods for a XAR archive wrapper format.
 */

#include <rpmiotypes.h>

extern int _xar_debug;

typedef struct rpmxar_s * rpmxar;

#ifdef	_RPMXAR_INTERNAL
#include "yarn.h"
struct rpmxar_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const void * x;		/*!< xar_t */
    const void * f;		/*!< xar_file_t */
    const void * i;		/*!< xar_iter_t */
    const char * member;	/*!< Current archive member. */
    unsigned char * b;		/*!< Data buffer. */
    size_t bsize;		/*!< No. bytes of data. */
    size_t bx;			/*!< Data byte index. */
    int first;
};
#endif

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Unreference a xar archive instance.
 * @param xar		xar archive
 * @param msg
 * @return		NULL on last dereference
 */
rpmxar rpmxarUnlink (rpmxar xar, const char * msg);
#define	rpmxarUnlink(_xar, _msg)	\
    ((rpmxar)rpmioUnlinkPoolItem((rpmioItem)(_xar), _msg, __FILE__, __LINE__))

/**
 * Reference a xar archive instance.
 * @param xar		xar archive
 * @param msg
 * @return		new xar archive reference
 */
rpmxar rpmxarLink (rpmxar xar, const char * msg);
#define	rpmxarLink(_xar, _msg)	\
    ((rpmxar)rpmioLinkPoolItem((rpmioItem)(_xar), _msg, __FILE__, __LINE__))

/**
 * Destroy a xar archive instance.
 * @param xar		xar archive
 * @param msg
 * @return		NULL on last dereference
 */
rpmxar rpmxarFree(rpmxar xar, const char * msg);
#define	rpmxarFree(_xar, _msg)	\
    ((rpmxar)rpmioFreePoolItem((rpmioItem)(_xar), _msg, __FILE__, __LINE__))

/**
 * Create a xar archive instance.
 * @param fn		xar file
 * @param fmode		"r" for reading, "w" for writing
 * @return		new xar instance
 */
rpmxar rpmxarNew(const char * fn, const char * fmode);

/**
 * Iterate a xar archive instance.
 * @param xar		xar archive
 * @return		0 on SUCCESS, 1 on end-of-iteration
 */
int rpmxarNext(rpmxar xar);

int rpmxarPush(rpmxar xar, const char * fn, unsigned char * b, size_t bsize);

int rpmxarPull(rpmxar xar, const char * fn)
	RPM_GNUC_CONST;

int rpmxarSwapBuf(rpmxar xar, unsigned char * b, size_t bsize,
		unsigned char ** obp, size_t * obsizep);

ssize_t xarRead(void * cookie, char * buf, size_t count);

/**
 * Return path of current archive member.
 * @param xar		xar archive
 * @return		path of xar member
 */
const char * rpmxarPath(rpmxar xar);

/**
 * Return stat(2) of current archive member.
 * @param xar		xar archive
 * @retval *st		stat(2) of current member
 * @return		0 on success
 */
int rpmxarStat(rpmxar xar, struct stat * st);

#ifdef __cplusplus
}
#endif

#endif  /* H_RPMXAR */
