#ifndef _H_RPMZLOG_
#define _H_RPMZLOG_

/** \ingroup rpmio
 * \file rpmio/rpmzlog.h
 */
#include <stdio.h>
#include <rpmdefs.h>

extern int _rpmzlog_debug;

/** trace log pointer */
typedef struct rpmzLog_s * rpmzLog;

#ifdef	_RPMZLOG_INTERNAL
#include <sys/time.h>
#include <yarn.h>

/** trace msg pointer */
typedef struct rpmzMsg_s * rpmzMsg;

/** trace msg */
struct rpmzMsg_s {
    struct timeval when;	/* time of entry */
    char *msg;			/* message */
    rpmzMsg next;		/* next entry */
};

/** trace log */
struct rpmzLog_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    struct timeval start;	/*!< starting time of day for tracing */
    rpmzMsg msg_head;
    rpmzMsg *msg_tail;
    int msg_count;
};
#endif	/* _RPMZLOG_INTERNAL */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Reference the log data.
 */
rpmzLog rpmzLogLink(rpmzLog zlog);

/**
 * Set up log (call from main thread before other threads launched).
 */
struct timeval;
rpmzLog rpmzLogNew(struct timeval *tv);

/**
 * Add entry to trace log.
 */
void rpmzLogAdd(rpmzLog zlog, const char *fmt, ...)
	RPM_GNUC_PRINTF(2, 3);

/**
 * Release a reference to the log data.
 */
rpmzLog rpmzLogFree(rpmzLog zlog);

/**
 * Show entries until no more, free log.
 */
rpmzLog rpmzLogDump(rpmzLog zlog, FILE * fp);

#ifdef __cplusplus
}
#endif

#endif /* _H_RPMZLOG_ */
