#ifndef	_H_RPMZSTD_
#define	_H_RPMZSTD_

/**
 */
extern int _rpmzstd_debug;

typedef	struct rpmzstd_s * rpmzstd;

/**
 */
#define _KFB(n) (1U << (n))
#define _MFB(n) (_KFB(n))
typedef enum rpmzstdFlags_e {
    RPMZSTD_FLAGS_NONE		= 0,
	/* bits 0 - 22 reserved  for linux O_* flags */
	/* bits 23 - 28 unused */
    RPMZSTD_FLAGS_DECOMPRESS	= _MFB(29),
    RPMZSTD_FLAGS_DEBUGREFS	= _MFB(30),
    RPMZSTD_FLAGS_DEBUGIO	= _MFB(31),
} rpmzstdFlags;
#undef  _MFB
#undef  _KFB

#ifdef __cplusplus
extern "C" {
#endif

/**
 */
#if defined(_RPMZSTD_INTERNAL)
struct rpmzstd_s {
    struct rpmioItem_s _item;	/*!< usage mutex and pool identifier. */
    const char * path;		/*!> open path. */
    const char * fmode;
    rpmzstdFlags flags;
    int fdno;
    int oflags;			/*!< open flags. */
    int omode;			/*!< open mode. */

    size_t nextnb;
    int level;

    FILE * fp;
    void * _stream;		/*!< ZSTD_{C,D}Stream */

    size_t nb;
    void * b;

#if defined(WITH_ZSTD)
    size_t nib;
    ZSTD_inBuffer zib;		/*!< ZSTD_inBuffer */
    size_t nob;
    ZSTD_outBuffer zob;		/*!< ZSTD_outBuffer */
#endif
};

#ifdef	NOTYET
int rpmzstdSeek(void * _zstd, _libio_pos_t pos, int whence)
#endif
rpmzstd rpmzstdOpen(const char * path, const char * fmode, int fdno, unsigned flags);
rpmzstd rpmzstdFdopen(void * _fdno, const char * fmode);
int rpmzstdDump(const char *msg, void * _zstd, FILE *fp);

#endif	/* _RPMZSTD_INTERNAL */

/**
 * Unreference a zstd wrapper instance.
 * @param zstd		zstd wrapper
 * @return		NULL on last dereference
 */
rpmzstd rpmzstdUnlink (rpmzstd zstd);
#define	rpmzstdUnlink(_zstd)	\
    ((rpmzstd)rpmioUnlinkPoolItem((rpmioItem)(_zstd), __FUNCTION__, __FILE__, __LINE__))

/**
 * Reference a zstd wrapper instance.
 * @param zstd		zstd wrapper
 * @return		new zstd wrapper reference
 */
rpmzstd rpmzstdLink (rpmzstd zstd);
#define	rpmzstdLink(_zstd)	\
    ((rpmzstd)rpmioLinkPoolItem((rpmioItem)(_zstd), __FUNCTION__, __FILE__, __LINE__))

/**
 * Destroy a zstd wrapper.
 * @param zstd		zstd wrapper
 * @return		NULL on last dereference
 */
rpmzstd rpmzstdFree(rpmzstd zstd);
#define	rpmzstdFree(_zstd)	\
    ((rpmzstd)rpmioFreePoolItem((rpmioItem)(_zstd), __FUNCTION__, __FILE__, __LINE__))

/**
 * Create a zstd wrapper.
 * @return		zstd wrapper
 */
rpmzstd rpmzstdNew(const char * path, const char * fmode, int fdno, unsigned flags);

ssize_t rpmzstdRead(rpmzstd zstd, void * b, size_t nb);

ssize_t rpmzstdWrite(rpmzstd zstd, const void * b, size_t nb);

int rpmzstdFlush(void * _zstd);

int rpmzstdClose(rpmzstd zstd);

#ifdef __cplusplus
}
#endif

#endif	/* _H_RPMZSTD_ */
