/**
 * \file rpmio/tcap.c
 */

#include "system.h"
#include <rpmio.h>
#include <poptIO.h>
#include <argv.h>
#include <rpmcap.h>
#include "debug.h"

static int _debug = 0;

static struct poptOption rpmcapOptionsTable[] = {
 { "debug", 'd', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN,		&_debug, 1,
	NULL, NULL },

 { NULL, '\0', POPT_ARG_INCLUDE_TABLE, rpmioAllPoptTable, 0,
	N_(" Common options for all rpmio executables:"), NULL },

  POPT_AUTOALIAS
  POPT_AUTOHELP
  POPT_TABLEEND
};

int
main(int argc, char *argv[])
{
    poptContext con = rpmioInit(argc, argv, rpmcapOptionsTable);
    ARGV_t av = poptGetArgs(con);
    int ac = argvCount(av);;
    pid_t pid = getpid();
    char * text;
    int rc;
    int ec = 0;

    text = rpmcapGetP(pid);
fprintf(stderr, "*** rpmcapGetP(%u) \"%s\"\n", (unsigned)pid, text);
    if (text) free(text);

    rc = rpmcapDrop("cap_chown");

    for (int i = 0; i < ac; i++) {
	const char * arg = av[i];
	rc = rpmcapTest(arg);
fprintf(stderr, "*** rpmcapGet(%s): %d\n", arg, rc);
    }

    con = rpmioFini(con);

#ifdef	DYING
fprintf(stderr, "==> rpmioMallopt(0, 0)\n");
    rpmioMallopt(0,0);
fprintf(stderr, "==> rpmioMallocstats()\n");
    rpmioMallocStats();

fprintf(stderr, "==> rpmioMalloSetState(rpmioMallocGetState())\n");
    rpmioMallocSetState(rpmioMallocGetState());
#endif

    return ec;
}
