#include "system.h"

#include <rpmdefs.h>
#include <stdlib.h>
#include <stdint.h>

#include "cpuid.h"

#define RPM_GNUC_CLEANUP( free )              \
  __attribute__((__cleanup__(free)))

#if 0
#define RPM_GNUC_ANNOTATE( attr )              \
  __attribute__((__annotate__(attr)))
#else
#define RPM_GNUC_ANNOTATE( attr )
#endif

#define RPM_GNUC_VARIABLE_SIZE \
  __attribute__((__bnd_variable_size__))

#define RPM_GNUC_ENDIAN_BIG \
  __attribute__((scalar_storage_order("big-endian")))

#define RPM_GNUC_ENDIAN_LITTLE \
  __attribute__((scalar_storage_order("little-endian")))

#define RPM_GNUC_MODE( mode )              \
  __attribute__((__mode__(mode)))

typedef	__uint128_t uint128_t;

#define RPM_GNUC_TRANSPARENT_UNION \
  __attribute__((__transparent_union__))

#define RPM_GNUC_DESIGNATED_INIT \
  __attribute__((__designated_init__))

#define RPM_GNUC_IFUNC( resolver ) \
  __attribute__((__ifunc__(resolver)))

typedef char const * argv_t[];

typedef union RPM_GNUC_TRANSPARENT_UNION tagData_u {
    void	 *ptr;
    uint64_t	 *ui64p;
    uint32_t	 *ui32p;
    uint16_t	 *ui16p;
    uint8_t	 *ui8p;
    const char	 *str;
    const char	**argv;
} tagData;

typedef struct rpmfoo_s * rpmfoo;
struct rpmfoo_s {
    tagData  p;
    int		i;
    uint64_t ui64;
    uint32_t ui32;
    uint16_t ui16;
    uint8_t   ui8;
    char     data[1]	RPM_GNUC_VARIABLE_SIZE;
} RPM_GNUC_DESIGNATED_INIT;
// RPM_GNUC_ENDIAN_BIG;


static void
xfree(char ** b)
{
fprintf(stderr, "--> %s(%p) *b %p\n", __FUNCTION__, b, *b);
    if (*b)
	free(*b);
}

#ifdef	__cplusplus
#undef	offsetof
#define	offsetof(STRUCT, MEMBER)	(size_t)&(((STRUCT *)NULL)->MEMBER)
#endif

#define	PRINT(FMT, PTR, MEMBER) \
    fprintf(stderr, "%19s: size(%u) offset(%u) %"#FMT"\n", \
	#MEMBER, \
	(unsigned) sizeof(PTR->MEMBER), \
	(unsigned) offsetof(struct rpmfoo_s, MEMBER), \
	PTR->MEMBER)

static void
foofree(rpmfoo * _foop)
{
    rpmfoo foo = (rpmfoo) *_foop;
fprintf(stderr, "--> %s(%p) *p %p\n", __FUNCTION__, _foop, foo);

    PRINT(lu, foo, ui64);
    PRINT(u, foo, ui32);
    PRINT(u, foo, ui16);
    PRINT(u, foo, ui8);
    PRINT(p, foo, data);

    if (foo)
	free(foo);
}

#define	CPUID_PRINT(ATTR) \
    fprintf(stderr, "%19s: %s\n", #ATTR, \
	(__builtin_cpu_supports(#ATTR) ? "YES" : "NO" ))

struct cpuid_regs_s {
    unsigned ret;
    unsigned eax;
    unsigned ebx;
    unsigned ecx;
    unsigned edx;
};
static struct cpuid_regs_s r;

static const char * cpuid_vendor = "unknown";
#define	CPUID_VENDOR(VENDOR) \
    if (__builtin_cpu_is(#VENDOR)) \
	cpuid_vendor = #VENDOR

static int tagGet_switch = 2;

#ifndef __cplusplus
static int
tagGet1(tagData tag)
{
fprintf(stderr, "--> %s\n", __FUNCTION__);
    return 1;
}
static int
tagGet2(tagData tag)
{
fprintf(stderr, "--> %s\n", __FUNCTION__);
    return 2;
}

extern int tagGet(tagData tag);
__asm__ (".type tagGet, @gnu_indirect_function");

typeof(tagGet) * tagGet_dispatch (void)  {
    return (tagGet_switch == 2 ? tagGet2 : tagGet1);
}

#else

int tagGet (tagData tag) __attribute__ ((ifunc ("tagGet_dispatch")));

static int
tagGet1(tagData tag)
{
fprintf(stderr, "--> %s\n", __FUNCTION__);
    return 1;
}
static int
tagGet2(tagData tag)
{
fprintf(stderr, "--> %s\n", __FUNCTION__);
    return 2;
}

extern "C" {
  typeof(tagGet) * tagGet_dispatch (void)
  {

    __builtin_cpu_init();

    r.ret = __get_cpuid(0x80000000,
	&r.eax,
	&r.ebx,
	&r.ecx,
	&r.edx);

#if 0
    CPUID_VENDOR(intel);
    CPUID_VENDOR(atom);
    CPUID_VENDOR(core2);
    CPUID_VENDOR(corei7);
    CPUID_VENDOR(nehalem);
    CPUID_VENDOR(westmere);
    CPUID_VENDOR(sandybridge);
    CPUID_VENDOR(amd);
    CPUID_VENDOR(amdfam10h);
    CPUID_VENDOR(barcelona);
    CPUID_VENDOR(shanghai);
    CPUID_VENDOR(istanbul);
    CPUID_VENDOR(btver1);
    CPUID_VENDOR(amdfam15h);
    CPUID_VENDOR(bdver1);
    CPUID_VENDOR(bdver2);
    CPUID_VENDOR(bdver3);
    CPUID_VENDOR(btver2);
#endif

    return (tagGet_switch == 2 ? &tagGet2 : &tagGet1);
  }
}

#endif

static void printCpuidPage(unsigned page)
{
    char b[64];
    size_t nb = 0;
    r.ret = __get_cpuid(page, &r.eax, &r.ebx, &r.ecx, &r.edx);
    
    if (r.ret && r.eax > 0x0) {
	int needeol = 0;
	unsigned max = r.eax;
	for (; page <= max; page++) {
	    r.ret = __get_cpuid(page,
				&r.eax,
				&r.ebx,
				&r.ecx,
				&r.edx);
	    if (r.ret == 0)
		continue;
fprintf(stderr, "eax in: %08x eax = %08x ebx = %08x ecx = %08x edx = %08x\n",
	page, r.eax, r.ebx, r.ecx, r.edx);
	    switch (page) {
	    default:
		break;
	    case 0x00000000:
		fprintf(stderr, "%19s: ", "genuine");
		fprintf(stderr, "%.*s", 4, (char *)&r.ebx);
		fprintf(stderr, "%.*s", 4, (char *)&r.edx);
		fprintf(stderr, "%.*s\n", 4, (char *)&r.ecx);
		break;
	    case 0x80000002:
	    case 0x80000003:
	    case 0x80000004:
		if ((nb+4*4) <= sizeof(b))
		    memcpy(b+nb, &r.eax, 4*4);
		nb += 4*4;
		break;
	    }
	}
	if (nb > 0)
	    fprintf(stderr, "%19s: %.*s\n", "branding", (int)nb, b);
    }
}

static void
printCpuid(void)
{
fprintf(stderr, "--> %s\n", __FUNCTION__);

#if 0
    __builtin_cpu_init();
#endif

    CPUID_PRINT(cmov);
    CPUID_PRINT(mmx);
    CPUID_PRINT(popcnt);
    CPUID_PRINT(sse);
    CPUID_PRINT(sse2);
    CPUID_PRINT(sse3);
    CPUID_PRINT(ssse3);
    CPUID_PRINT(sse4.1);
    CPUID_PRINT(sse4.2);
    CPUID_PRINT(avx);
    CPUID_PRINT(avx2);

    CPUID_PRINT(sse4a);
    CPUID_PRINT(fma4);
    CPUID_PRINT(xop);
    CPUID_PRINT(fma);
    CPUID_PRINT(avx512f);
    CPUID_PRINT(bmi);
    CPUID_PRINT(bmi2);

    CPUID_PRINT(aes);
    CPUID_PRINT(pclmul);
    CPUID_PRINT(avx512vl);
    CPUID_PRINT(avx512bw);
    CPUID_PRINT(avx512dq);
    CPUID_PRINT(avx512cd);
    CPUID_PRINT(avx512er);
    CPUID_PRINT(avx512pf);
    CPUID_PRINT(avx512vbmi);
    CPUID_PRINT(avx512ifma);

#if 0	/* IBM s390 */
    CPUID_PRINT(longdisplacemnt);
    CPUID_PRINT(etf2);
    CPUID_PRINT(etf3);
    CPUID_PRINT(dfp);
    CPUID_PRINT(prefetch);
    CPUID_PRINT(storeclockfast);
    CPUID_PRINT(loadstoreoncond);
    CPUID_PRINT(popcount);
    CPUID_PRINT(interlocked);
    CPUID_PRINT(tx);
    CPUID_PRINT(dfpzoned);
    CPUID_PRINT(vector128);
#endif

    printCpuidPage(0x00000000);
    printCpuidPage(0x80000000);
    printCpuidPage(0x80860000);
    printCpuidPage(0xc0000000);

#define	CPUID_PRINT_VENDOR(VENDOR) \
    if (__builtin_cpu_is(#VENDOR)) \
	fprintf(stderr, "%19s: %s\n", "vendor", #VENDOR)

    CPUID_PRINT_VENDOR(intel);
    CPUID_PRINT_VENDOR(atom);
    CPUID_PRINT_VENDOR(core2);
    CPUID_PRINT_VENDOR(corei7);
    CPUID_PRINT_VENDOR(nehalem);
    CPUID_PRINT_VENDOR(westmere);
    CPUID_PRINT_VENDOR(sandybridge);
    CPUID_PRINT_VENDOR(amd);
    CPUID_PRINT_VENDOR(amdfam10h);
    CPUID_PRINT_VENDOR(barcelona);
    CPUID_PRINT_VENDOR(shanghai);
    CPUID_PRINT_VENDOR(istanbul);
    CPUID_PRINT_VENDOR(btver1);
    CPUID_PRINT_VENDOR(amdfam15h);
    CPUID_PRINT_VENDOR(bdver1);
    CPUID_PRINT_VENDOR(bdver2);
    CPUID_PRINT_VENDOR(bdver3);
    CPUID_PRINT_VENDOR(btver2);
}

int
main(int argc, char *argv[])
{
    int ec RPM_GNUC_ANNOTATE(ec) = -1;
    char * b RPM_GNUC_CLEANUP(xfree) = (char *) xmalloc(10);
    size_t ndata = 8;
    rpmfoo foo RPM_GNUC_CLEANUP(foofree) = (rpmfoo) xmalloc(sizeof(*foo)+ndata);

    struct rpmfoo_s _foo = {};
    (void)_foo;

    uint8_t ui8;
    uint16_t ui16;
    uint32_t ui32;
    uint64_t ui64;
    uint128_t ui128;
    char str[] = "str";
    argv_t av = { "foo", "bar", NULL};

    printCpuid();

    fprintf(stderr, "vendor(%s) tagGet(%d)\n", cpuid_vendor, tagGet(foo->p));
#ifndef	__cplusplus
    tagGet(foo->p.ui32p);
    tagGet(&ui8);
    tagGet(&ui16);
    tagGet(&ui32);
    tagGet(&ui64);
    tagGet(&ui128);
    tagGet(str);
    tagGet(av);
#endif

    *foo = _foo;	/* structure assignment */

    foo->i = 0;
    foo->ui64 = 64;
    foo->ui32 = 32;
    foo->ui16 = 16;
    foo->ui8 = 8;
    for (int i = 0; i < (int)ndata; i++)
	foo->data[i] = i;

    return ec;
}
